/******************************************************************************
  aux.c -- Auxiliary functions
  
  Author: Mark Baranowski (markb@elen.utah.edu)

  Last significant change: July 17, 2000
  
  These functions are provided "as is" in the hopes that they might serve some
  higher purpose.  If you want these functions to serve some lower purpose,
  then that too is all right.  So as to keep these hopes alive you may
  freely distribute these functions so long as this header remains intact.
  You may also freely distribute modified versions of these functions so long
  as you indicate that such versions are modified and so long as you
  provide access to the unmodified original copy.
  
  Note: The most recent version of these functions may be obtained from
  ftp://auriga.elen.utah.edu/pub/markb/els.tar.Z

  ****************************************************************************/

#include "sysdefs.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "aux.h"

extern void give_usage_or_help(char U_or_H);


void *mem_alloc(int size)
{
  char *ptr;
  
  ptr = malloc(size);
  if (ptr == NULL)
  {
    fprintf(stderr, "%s: Unable to allocate working memory\n",
	    Progname);
    exit(1);
  }
  
  return((void *)ptr);
}


void mem_free(void *mem)
{
  free(mem);
  return;
}


/* strdup_ns: same as strdup, except duplicate string contains No Spaces */
char *strdup_ns(register const char *str)
{
  char *newstr = mem_alloc(strlen(str)+1);
  register char *dup = newstr;
  while (*str != CNULL)
  {
    if (!isspace(*str)) *dup++ = *str;
    str++;
  }
  *dup = CNULL;
  return(newstr);
}


/* strcmp_ci: same as strcmp, except comparisons are case-insensitive */
int strcmp_ci(register const char *s1,
	      register const char *s2)
{
  register char c1, c2;
  int diff = 0;
  do
  {
    c1 = *s1++;
    c2 = *s2++;
    if (isupper(c1)) c1 = tolower(c1);
    if (isupper(c2)) c2 = tolower(c2);
  } while ((diff = c1 - c2) == 0 && c1 && c2);
  return(diff);
}


/* strncmp_ci: same as strncmp, except comparisons are case-insensitive */
int strncmp_ci(register const char *s1,
	       register const char *s2,
	       register int n)
{
  register char c1, c2;
  int diff = 0;
  do
  {
    if (n-- <= 0) break;
    c1 = *s1++;
    c2 = *s2++;
    if (isupper(c1)) c1 = tolower(c1);
    if (isupper(c2)) c2 = tolower(c2);
  } while ((diff = c1 - c2) == 0 && c1 && c2);
  return (diff);
}


/* Not all unixes have strtoul, so here's my version: */
Local Ulong _strtoul(const char *str, char **ptr, int base,
		     Boole scan_ws, Boole is_signed,
		     Boole *negative, Boole *overflow)
{
  int digit;
  Ulong result, last_result;
  const char *next = str;

  result = 0;
  last_result = 0;
  *negative = FALSE;
  *overflow = FALSE;
  
  /* Scan leading whitespace only if specified: */
  if (scan_ws) while (isspace(*str)) str++;
  
  if (*str == '-')
  {
    *negative = TRUE;
    str++;
  }
  else if (*str == '+')
    str++;
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * Base may be either explicitly specified (i.e. the parameter "base" is
   * non-zero) or implicitly discerned (e.g. the number starts with 0x,
   * etc.).  If the base is explicit then the number may or may not be
   * preceded with an implicit base; however, if both an implicit and an
   * explicit base are specified they both need to agree.
   */
  
  if (*str == '0')
  {
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Attempt to discern an implicit base:
     *	0x... --> implicit base = 16
     *	0...  --> implicit base = 8
     */
    str++;
    next = str;
    if ((base == 0 || base == 16) && (*str == 'x' || *str == 'X'))
    {
      str++;
      base = 16;
    }
    else if ((base == 0 || base == 8) && (*str != '\0'))
    {
      base = 8;
    }
    else
    {
      /* We have either a single zero or an explicit base: */
      str--;	/* restore the zero, we may need it later */
      next = str;
    }
  }
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * If no explicit base was specified and no implicit base was discerned,
   * then use base 10:
   */
  if (base == 0) base = 10;

  if (base < 2 || base > 36)
    errno = EINVAL;
  else
  {
    Boole digit_ok;
    do {
      digit = base;
      if (isascii(*str))
      {
	if (isdigit(*str))
	  digit = *str - '0';
	else if (islower(*str))
	  digit = *str - 'a' + 10;
	else if (isupper(*str))
	  digit = *str - 'A' + 10;
      }
      digit_ok = (digit < base);
      
      if (digit_ok)
      {
	result *= base;
	result += digit;
	str++;
	next = str;
	if (result / base != last_result) *overflow = TRUE;
	if (is_signed && (long)result < 0) *overflow = TRUE;
	last_result = result;
      }
    } while (digit_ok && *str);
    
    if (*negative) result = (Ulong)(-(long)result);
  }
    
  /* Certain compilers complain about de-consting "next" even though the
     pointer VARIABLE itself is non-const!: */
  if (ptr) *ptr = (char *)next;
  
  return(result);
}


long strtol(const char *str, char **ptr, int base)
{
  Boole negative, overflow;
  long l;
  l = (long)_strtoul(str, ptr, base, TRUE, TRUE, &negative, &overflow);
  if (overflow)
  {
    /* NB: certain processors (e.g. DEC_ALPHA) *require* 1L */
    long maxneg = 1L << (8*sizeof(long)-1);
    errno = ERANGE;
    return(negative ? maxneg : ~maxneg);
  }
  else
    return(l);
}


Ulong strtoul(const char *str, char **ptr, int base)
{
  Boole negative, overflow;
  Ulong ul;
  ul = _strtoul(str, ptr, base, TRUE, FALSE, &negative, &overflow);
  if (overflow)
  {
    errno = ERANGE;
    return(~(Ulong)0);
  }
  else
    return(ul);
}


#ifdef HAVE_STAT64
/* Not all unixes have strtoull, so here's my version: */
Local unsigned long long _strtoull(const char *str, char **ptr, int base,
				   Boole scan_ws, Boole is_signed,
				   Boole *negative, Boole *overflow)
{
  int digit;
  unsigned long long result, last_result;
  const char *next = str;

  result = 0;
  last_result = 0;
  *negative = FALSE;
  *overflow = FALSE;
  
  /* Scan leading whitespace only if specified: */
  if (scan_ws) while (isspace(*str)) str++;
  
  if (*str == '-')
  {
    *negative = TRUE;
    str++;
  }
  else if (*str == '+')
    str++;
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * Base may be either explicitly specified (i.e. the parameter "base" is
   * non-zero) or implicitly discerned (e.g. the number starts with 0x,
   * etc.).  If the base is explicit then the number may or may not be
   * preceded with an implicit base; however, if both an implicit and an
   * explicit base are specified they both need to agree.
   */
  
  if (*str == '0')
  {
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Attempt to discern an implicit base:
     *	0x... --> implicit base = 16
     *	0...  --> implicit base = 8
     */
    str++;
    next = str;
    if ((base == 0 || base == 16) && (*str == 'x' || *str == 'X'))
    {
      str++;
      base = 16;
    }
    else if ((base == 0 || base == 8) && (*str != '\0'))
    {
      base = 8;
    }
    else
    {
      /* We have either a single zero or an explicit base: */
      str--;	/* restore the zero, we may need it later */
      next = str;
    }
  }
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * If no explicit base was specified and no implicit base was discerned,
   * then use base 10:
   */
  if (base == 0) base = 10;

  if (base < 2 || base > 36)
    errno = EINVAL;
  else
  {
    Boole digit_ok;
    do {
      digit = base;
      if (isascii(*str))
      {
	if (isdigit(*str))
	  digit = *str - '0';
	else if (islower(*str))
	  digit = *str - 'a' + 10;
	else if (isupper(*str))
	  digit = *str - 'A' + 10;
      }
      digit_ok = (digit < base);
      
      if (digit_ok)
      {
	result *= base;
	result += digit;
	str++;
	next = str;
	if (result / base != last_result) *overflow = TRUE;
	if (is_signed && (long long)result < 0) *overflow = TRUE;
	last_result = result;
      }
    } while (digit_ok && *str);
    
    if (*negative) result = (unsigned long long)(-(long long)result);
  }
    
  /* Certain compilers complain about de-consting "next" even though the
     pointer VARIABLE itself is non-const!: */
  if (ptr) *ptr = (char *)next;
  
  return(result);
}


long long strtoll(const char *str, char **ptr, int base)
{
  Boole negative, overflow;
  long long l;
  l = (long long)_strtoull(str, ptr, base, TRUE, TRUE, &negative, &overflow);
  if (overflow)
  {
    /* NB: certain processors (e.g. DEC_ALPHA) *require* 1LL */
    long long maxneg = 1LL << (8*sizeof(long long)-1);
    errno = ERANGE;
    return(negative ? maxneg : ~maxneg);
  }
  else
    return(l);
}


unsigned long long strtoull(const char *str, char **ptr, int base)
{
  Boole negative, overflow;
  unsigned long long ul;
  ul = _strtoull(str, ptr, base, TRUE, FALSE, &negative, &overflow);
  if (overflow)
  {
    errno = ERANGE;
    return(~(unsigned long long)0);
  }
  else
    return(ul);
}
#endif /*HAVE_STAT64*/


/******************************************************************************
 * envSet()/envUnset(): set and unset environment variables
 *
 * envSet() acts the same and returns the same value putenv(), plus it also
 * tries to recycle any env slots that were made available by envUnset().
 * envUnset returns 0 if the env var is unset, otherwise it returns 1.
 */

/* Define the following if you want to allow spaces to precede '=' in
   the var name, otherwise spaces trailing the var name are considered
   part of the name.  Some OSes (e.g. VxWorks5.3.1) treat putenv("X  =1")
   to be the same as putenv("X=1"), whereas other OSes (e.g. SunOS5.7) treat
   these as two separate vars named "X  " and "X". */
/*#define IGNORE_SPACES_FOLLOWING_VARNAME*/

extern char **environ;
Local char *cannon_fodder = "CaNnOn_FoDdEr=0";
Local char **recyclable_env_slot = NULL;

int envSet(char *var)
{
  char **e;
  Boole found;
  int i, sts = 0;

  found = FALSE;
  for (e = environ, i = 0;  *e != NULL && !found;  i++, e++)
  {
    Fast char *cp = *e, *vp = var;
    while (*cp && *vp && *cp != '=' && *vp == *cp) cp++, vp++;
# ifdef IGNORE_SPACES_FOLLOWING_VARNAME
    while (isspace(*cp)) cp++;
    while (isspace(*vp)) vp++;
# endif
    found = (*vp == *cp && (*vp == '\0' || *vp == '='));
  }

  if (found)
  {
    /* Replace existing variable: */
    i--;
    environ[i] = var;
  }
  else if (recyclable_env_slot != NULL)
  {
    /* Recycle last remembered env slot: */
    *recyclable_env_slot = var;
    recyclable_env_slot = NULL;
  }
  else
  {
    /* Look for any recyclable env slots that might have been forgotten: */
    found = FALSE;
    for (e = environ, i = 0;  *e != NULL && !found;  i++, e++)
      found = (*e == cannon_fodder);
    
    if (found)
    {
      /* A recyclable env slot was discovered: */
      i--;
      environ[i] = var;
    }
    else
      /* No recyclable env slots found; create a new variable: */
      sts = putenv(var);
  }
  return(sts);
}


int envUnset(char *var)
{
  char **e;
  Boole found;
  int i, sts = 1;

  /* Note: If var not found, or if var itself is NULL then it might not
     be considered a serious error as it will essentially be unset. */

  if (var != NULL)
  {
    found = FALSE;
    for (e = environ, i = 0;  *e != NULL && !found;  i++, e++)
    {
      Fast char *cp = *e, *vp = var;
      while (*cp && *vp && *cp != '=' && *vp == *cp) cp++, vp++;
#   ifdef IGNORE_SPACES_FOLLOWING_VARNAME
      while (isspace(*cp)) cp++;
      while (isspace(*cp)) vp++;
#   endif
      found = (*vp == '\0' && *cp == '=');
    }
    
    if (found)
    {
      /* Unset an existing variable by marking it as fodder (it might be
	 easier to just set it to '\0', but some OSes might get disturbed
	 by this): */
      i--;
      environ[i] = cannon_fodder;
      recyclable_env_slot = &environ[i];
      sts = 0;
    }
  }    
  return(sts);
}


char *find_first_occurrence(char *look_for, char *str)
{
  char *found = NULL;
  while (*str && (found = strchr(look_for, *str)) == NULL) str++;
  return(found);
}


void carrot_msg(char *tag, char *opt, char *arg, char *msg, char *ptr)
{
  char preface[128];
  sprintf(preface, "%s: %s%s%s",
	  Progname,
	  tag ? tag : "",
	  tag ? ": " : "",
	  opt ? opt : "");
  if (arg)
  {
    int carrot, indent;
    char *qarg, *qptr = ptr;
    if (ptr == NULL || ptr < arg) ptr = arg; /*JIC*/

    if ((qarg = quote_str(arg, &qptr, QS_ARG)) != NULL)
    {
      fprintf(stderr, "%s%s\n", preface, qarg);
      carrot = strlen(preface) + ((Ulong)qptr - (Ulong)qarg);
    }
    else
    {
      fprintf(stderr, "%s%s\n", preface, arg);
      carrot = strlen(preface) + ((Ulong)ptr - (Ulong)arg);
    }

    indent = carrot - strlen(msg);
    if (indent > 0 && indent < 256)
      fprintf(stderr, "%*c%s^\n", indent, ' ', msg);
    else if (carrot > 0 && carrot < 256)
      fprintf(stderr, "%*c^%s\n", carrot, ' ', msg);
    else
      fprintf(stderr, "%s\n", msg);
  }
  else
    fprintf(stderr, "%s: %s\n", preface, msg);

  return;
}


#define FNAME_TROUBLE(c) \
((!isgraph(c) && (c) != ' ') || (dash_Q && IS_MEMBER(c, "\\\"")))
/* FNAME_TROUBLE determines set of troublesome characters that are to be
   printed in octal and/or c-style, e.g.:
   dash_b: BAD<tab>FILE --> BAD\011FILE	
   dash_Q: B\A\D<tab>FILE --> "B\\A\\D\tFILE"
   dash_Q: A B\A\D<tab>"FILE" --> "A B\\A\\D\t\"FILE\""
   */

char *quote_fname(char *fname, Boole dash_b, Boole dash_Q, Boole plus_q)
{
  if (dash_b || dash_Q)
  {
    char *cp = fname;
    int trouble = 0;

    while (*cp != CNULL)
    {
      if (FNAME_TROUBLE(*cp)) trouble++;
      cp++;
    }
    if (dash_Q || trouble > 0)
    {
      static char *last_qstr = NULL;
      static int last_qstr_len = 0;
      int qstr_len;
      char *qstr;
      
      qstr_len = strlen(fname) + 2 + 3*trouble + 1;
      if (last_qstr != NULL && last_qstr_len >= qstr_len)
      {
	qstr_len = last_qstr_len;
	qstr = last_qstr;
      }
      else
      {
	if (last_qstr != NULL) mem_free(last_qstr);
	qstr = mem_alloc(qstr_len);
	/*printf("qstr_len, qstr: %d, %x\n", qstr_len, (Ulong)qstr);*/
      }
      
      last_qstr_len = qstr_len;
      last_qstr = qstr;
      
      cp = fname;
      if (dash_Q) *qstr++ = '"';
      while (trouble > 0 && *cp != CNULL)
      {
	char c_char;
	if (FNAME_TROUBLE(*cp))
	{
	  trouble--;
	  c_char = CNULL;
	  if (dash_Q)
	  {
	    if (*cp == '\0') c_char = '0';
	    if (*cp == '\\') c_char = '\\';
	    if (*cp == '"')  c_char = '"';
	    if (*cp == '\b') c_char = 'b';
	    if (*cp == '\n') c_char = 'n';
	    if (*cp == '\r') c_char = 'r';
	    if (*cp == '\t') c_char = 't';
	  }
	  if (c_char != CNULL)
	  {
	    sprintf(qstr, "\\%c", c_char);
	    qstr += 2;
	  }
	  else
	  {
	    sprintf(qstr, "\\%03o", *cp);
	    qstr += 4;
	  }
	}
	else
	{
	  *qstr++ = *cp;
	}
	cp++;
      }
      sprintf(qstr, "%s%s", cp, dash_Q ? "\"" : "");

      if (plus_q)
      {
	char *q = quote_str(last_qstr, NULL, QS_FILE);
	if (q != NULL) return(q);
      }
      return(last_qstr);
    }
    else
      return(NULL);
  }

  else if (plus_q)
    return(quote_str(fname, NULL, QS_FILE));

  else
    return(NULL);
}


#define STR_TROUBLE(c) IS_MEMBER(c, "$`\\\"'!")
/* STR_TROUBLE determines set of troublesome characters that can't be
   contained inside single quotes, e.g.:
   BAD`FILE'NAME    -->  BAD\`FILE\'NAME
   A BAD FILE NAME  -->  'A BAD FILE NAME'
   A BAD`FILE'NAME  -->  'A BAD'\`'FILE'\''NAME'
   !pwd    -->  \!pwd
   \!*pwd  -->  '\'\!'*pwd'
   a!=b    -->  a!=b
   a ! b   -->  'a ! b'

   Note: some strings echo differently under various shells.
   sh and bash:  echo '\!pwd' --> \!pwd
   csh and tcsh: echo '\!pwd' --> !pwd
   csh and tcsh: echo a!=b    --> a!=b	 (does not require \ or ' quotes!)
   csh and tcsh: echo a ! b   --> a ! b	 (does not require \ or ' quotes!)
   */
char *quote_str(char *str, char **ptr, Ulong mode)
{
  char *cp = str;
  char *sptr = (ptr != NULL ? *ptr : NULL);
  Boole pre;
  Boole quote;
  int trouble = 0;

  /* If file, the following characters need prefacing: */
  pre = ((mode & QS_FILE) && IS_MEMBER(*cp, "-+"));

  /* The following leading characters need quoting: */
  quote = IS_MEMBER(*cp, "~#");

  /* Look for troublesome characters.  Also, if any characters from the
     special set are detected then the entire string needs quoting: */
  while (quote == FALSE && *cp != CNULL)
  {
    quote = !isgraph(*cp) || IS_MEMBER(*cp, "{}[]()<>&*|?;$\\\"'");
    if (STR_TROUBLE(*cp)) trouble++;
    cp++;
  }

  /* Finish looking for troublesome characters: */
  while (*cp != CNULL)
  {
    if (STR_TROUBLE(*cp)) trouble++;
    cp++;
  }

  if (pre || quote || trouble > 0)
  {
    static char *last_qstr = NULL;
    static int last_qstr_len = 0;
    int qstr_len;
    char *qstr;
    char *qptr = NULL;
    char qchar, *qnext;

    /* Worst case: qstr_len = strlen(str) + strlen("./") +
       sizeof(start_quote) + sizeof(final_quote) + 
       sizeof(trouble_overhead) + sizeof(CNULL); */
    qstr_len = strlen(str) + 4 + 3*trouble + 1;
    if (last_qstr != NULL && last_qstr_len >= qstr_len)
    {
      qstr_len = last_qstr_len;
      qstr = last_qstr;
    }
    else
    {
      if (last_qstr != NULL) mem_free(last_qstr);
      qstr = mem_alloc(qstr_len);
      /*printf("qstr_len, qstr: %d, %x\n", qstr_len, (Ulong)qstr);*/
    }
    
    last_qstr_len = qstr_len;
    last_qstr = qstr;

    cp = str;

    qchar = '\'';
    if ((qnext = find_first_occurrence("$`\\\"'", cp)) != NULL)
      if (*qnext == '\'') qchar = '"';

    if (quote) *qstr++ = qchar;
    if (pre) {*qstr++ = '.'; *qstr++ = '/';}
    while (trouble > 0 && *cp != CNULL)
    {
      if (STR_TROUBLE(*cp))
      {
	trouble--;
	if ((quote && ((qchar == '\'' && IS_MEMBER(*cp, "$`\\\"")) ||
		       (qchar == '"'  && IS_MEMBER(*cp, "'"))
		       )) ||
	    (*cp == '!' && IS_MEMBER(*(cp+1), "= ")))  /* NO quotes needed! */
	{
	  /* No trouble after all: */
	  if (cp == sptr) qptr = qstr;
	  *qstr++ = *cp;
	}
	else
	{
	  if (quote) *qstr++ = qchar;

	  *qstr++ = '\\';
	  if (cp == sptr) qptr = qstr;
	  *qstr++ = *cp;

	  qchar = '\'';
	  if ((qnext = find_first_occurrence("$`\\\"'", cp+1)) != NULL)
	    if (*qnext == '\'') qchar = '"';

	  if (quote) *qstr++ = qchar;
	}
      }
      else
      {
	if (cp == sptr) qptr = qstr;
	*qstr++ = *cp;
      }
      cp++;
    }
    if (qptr == NULL) qptr = qstr + (sptr - cp);
    sprintf(qstr, "%s%c", cp, quote ? qchar : CNULL);

    if (ptr != NULL) *ptr = qptr;
    return(last_qstr);
  }
  else
    return(NULL);
}


/* Display error message and exit: */
void error_msg(char *msg)
{
  fprintf(stderr, "%s: %s\n", Progname, msg);
  exit(1);
}


/* Display errno message and exit: */
void error_num(char *file)
{
  char msg[256];
  if (file != NULL)
    sprintf(msg, "%s: %s", Progname, file);
  else
    sprintf(msg, "%s", Progname);
  if (errno != 0)
    perror(msg);
  else
    fprintf(stderr, "%s: Unknown error\n", msg);
  exit(1);
}


extern char *Current_Arg;
void opt_error_msg(char *msg, char *ptr)
{
  fprintf(stderr, "\n");
  carrot_msg(NULL, NULL, Current_Arg, msg, ptr-1);
  give_usage_or_help('U');
  exit(1);
}


void arg_error_msg(char *msg, char *arg, char *ptr)
{
  carrot_msg(NULL, NULL, arg, msg, ptr-1);
  exit(1);
}


void usage_error(void)
{
  give_usage_or_help('U');
  exit(1);
}
