/******************************************************************************
  els.c -- An Enhanced LS look-alike with many additional features.
  
  Author: Mark Baranowski (markb@elen.utah.edu) and James M. Gleason

  Last significant change: March 6, 2002
  
  This program is provided "as is" in the hopes that it might serve some
  higher purpose.  If you want this program to serve some lower purpose,
  then that too is all right.  So as to keep these hopes alive you may
  freely distribute this program so long as this header remains intact.
  You may also freely distribute modified versions of this program so long
  as you indicate that such versions are modified and so long as you
  provide access to the unmodified original copy.
  
  Acknowledgements and contributions:
  This program is based on an earlier program written by James M. Gleason.
  Access filtering idea contributed by Larry Gensch <gensch@zk3.dec.com>.
  AIX ACL feature contributed by Maurizio Sartori "Masar" (masar@ibm.it).

  Note: The most recent version of this program may be obtained from
  ftp://auriga.elen.utah.edu/pub/markb/els.tar.Z

  ****************************************************************************/

#include "version.h"
#include "sysdefs.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
#include <utime.h>
#include <dirent.h>
#include <sys/param.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#if defined(GEN_BSD) || defined(FREEBSD)
/* Generic BSD defines major/minor in sys/types.h */
#elif SUNOS >= 5
/* Something ain't quite right with SunOS5.7: */
#  define major(x)  ((x) >> 18)
#  define minor(x)  ((x) & 0x3ffff)
#else
/* Everybody else (including SunOS4!): */
#  include <sys/sysmacros.h>
#endif

#include "defs.h"
#include "getdate32.h"
#include "time32.h"
#include "aux.h"
#include "els.h"
#include "elsMisc.h"
#include "sysdep.h"
#include "phLib.h"
#include "format.h"

char LGC = ' '; /* LGC default value used by convert_iso8601 */


/* Add /usr/include/sys/stat.h defs if necessary: */
#if !defined(S_ISLNK)
#  define S_ISLNK(m)  (((m) & S_IFMT) == S_IFLNK)
#endif
#if !defined(S_ISDIR)
#  define S_ISDIR(m)  (((m) & S_IFMT) == S_IFDIR)
#endif
#if !defined(S_ISBLK)
#  define S_ISBLK(m)  (((m) & S_IFMT) == S_IFBLK)
#endif
#if !defined(S_ISCHR)
#  define S_ISCHR(m)  (((m) & S_IFMT) == S_IFCHR)
#endif
#if defined(SUNOS) && !defined(S_IFDOOR)
/* This definition will not upset SUNOS4: */
#  define S_IFDOOR  0xd000
#endif

BLKCNT_T dir_block_total;
int dir_file_count;


/********** Global Routines Referenced **********/

extern char *getenv();

/********** Global Routines Defined **********/

void do_options(char *options);
void do_options_minus(char *options);
void do_options_plus(char *options);
void stamp_setup(int pass);
void untouch_setup(int pass);
void give_usage_or_help(char U_or_H);
void analyze_options(void);
void append_dir(Dir_List *dlist, char *fname, int fname_size);
void copy_info(Dir_Item *dest, Dir_Item *src);
void read_dir(Dir_List *dlist, char *dname, DIR *dir);
void stat_dir(Dir_List *dlist, char *dname);
void name_dir(Dir_List *dlist, char *dname);
void sort_dir(Dir_List *dlist, char *dname);
void sort_dir_time(Dir_List *dlist, char *dname);
void sort_dir_name(Dir_List *dlist, char *dname);
void list_dir(Dir_List *dlist, char *dname);
void free_dir(Dir_List *dlist);
struct stat *getCwdInfo(char *path);
int getCwdBlocksize(char *path);
void filter_file(Dir_Item *file, char *path);
Boole filter_fexpr(Dir_Item *file, char *path, char *ptr, char **end);
Boole filter_access(Dir_Item *file, char *path, char *ptr, char **end);
Boole filter_type(Dir_Item *file, char *path, char *ptr, char **end);
Boole filter_perm(Dir_Item *file, char *path, char *ptr, char **end);
Boole filter_quantity(Dir_Item *file, char *path, char *ptr, char **end);
Boole filter_unusual(Dir_Item *file, char *path, char *ptr, char **end);
void list_file(Dir_Item *file, char *dname);
char *find_first_directive(char *list, char *fmt);
char *G_print(char *buff, char *fmt, char *dname, Dir_Item *file);
char *N_print(char *buff, char *fmt, char *dname, Dir_Item *file);
char *T_print(char *buff, char *fmt, time_t ftime,
	      struct tm *fdate, Boole gmt, Boole meridian);
char *T_print_width(char *buff, char *fmt, time_t ftime,
		    struct tm *fdate, Boole gmt, Boole meridian, int width);
char *separate(char *bp, char icase);
char *finish(char *bp);
char *full_name(char *dname, char *fname);
Dir_Item *dir_item_alloc(void);
void dir_item_free(Dir_Item *file);
void filter_error_msg(char *msg, char *ptr);
Boole filter_warning_msg(char *msg, char *ptr);
void redundancy_warning(char *types, char *ptr);
void expecting_error(Boole negate, char *items, Boole logic, Boole end,
		     char *ptr);
void watchSigHandler(int sig);
int sigSafe_stat(char *path, struct stat *buf);
int sigSafe_lstat(char *path, struct stat *buf);
void spin_wheel(void);

/********** Global Variables Defined **********/

int  Iarg;
int  Argc;
char **Argv;
char *Progname;
uid_t Whoami;
time_t The_Time;
time_t The_Time_in_an_hour;
char *Time_Zone = NULL;
int Debug = 0;
int MaxVersionLevel;
Boole ArgOrder = FALSE;
Boole TruncateId = FALSE;
Boole First_listing = TRUE;
char *Current_Opt = NULL;
char *Current_Arg = NULL;
Dir_Item *CwdItem = NULL;
int CwdBlocksize = 0;
char *CwdPath = ".";

/* Uncomment the desired signal type for +W: */
/*#define WATCH_SIGNAL  SIGINT*/	/* watch == ^C */
#define WATCH_SIGNAL  SIGQUIT		/* watch == ^\ (^| on some systems) */
Void_Function defaultSigHandler = NULL;
Boole sigEvent = FALSE;
Boole sigAskAbort = FALSE;
Boole sigAskNoMore = FALSE;

#define LISTING_ERROR   2
#define USAGE_ERROR     1
#define NORMAL_EXIT     0
Boole listingError = FALSE;

Boole zero_file_args;
Boole multiple_file_args;
Boole using_full_names;
Boole avoid_trimmings;
char first_mac;
int recursion_level = 0;

/* SEMANTICS: In general, when intrepreting options ELS leans more
 * towards SYS5 (with a little GNU sprinkled in) than it does towards
 * BSD.  On the other hand, when it comes to listing format and
 * spacing, ELS leans more towards BSD.
 *
 * Meaningful combinations are:
 * Flags  Bits             Behavior
 * +4     SEM_ELS|SEM_BSD  Interpret options a la BSD, output a la ELS.
 * +5     SEM_ELS|SEM_SYS5 Interpret options a la SYS5, output a la ELS.
 * +g     SEM_ELS|SEM_GNU  Interpret options a la GNU, output a la ELS.
 * +l4    SEM_LS|SEM_BSD   Behave as much like BSD's /bin/ls as is reasonable.
 * +l5    SEM_LS|SEM_SYS5  Behave as much like SYS5's /bin/ls as is reasonable.
 * +lg    SEM_LS|SEM_GNU   Behave as much like GNU's /bin/ls as is reasonable.
 * +l     SEM_LS           Output as much like native /bin/ls as is reasonable.
 * none   SEM_ELS          Interpret options a la SYS5 (except for
 *                         -g, -G, and -s options), output a la ELS.
 */

#define SEM_ELS  0x01
#define SEM_LS   0x02
#define SEM_BSD  0x04
#define SEM_SYS5 0x08
#define SEM_GNU  0x10
Uint Sem;

Boole list_hidden;
Boole list_dotdir;
Boole list_long;
Boole list_long_numeric;
Boole list_long_omit_gid;
Boole list_atime;
Boole list_ctime;
Boole list_inode;
Boole list_size_in_blocks;
Boole mark_files;
Boole mark_dirs;
Boole dash_b;
Boole dash_Q;
Boole recursive;
Boole reverse_sort;
Boole time_sort;
Boole unsort;
Boole g_flag;
Boole G_flag;
Boole expand_symlink;
Boole expand_directories;
Boole list_directories;
Boole plus_q;
Boole traverse_mp;
Boole traverse_expanded_symlink;
Boole traverse_specified;
Boole warning_suppress;
Boole watch_progress;
FILE *ttyin = NULL, *ttyout = NULL;
Boole zero_st_info;
Ulong zero_st_mask;
#define ZERO_DIR_MTIME   0001
#define ZERO_DIR_ATIME   0002
#define ZERO_DIR_SIZE    0004
#define ZERO_LINK_MTIME  0010
#define ZERO_LINK_ATIME  0020
#define ZERO_LINK_OWNER  0040
#define ZERO_DEV_MTIME   0100
#define ZERO_DEV_ATIME   0200

Boole G_option = FALSE;
Boole N_option = FALSE;
Boole T_option = FALSE;

char *G_format;
#define INODE			'i'
#define TYPE_IN_ALPHA		't'
#define TYPE_IN_SYMBOLIC	'T'
#define PERMISSION_IN_ALPHA	'p'
#define PERMISSION_IN_NUMERIC	'P'
#define PERMISSION_IN_SYMBOLIC	'M'
#define ACL_INDICATOR		'A'
#define SIZE_IN_BYTES		's'
#define SIZE_IN_BLOCKS		'S'
#define TIME_MODIFIED		'm'
#define TIME_ACCESSED		'a'
#define TIME_MODE_CHANGED	'c'
#define UID_IN_ALPHA		'u'
#define UID_IN_NUMERIC		'U'
#define GID_IN_ALPHA		'g'
#define GID_IN_NUMERIC		'G'
#define _16_BIT_CHECKSUM	'C'
#define NAME_FORMAT		'n'
#define FULL_NAME_FORMAT	'N'
#define LINK_COUNT		'l'
#define LINK_NAME		'L'

char *N_format;
#define LINK_NAME2	'l'
#define QUOTE_NAME	'q'
#define FULL_NAME	'F'
#define DIR_NAME	'd'
#define FILE_NAME	'f'

char *T_format;
#define ABS_TIME	'a'
#define DELTA_TIME	'd'
#define REL_TIME	'r'
#define ALPHA_DATE	'A'
#define NUM_DATE	'N'
#define GMT_DATE	'G'
#define LOCAL_DATE	'L'
#define MERIDIAN_TIME	'M'

#define ISO8601_DATE	'I'
#define ELS_DATE	'e'
#define LS_DATE		'l'
#define DOS_DATE	'd'
#define WINDOWS_DATE	'w'
#define VERBOSE_DATE	'v'

#define YEARS		'Y'
#define MONTHS		'M'
#define WEEKS		'W'
#define DAYS		'D'
#define HOURS		'h'
#define MINS		'm'
#define SECS		's'
#define TIME_2		't'	/* h:m */
#define TIME_3		'T'	/* h:m:s */
#define TIME_OR_YEAR	'Q'
#define YEARS_MOD_100	'y'
#define ZONE_NAME	'Z'

#define MAX_FEXPR_FILTER  16
int fexpr_count;
char *fexpr_filter[MAX_FEXPR_FILTER];
#define fexpr_FNOT	'~'
#define fexpr_FAND	','
#define fexpr_FOR	':'
#define fexpr_NOT	'!'
#define fexpr_AND	'&'
#define fexpr_OR	'|'
#define fexpr_BEGIN	'{'
#define fexpr_END	'}'

#define type_REG	'r'
#define type_DIR	'd'
#define type_CHR	'c'
#define type_BLK	'b'
#define type_FIFO	'p'
#define type_LNK	'l'
#define type_SOCK	's'
#define type_DOOR	'D'
#define type_SPECIAL	'S'	/* c:b */
#define symtype_REG	'-'
#define symtype_DIR	'/'
#define symtype_FIFO	'|'
#define symtype_LNK	'@'
#define symtype_SOCK	'='
#define symtype_DOOR	'>'

/* List of wildcard patterns for "E"xcluding file/dir names: */
#define MAX_EXC_FILTER  16
int exc_file_count;
char *exc_file_filter[MAX_EXC_FILTER];
int exc_dir_count;
char *exc_dir_filter[MAX_EXC_FILTER];

/* List of wildcard patterns for "I"ncluding file/dir names: */
#define MAX_INC_FILTER  16
int inc_file_count;
char *inc_file_filter[MAX_INC_FILTER];
int inc_dir_count;
char *inc_dir_filter[MAX_INC_FILTER];

int munge = 0;
int untouch = 0;
Boole filtering = FALSE;
Boole stamping = FALSE;
char stamp_option;
Boole ver_level_given = FALSE;
Boole verbose_mode = FALSE;
Boole execute_mode = FALSE;
Boole execute_symlinks = FALSE;
Boole execute_sts_check = FALSE;
int execute_sts_good = 0;

char output_buff[MAX_OUTPUT];
int  output_nlines = 0;

int squeeze_cnt = 0;
Boole squeeze = FALSE;
Boole fielding = FALSE;
Boole as_is;
Boole separated;
char separator = ' ';

/* Static messages: */
char *NON_NEGATABLE = "Non-negatable option";
char *TOO_MANY_FILTERS = "Too many filters specified";

/*****************************************************************************/

int main(int argc, char *argv[])
{
  char *dname;
  Dir_List dlist;
  
  Iarg = 1;
  Argc = argc;
  Argv = argv;
  
  /* Simplify access to the program name */
  {
    char *cp;
    if ((cp = strrchr(Argv[0], '/')) != NULL)
      Progname = cp + 1;
    else
      Progname = Argv[0];
  }

  Whoami = geteuid();
  
  /* Get the the current time just once during the program: */
  The_Time = time(NULL);
  The_Time_in_an_hour = The_Time + SECS_PER_HOUR;

#ifdef Wformat
  fprintf(stderr, "%s: Warning: compiled with -DWformat (zero_pad disabled)\n",
	  Progname);
#endif

  /* Set any debugging bits: */
  {
    char *debug = getenv("ELS_DEBUG");
    if (debug != NULL) Debug = strtoul(debug, NULL, 0);
  }

  /* Enable behavior of earlier ELS releases if so requested: */
  MaxVersionLevel = VersionLevel;
  {
    char *els_ver_level = getenv("ELS_VER_LEVEL");
    if (els_ver_level != NULL)
    {
      char version[16];
      sprintf(version, "+v%.8s", els_ver_level);
      do_options(version);
      ver_level_given = FALSE; /* Env var considered different than arg */
    }
    if (Argc >= 2 && strncmp(Argv[1], "+v", 2) == 0)
      do_options(Argv[1]);
    if (VersionLevel > MaxVersionLevel)
      fprintf(stderr, "\
%s: Warning: Requested level %s exceeds this version's supported level %s\n\
	Consider upgrading to a newer version\n",
	      Progname, verToStr(VersionLevel), verToStr(MaxVersionLevel));
  }

  /* Parse options and file args in their given order if env var set: */
  ArgOrder = getenv("ELS_ARG_ORDER") != NULL || VersionLevel <= 142;

  /* Truncate annoyingly long USER/GROUP IDs if env var set: */
  TruncateId = getenv("ELS_TRUNCATE_ID") != NULL;

#ifndef BUFFER_STDOUT
  /* Define BUFFER_STDOUT if you desire buffering on stdout and stderr
     (buffering stdout and stderr will improve performance, but it will
     also cause errors to be randomly placed within the output): */
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);
#endif /*!BUFFER_STDOUT*/
  
#if defined(SUNOS)
  /* Determine at runtime which version of SunOS is being run (this is
     necessary to decide whether the OS release supports ACLs): */
  osVersion = get_osVersion();
#endif

  /* Set the default OPTION values: */

  if (strcmp(Progname, "els5") == 0)
    Sem = SEM_ELS|SEM_SYS5;
  else if (strcmp(Progname, "els4") == 0)
    Sem = SEM_ELS|SEM_BSD;
  else if (strcmp(Progname, "ls5") == 0)
    Sem = SEM_LS|SEM_SYS5;
  else if (strcmp(Progname, "ls4") == 0)
    Sem = SEM_LS|SEM_BSD;
  else if (strcmp(Progname, "ls") == 0)
#if defined(GEN_BSD) || defined(FREEBSD) || SUNOS==4
    Sem = SEM_LS|SEM_BSD;
#elif defined(LINUX)
    Sem = SEM_LS|SEM_GNU;
#else
    Sem = SEM_LS|SEM_SYS5;
#endif
  else /* default is ELS with neither SYS5 nor BSD specific flavorings: */
    Sem = SEM_ELS;

  list_hidden = FALSE;
  list_dotdir = FALSE;
  list_long = FALSE;
  list_long_numeric = FALSE;
  list_long_omit_gid = FALSE;
  list_atime = FALSE;
  list_ctime = FALSE;
  list_inode = FALSE;
  list_size_in_blocks = FALSE;
  mark_files = FALSE;
  mark_dirs = FALSE;
  dash_b = FALSE;
  dash_Q = FALSE;
  recursive = FALSE;
  reverse_sort = FALSE;
  time_sort = FALSE;
  unsort = FALSE;
  g_flag = FALSE;
  G_flag = FALSE;
  expand_symlink = FALSE;
  expand_directories = TRUE;
  list_directories = TRUE;
  plus_q = FALSE;
  traverse_mp = FALSE;
  if (VersionLevel <= 145) traverse_mp = TRUE;
  traverse_expanded_symlink = FALSE;
  traverse_specified = FALSE;
  warning_suppress = FALSE;
  watch_progress = FALSE;
  zero_st_info = FALSE;

  G_format = "n";
  N_format = "f";
  T_format = "M%_DYt";
  
  fexpr_count = 0;
  exc_file_count = 0;
  exc_dir_count = 0;
  inc_file_count = 0;
  inc_dir_count = 0;
  
  /* How many files? */
  {
    int i, n_files = 0;
    for (i = 1; i < Argc; i++)
    {
      if (IS_NOT_MEMBER(Argv[i][0], "+-"))
	n_files++;
    }
    zero_file_args = (n_files == 0);
    multiple_file_args = (n_files > 1);
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  /* This section processes command and file arguments in the order they
     arrive--this was the standard mode for ELS version 1.42 and below
     (enabled by setting the environment variable ELS_ARG_ORDER): */
  if (ArgOrder)
  {
    Iarg = 1;
    do
    {
      /* Process options in the order they arive: */
      while (Iarg < Argc && IS_MEMBER(Argv[Iarg][0], "+-"))
      {
	do_options(Argv[Iarg]);
	Iarg++;
      }
      analyze_options();
      
      /* Process files and directories in the order they arrive: */
      dlist.head = NULL;
      while (Iarg < Argc && IS_NOT_MEMBER(Argv[Iarg][0], "+-"))
      {
	if (Argv[Iarg][0] != CNULL)
	  append_dir(&dlist, Argv[Iarg], 0);
	Iarg++;
      }
      
      /* If there were no valid files and this is the first listing, then by
	 default list all files.  However, prevent "els +G... f1 f2 +X" from
	 listing "f1 f2" and then applying +X to every file the second time
	 around.  On the other hand, "els +G... f1 f2 +X f3 f4" is
	 considered valid (although f1 f2 will be listed while +X will apply
	 only to f3 f4). */
      if (dlist.head == NULL)
      {
	if (First_listing)
	  append_dir(&dlist, ".", 0);
	else
	{
	  Current_Arg = Argv[Argc-1];
	  opt_error_msg("Option must precede files to have any effect",
			Current_Arg+1);
	}
      }
      
      /* Perform the listing: */
      dname = "";
      stat_dir(&dlist, dname);
      sort_dir(&dlist, dname);
      list_dir(&dlist, dname);
      free_dir(&dlist);
      First_listing = FALSE;
    } while (Iarg < Argc);
  }  

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  /* This section processes command arguments during the first pass and then
     lists any file arguments during the second pass--this is the standard
     mode for ELS version 1.43 and above: */
  else
  {
    /* Process options in the first pass: */
    for (Iarg = 1; Iarg < Argc; Iarg++)
    {
      if (IS_MEMBER(Argv[Iarg][0], "+-"))
	do_options(Argv[Iarg]);
    }
    analyze_options();
      
    /* Process files and directories in the second pass: */
    dlist.head = NULL;
    for (Iarg = 1; Iarg < Argc; Iarg++)
    {
      if (IS_NOT_MEMBER(Argv[Iarg][0], "+-") && Argv[Iarg][0] != CNULL)
	  append_dir(&dlist, Argv[Iarg], 0);
    }
    if (dlist.head == NULL)
      append_dir(&dlist, ".", 0);
    
    /* Perform the listing: */
    dname = "";
    stat_dir(&dlist, dname);
    sort_dir(&dlist, dname);
    list_dir(&dlist, dname);
    free_dir(&dlist);
  }
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

  /* Return status back to SHELL: */
  if (ANYBIT(Sem,SEM_BSD) || VersionLevel <= 147)
    /* BSD returns zero status regardless of listing errors;
       ELS prior to 1.47 behaved similar to BSD: */
    return(NORMAL_EXIT);
  else
    /* Return non-zero status if any error occured during listing: */
    return(listingError ? LISTING_ERROR : NORMAL_EXIT);
}


void do_options(char *options)
{
  static int level = 0;
  if (level++ == 0) Current_Arg = options;

  if (*options == '-')
    do_options_minus(&options[1]);
  else if (*options == '+')
    do_options_plus(&options[1]);

  if (--level == 0) Current_Arg = NULL;
}


#define OPTION(val)  (negate ? !(val) : (val))
void do_options_minus(char *options)
{
  char opt;
  while ((opt = *options++) != CNULL)
  {
    Boole negate = (opt == fexpr_FNOT || opt == fexpr_NOT);
    if (negate) opt = *options++;
    switch (opt)
    {
    case 'a': /* List all files including "." and ".." */
      list_hidden = OPTION(TRUE);
      list_dotdir = OPTION(TRUE);
      break;

    case 'A': /* List all files except "." and ".." */
      list_hidden = OPTION(TRUE);
      list_dotdir = OPTION(FALSE);
      /* "els -~A" will ignore all hidden files except "." and ".."! */
      break;
      
    case 'd': /* Inhibit expansion of directories specified as params */
      expand_directories = OPTION(FALSE);
      break;
      
    case 'l': /* Long listing as in ls -l */
      list_long = OPTION(TRUE);
      break;

    case 'n': /* Long listing with UID and GID in numeric */
      list_long = OPTION(TRUE);
      list_long_numeric = OPTION(TRUE);
      break;

    case 'o': /* Long listing omitting GID */
      list_long = OPTION(TRUE);
      list_long_omit_gid = OPTION(TRUE);
      break;

    case 'g': /* Interpret -g flag using either BSD or SYS5 semantics */
      g_flag = OPTION(TRUE);
      break;
      
    case 'G': /* Interpret -G flag using GNU semantics */
      G_flag = OPTION(TRUE);
      break;
      
    case 'L': /* Expand symbolic link */
      expand_symlink = OPTION(TRUE);
      break;
      
    case 'u': /* List atime */
      list_atime = OPTION(TRUE);
      /* Both BSD and SYS5/GNU agree that -u always supersedes -c: */
      if (list_atime) list_ctime = FALSE;
      break;

    case 'c': /* List ctime */
      list_ctime = OPTION(TRUE);
      if (list_ctime)
      {
	/* In SYS5/GNU the last -c or -u specified supersedes the previous,
	   while in BSD/ELS if -u is specified then -c is ignored: */
        if (!ANYBIT(Sem,SEM_BSD|SEM_ELS)) list_atime = FALSE;
        if (ANYBIT(Sem,SEM_BSD|SEM_ELS) && list_atime) list_ctime = FALSE;
      }
      break;

    case 'i': /* List inode */
      list_inode = OPTION(TRUE);
      break;

    case 's': /* List size in blocks */
      list_size_in_blocks = OPTION(TRUE);
      break;

    case 'r': /* Reverse sort */
      reverse_sort = OPTION(TRUE);
      break;
      
    case 't': /* Sort by time */
      time_sort = OPTION(TRUE);
      break;
      
    case 'U': /* Unsort */
      unsort = OPTION(TRUE);
      break;
      
    case 'R': /* Recursively list all directories found (same as +R) */
      recursive = OPTION(TRUE);
      break;

    case 'F': /* Mark files */
      mark_files = OPTION(TRUE);
      break;
      
    case 'p': /* Mark directories */
      mark_dirs = OPTION(TRUE);
      break;
      
    case 'b': /* List non-printable characters using \nnn octal */
      dash_b = OPTION(TRUE);
      break;
      
    case 'Q': /* List names using C-syntax a la GNU */
      dash_Q = OPTION(TRUE);
      break;
      
    case '1': /* List single column */
      /* Currently the default */
      break;
      
    case 'C': /* List multi column */
      /* Unimplemented */
      break;
      
    default: /* Give error message and usage, then exit: */
      opt_error_msg("Unrecognized option", options);
      break;
    }
  }
  
  return;
}


void do_options_plus(char *options)
{
  char opt;
  while ((opt = *options++) != CNULL)
  {
    Boole negate = (opt == fexpr_FNOT || opt == fexpr_NOT);
    if (negate) opt = *options++;
    switch (opt)
    {
      /* Exactly one of either SEM_ELS or SEM_LS must always be set;
	 thus, clearing one implies setting the other: */
    case 'l':
      CLRBITS(Sem, SEM_ELS|SEM_LS);
      if (negate)
	SETBIT(Sem, SEM_ELS);
      else
	SETBIT(Sem, SEM_LS);
      break;

      /* Either one or none of SEM_ELS, SEM_SYS5, or SEM_GNU can be set;
	 thus, clearing one does not imply setting any of the others: */
    case 'b': /* '+b' is DEPRECATED, use +4 instead */
    case '4':
      if (negate)
	CLRBIT(Sem, SEM_BSD);
      else
      {
	CLRBITS(Sem, SEM_BSD|SEM_SYS5|SEM_GNU);
	SETBIT(Sem, SEM_BSD);
      }
      break;

    case '5':
      if (negate)
	CLRBIT(Sem, SEM_SYS5);
      else
      {
	CLRBITS(Sem, SEM_BSD|SEM_SYS5|SEM_GNU);
	SETBIT(Sem, SEM_SYS5);
      }
      break;

    case 'g':
      /* undocumented option */
      if (negate)
	CLRBIT(Sem, SEM_GNU);
      else
      {
	CLRBITS(Sem, SEM_BSD|SEM_SYS5|SEM_GNU);
	SETBIT(Sem, SEM_GNU);
      }
      break;

    case 'D': /* '+D' is DEPRECATED, use +d instead */
    case 'd': /* List directories as files */
      /* NB: This does not affect the expansion of directories, rather, the
	 purpose is to generate directory listings without directory files. */
      list_directories = OPTION(FALSE);
      break;
      
    case 'h': /* Allow +X to act on symbolic links */
      execute_symlinks = OPTION(TRUE);
      break;

    case 'Q': /* '+Q' is DEPRECATED, use +q instead */
    case 'q': /* Quote file names having special characters */
      plus_q = OPTION(TRUE);
      break;

    case 'R': /* Recursively list all directories found (same as -R) */
      recursive = OPTION(TRUE);
      break;

    case 't': /* Traverse mount-points during recursive listings */
      /* Look for and read undocumented "=type": */
      traverse_specified = TRUE;
      if (*options == '=')
      {
	options++;
	if (IS_NOT_MEMBER(*options, "~ML"))
	    opt_error_msg("Expected one of '~ML'", options+1);

	while (IS_MEMBER(*options, "~ML"))
	{
	  Boole save_negate = negate;
	  char type = *options++;
	  
	  if (type == '~')
	  {
	    negate = !negate;
	    type = *options++;
	  }
	  
	  if (type == 'M')
	    traverse_mp = OPTION(TRUE);
	  else if (type == 'L')
	    /* If a directory is symbolically linked to a parent directory,
	       this will cause els (and /bin/ls) to loop infinitely.
	       Possible infinite symlink loops can be avoided by disabling
	       symlink traversal by specifying "+~t=L" or "+t=~L". */
	    traverse_expanded_symlink = OPTION(TRUE);
	  else
	    opt_error_msg("Expected one of 'ML'", options);
	  
	  negate = save_negate;
	}
      }
      else
      {
	traverse_mp = OPTION(TRUE);
	if (VersionLevel <= 145) traverse_mp = OPTION(FALSE);
	
	/* Changing the meaning of +t was an agonizing decision but was
	   necessary to allow +t=<type> specifying the types of directories
	   and/or filesystems to traverse: */
	if (VersionLevel >= 147)
	{
	  static Boole warned = FALSE;
	  if (! warned)
	  {
	    fprintf(stderr, "\
\n\
%s: Warning: The meaning of +t has changed as of release 1.47 (May 2000)\n\
Specifying +t under release 1.47 causes file-systems to be traversed whereas\n\
before it inhibited traversal.  Moreover, as of release 1.48 (March 2002)\n\
you should specify '+t=M' for traversing mount-points and +t=L for\n\
traversing expanded symbolic-links.\n\
\n\
", Progname);
	    warned = TRUE;
	  }
	}
      }
      break;
      
    case 'w': /* Warning suppression */
      warning_suppress = OPTION(TRUE);
      /* TBD: 'w' will eventually take a mask argument and maybe have
	 its logic reversed? */
      break;

    case 'W': /* Watch progress */
      watch_progress = OPTION(TRUE);
      /* TBD: 'W' might eventually take an argument to specify report
	 interval or report type (e.g. number, spinning wheel). */
      if (watch_progress)
      {
	if (defaultSigHandler == NULL)
	  defaultSigHandler = signal(WATCH_SIGNAL, watchSigHandler);
	else
	  (void) signal(WATCH_SIGNAL, watchSigHandler);
	if (ttyin  == NULL) ttyin  = fopen("/dev/tty", "r");
	if (ttyout == NULL) ttyout = fopen("/dev/tty", "w");
	if (ttyin == NULL || ttyout == NULL)
	{
	  fprintf(stderr, "%s: Warning: Unable to open /dev/tty\n" , Progname);
	  if (ttyin == NULL) ttyin = stdin;
	  if (ttyout == NULL) ttyout = stderr;
	}
      }
      else
      {
	if (defaultSigHandler != NULL)
	  (void) signal(WATCH_SIGNAL, defaultSigHandler);
      }
      break;

    case 'z': /* Zero-out meaningless stat info */
      zero_st_info = OPTION(TRUE);
      /* Look for and read optional/undocumented mask: */
      {
	char *tmp = options;
	zero_st_mask = strtol(options, &options, 8);	/* use octal */
	if (tmp == options)
	  zero_st_mask = 037777777777;
      }
      break;

    case 'M': /* MUNGE dates and file names */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      munge++;	/* Go to next munge-level */
      do_options("+d");

      {
	char *g_opt;
	if (munge == 1)
	{
	  g_opt = "+GmnL";
	}
	else if (munge == 2 || munge == 3)
	{
	  g_opt = "+Gan";
	}
	else if (munge == 4)
	{
	  g_opt = "+G'chdate -m %m -a %a%n'";
	}
	else
	{
	  g_opt = "";
	}

	/* Defer to any previous G option from the command-line: */
	if (!G_option) { do_options(g_opt); G_option = FALSE; }

	/* Defer to any previous N option from the command-line: */
	if (!N_option) { do_options("+NF");  N_option = FALSE; }

	/* Do not defer to any previous T option, as all munge levels imply
	   that we want ISO8601 dates: */
	do_options("+TI");
      }

      /* Additional options may follow any MUNGE command */
      break;
      
    case 'U': /* Generate Untouch script commands */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      untouch++;
      untouch_setup(1);
      break;

    case 'f': /* Specify field separtor */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      fielding = TRUE;
      separator = *options++;
      break;

    case 'G': /* Specify the general format */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      G_option = TRUE;	/* Command-line args take precedence */
      G_format = options;
      options = "";	/* No more options */
      break;
      
    case 'N': /* Specify the name format */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      N_option = TRUE;	/* Command-line args take precedence */
      N_format = options;
      options = "";	/* No more options */
      break;
      
    case 'T': /* Specify the time & date format */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      T_option = TRUE;	/* Command-line args take precedence */
      T_format = options;
      options = "";	/* No more options */
      break;
      
    case 'E': /* Exclude file names filter */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      filtering = TRUE;
      if (exc_file_count < MAX_EXC_FILTER)
	exc_file_filter[exc_file_count++] = options;
      else
	opt_error_msg(TOO_MANY_FILTERS, options);
      options = "";	/* No more options */
      break;

    case 'e': /* Exclude directory names filter */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      filtering = TRUE;
      if (exc_dir_count < MAX_EXC_FILTER)
	exc_dir_filter[exc_dir_count++] = options;
      else
	opt_error_msg(TOO_MANY_FILTERS, options);
      options = "";	/* No more options */
      break;

    case 'I': /* Include file names filter */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      filtering = TRUE;
      if (inc_file_count < MAX_INC_FILTER)
	inc_file_filter[inc_file_count++] = options;
      else
	opt_error_msg(TOO_MANY_FILTERS, options);
      options = "";	/* No more options */
      break;
      
    case 'i': /* Include directory names filter */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      filtering = TRUE;
      if (inc_dir_count < MAX_INC_FILTER)
	inc_dir_filter[inc_dir_count++] = options;
      else
	opt_error_msg(TOO_MANY_FILTERS, options);
      options = "";	/* No more options */
      break;
      
    case 'F': /* Filter expression */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      filtering = TRUE;
      if (fexpr_count < MAX_FEXPR_FILTER)
	fexpr_filter[fexpr_count++] = strdup_ns(options);
      else
	opt_error_msg(TOO_MANY_FILTERS, options);
      options = "";	/* No more options */
      break;

    case 'S': /* Select a Stamp format */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      stamping = TRUE;
      stamp_option = *options;
      stamp_setup(1);
      options++;
      break;
      
    case 'v': /* Give version and bits of information: */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      if (*options == '=' || isdigit(*options))
      {
	if (Iarg != 1 || options != Current_Arg+2)
	  opt_error_msg("If specified, +v must be first argument", options);
	if (*options == '=') options++;
	VersionLevel = strToVer(options, &options);
	if (*options != CNULL)
	  opt_error_msg("Invalid version", options+1);
	ver_level_given = TRUE;
      }
      else
      {
	puts(VersionID);
	printf("	ELS version: %s\n", verToStr(MaxVersionLevel));
	printf("	Max dir name: %d\n", MAX_DNAME);
	printf("	Max file name: %d\n", MAX_FNAME);
	show_options();
	printf("\n");
	exit(USAGE_ERROR);
      }
      break;

    case 'V': /* Verbose mode */
      verbose_mode = OPTION(TRUE);
      break;

    case 'X': /* Execute mode */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      execute_mode = TRUE;
      /* +X must always use names quoted using +q whenever unusual file names
	 occur (-b and -Q are inadequate for these purposes!): */
      do_options("+q");
      do_options("-~b~Q");
      /* Detect the following, otherwise unpleasant surprises may occur: */
      if (!First_listing)
	error_msg("+X must precede any file name arguments");
      /* Look for and read optional status check value: */
      {
	char *tmp = options;
	/* ??? Maybe issue warning if > 255 ??? */
	execute_sts_good = (strtol(options, &options, 0) & 0xff);
	execute_sts_check = (tmp != options);
      }
      break;

    case 'Z': /* Set zonename */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      if (*options == '=')
      {
	options++;
	Time_Zone = get_validTZ(options);
	set_currentTZ(Time_Zone);
	options = "";	/* No more options */
      }
      else
      {
	opt_error_msg("Missing =ZONENAME", options);
      }
      break;

    case 'H': /* Give help and exit: */
      if (negate) opt_error_msg(NON_NEGATABLE, options);
      give_usage_or_help('H');
      exit(USAGE_ERROR);
      break;

    default: /* Give error message and usage, then exit: */
      opt_error_msg("Unrecognized option", options);
      break;
    }
  }
  
  return;
}


void stamp_setup(int pass)
{
  if (pass == 1)
  {
    if (stamp_option == 'o')
    {
      do_options("+GCtPUG~%+8s~mNL"); /* ~%+8s~ lists device files as X,Y */
      do_options("+TI");
      do_options("-A");
    }
    else if (stamp_option == 'O')
    {
      do_options("+GtPUG~%+8s~mNL");  /* ~%+8s~ lists device files as X,Y */
      do_options("+TI");
      do_options("-A");
    }
    else if (stamp_option == 'f')
    {
      do_options("+GCtP~%+8s~mNL"); /* ~%+8s~ lists device files as X,Y */
      do_options("+TI");
      do_options("-A");
    }
    else if (stamp_option == 'F')
    {
      do_options("+GtP~%+8s~mNL");  /* ~%+8s~ lists device files as X,Y */
      do_options("+TI");
      do_options("-A");
    }
    else if (stamp_option == 'C' || stamp_option == 'c')
    {
      /* Tracking directory dates is pointless for sorce code: */
      do_options("+d");
      
      do_options("+Gm  u~%+-6s~NL");	/* ~%+-6s~ lists device files as X,Y */
      do_options("+TI");
      do_options("-A");
      
      if (stamp_option == 'C')
      {
	/* Exclude libraries, object, lint files: */
	do_options("+E*.[ao]");
	do_options("+E*.ln");
	
	/* Exclude SCCS, RCS, CVS directories: */
	do_options("+eSCCS");
	do_options("+eRCS");
	do_options("+eCVS");
	/* ??? Perhaps exclude CVSROOT/cvsroot??? */
	
	/* Exclude tilde backup files generated by Emacs: */
	do_options("+E*~");
	
#ifdef MARKB_PREFERENCES
	/* I like to use an older (and faster) version of GNU emacs that
	   generates .~* backup files instead of *~ */
	do_options("+E.~*");
	
	/* I use numbers on the end of my files to designate versions; e.g.
	   els.c.49, els.c.50, els.c.51, etc. thus I choose to ignore them */
	do_options("+E*.[0-9][0-9]");
	do_options("+E*.[0-9][0-9][0-9]");
#endif
      }
    }
    else
    {
      do_options("+H");	/* Give help and exit */
    }
  }
  else
  {
    char vdate[42];
    char *cwd, dir[MAX_DNAME];
    char *tz;
    T_print(vdate, "v", The_Time, NULL, FALSE, FALSE);
    cwd = dir;
    getcwd(cwd, MAX_DNAME);
    if (strncmp("/tmp_mnt/", cwd, 9) == 0) cwd += 8; /* automount fodder */
    printf("	Stamp performed by %s\n", get_userAtHost());
    printf("	on %s\n", vdate);
    printf("	from the directory %s\n", cwd);
    /* Note: root's TZ setting under SunOS5 is "" */
    if (Time_Zone == NULL && (tz = getenv("TZ")) != NULL && *tz != CNULL)
      printf("	with the environment variable TZ=%s\n", tz);
    printf("	using the command:");
    print_args(Argv, Argc, ver_level_given ? 0 : VersionLevel);
    printf("\n \n"); /* The embedded space is deliberate */
    
    if (stamp_option == 'o')
    {
      /* +GCtPUG~%+8s~mNL */
      printf("\
CKSum T Mode   UID   GID    Size       Date          File\n\
----- - ----   ---   ---   ------ --------------- ----------\n\
");
    }
    else if (stamp_option == 'O')
    {
      /* +GtPUG~%+8s~mNL */
      printf("\
T Mode   UID   GID    Size       Date          File\n\
- ----   ---   ---   ------ --------------- ----------\n\
");
    }
    else if (stamp_option == 'f')
    {
      /* +GCtP~%+8s~mNL */
      printf("\
CKSum T Mode    Size       Date          File\n\
----- - ----   ------ --------------- ----------\n\
");
    }
    else if (stamp_option == 'F')
    {
      /* +GtP~%+8s~mNL */
      printf("\
T Mode    Size       Date          File\n\
- ----   ------ --------------- ----------\n\
");
    }
    else if (stamp_option == 'C' || stamp_option == 'c')
    {
      /* +Gm  u~%+-6s~NL */
      printf("\
     Date        Owner    Size      File\n\
---------------  -----    -----  ----------\n\
");
    }
  }
  return;
}

      
void untouch_setup(int pass)
{
  if (pass == 1)
  {
    if (untouch == 1)
      do_options("\
+G'[ -f %N ] && [ `els +GC %N` -eq %C ] && chdate -m %m %N'");
    else
      do_options("\
+G'[ -f %N ] && [ `els +GC %N` -eq %C ] && chdate -m %m -a %a %N'");
    do_options("+q");		/* Use quoted names */
    do_options("+TI");		/* Use Iso8601 dates */
    do_options("+FT{r}");	/* List regular files only */
  }
  else
  {
    char vdate[42];
    char *cwd, dir[MAX_DNAME];
    char *tz;
    T_print(vdate, "v", The_Time, NULL, FALSE, FALSE);
    cwd = dir;
    getcwd(cwd, MAX_DNAME);
    if (strncmp("/tmp_mnt/", cwd, 9) == 0) cwd += 8; /* automount fodder */
    printf("\
#!/bin/sh\n\
#\n\
# 'untouch' shell script created by %s\n\
# on %s\n\
# from the directory %s\n\
# using the command:", get_userAtHost(), vdate, cwd);
    print_args(Argv, Argc, ver_level_given ? 0 : VersionLevel);
    printf("\n\
#\n\
# NB: Just because you are able to 'touch' a file doesn't necessarily mean\n\
# that you can 'untouch' it.  The ability to untouch a file may require\n\
# that you be the direct owner of the file or have super-user privileges.\n\
#\n\
# Save the following output into a script file so that it can then be run\n\
# at a later time in order to recover file dates following a 'touch':\n\
#\n\
");
    /* Include TZ setting regardless of whether it's from the environment
       or from the command line, as the script file needs to be told: */
    if ((tz = getenv("TZ")) != NULL && *tz != CNULL)
      printf("\nTZ=%s; export TZ\n\n", tz);
  }
  return;
}


void give_usage_or_help(char U_or_H)
{
  /* Print usage to stderr so as to avoid any redirection.
     Print help to stdout so as to make it pipe-able through more: */
  FILE *out, *fopen(), *popen();
  Boole more = FALSE;

  if (U_or_H == 'H')
  {
    char *places[] = {"/usr/ucb/more", "/usr/bin/more", "/bin/more", NULL};
    char *pager, **path;
    pager = getenv("PAGER");
    if (pager == NULL)
    {
      for (path = places; *path && fopen(*path, "r") == NULL; path++) /*...*/;
      pager = *path;
    }
    if (pager != NULL && (out = popen(pager, "w")) != NULL)
      more = TRUE;
    else
      out = stdout;
  }
  else
    out = stderr;

  Void fprintf(out, "\
\n\
Usage: %s [-~aAbcCdFgGilLnopQrRstuU1] [+~dhHlqRtvVwz45] \\\n\
	[+M[M]] [+Z=zonename] [+f char] [+GTN format] \\\n\
	[+IEie pattern] [+F fexpr] \\\n\
	[+S char] [+U] [+W] [+X status] files\n\
\n\
", Progname);

  if (U_or_H == 'U')
  {
    Void fprintf(out, "\
For more help: %s +H\n\
\n\
", Progname);
  }
  else
  {
    Void fprintf(out, "\
\n\
LS -- STANDARD OPTIONS:\n\
\n\
    -a: List all files including . and ..\n\
    -A: List all files except . and ..\n\
    -b: List non-graphic characters using octal notation\n\
    -c: List time of last status change(SYS5, ELS),\n\
        List time of last status change and sort(BSD)\n\
    -C: List multi-columns (unimplemented)\n\
    -d: List directories as files, but don't list their contents\n\
    -F: Mark files\n\
    -g: List GIDs(BSD), long listing omitting UIDs(SYS5), ignored(ELS)\n\
    -G: Don't list GIDs(ELS)\n\
    -i: List inode number\n\
    -l: Long listing using BSD, SYS5, or ELS semantics\n\
    -L: List actual file rather than symbolic link\n\
    -n: Long listing using numeric UIDs and GIDs(SYS5, ELS)\n\
    -o: Long listing omitting GIDs(SYS5, ELS)\n\
    -p: Mark directories\n\
    -Q: List non-graphic characters using double-quotes and C-style notation\n\
    -r: Reverse the sort\n\
    -R: Recursively list contents of all directories\n\
    -s: List size in KBytes(BSD, ELS) or in blocks(SYS5)\n\
    -t: Sort files according to time\n\
    -u: List time of last access\n\
    -U: List unsorted(ELS)\n\
    -1: List single column\n\
     ~: If a tilde precedes any of the above options then that option is\n\
	reset, effectively causing it to be canceled.\n\
\n\
");

    Void fprintf(out, "\
\n\
Enhanced LS -- MISCELLANEOUS:\n\
\n\
    +H: Give HELP\n\
    +l: Mimic /bin/ls behavior\n\
    +4: Use BSD semantics\n\
    +5: Use SYS5 semantics\n\
    +d: Don't list directories as files, but list their contents (has the\n\
	opposite effect of -d)\n\
    +h: Allow +X to act upon symbolically linked files\n\
    +q: Quote unusual and troublesome file names (besides doing a more\n\
	thorough job than either -b or -Q, this option also quotes special\n\
	and non-graphic characters so that the file name can subsequently be\n\
	used as an argument for most Unix commands; thus, the +q option is\n\
	automatically implied whenever using the +X option).\n\
    +R: Recursively list contents of all directories (same as -R)\n\
    +t: Traverse mount-points during recursive listings\n\
    +v: Print els version and information\n\
    +V: Verbose mode\n\
    +w: Suppress warnings about unreadable files, etc.\n\
    +z: List volatile file information as zero (volatile file information is\n\
	defined as those values that change whenever a file is listed or\n\
	those values that aren't preserved after a hierarchy gets copied\n\
	or restored; for example, a directory or a symbolic link changes its\n\
	access time whenever being listed; a symbolic link changes its\n\
	modification time and/or ownership after being copied; a directory\n\
	doesn't always preserve its size after being copied).  Also, since\n\
	volatile file information can change over time, this option is\n\
	particularly useful when stamping a hierarchy (see +S option), or\n\
	when comparing the listings of two very similar hierarchies.\n\
     ~: If a tilde precedes any of the above options then that option is\n\
	reset, effectively causing it to be canceled.\n\
\n\
    +M: Macro 1 for listing ISO8601 *modification* dates of all files\n\
	except directories (i.e. els +GmnL +NF +TI +d).  For example, to\n\
	list the 30 most recently modified source code files in a hierarchy:\n\
		els +MR +I'*.{c,h,cc,hh}' | sort | tail -30\n\
   +MM: Macro 2 for listing ISO8601 *access* dates of all files except\n\
	directories and symbolic links (i.e. els +Gan +NF +TI +d +F'T{~l}').\n\
	For example, to list the 100 most recently accessed executable files\n\
	under the /usr hierarchy:\n\
		els +MMR +F'P{+x}' /usr | sort | tail -100\n\
\n\
    +Z: Specify timezone to be used in place of current TZ setting.\n\
	The timezone should be in the form of:\n\
	    All OSes:		STDoffset[DST[offset][,rule]]\n\
	Additionally, if your host provides zoneinfo then you can also\n\
	use names from the appropriate zoneinfo directory:\n\
	    SunOS/Solaris:	/usr/share/lib/zoneinfo\n\
	    Linux2, FreeBSD:	/usr/share/zoneinfo\n\
	    OSF1:		/etc/zoneinfo\n\
	For example:\n\
	    els -la +Z=EST5EDT +TIZ	(Available on *all* OSes)\n\
	    els -la +Z=US/Eastern +Tv	(Available on OSes with zoneinfo)\n\
\n\
");

    Void fprintf(out, "\
\n\
Enhanced LS -- FORMATTING:\n\
\n\
    +f: Field separator character, e.g.:  els -la +f: +TI\n\
\n\
    +G: GENERAL format, e.g.:  els +G~tp~lusmnL (same as els -l)\n\
	i:  inode number\n\
      t,T:  type of file in alpha(t), symbolic(T)\n\
      p,P:  permission in alpha(p), numeric(P)\n\
	M:  permission in chmod format (e.g. u=rwxs,g=x,o=x)\n\
	A:  ACL indicator '+' if present\n\
	l:  link count\n\
      u,U:  UID in alpha(u), numeric(U)\n\
      g,G:  GID in alpha(g), numeric(G)\n\
      s,S:  size in bytes(s), blocks(S)\n\
    m,a,c:  time modified(m), accessed(a), status changed(c)\n\
	    (time displayed using +T format)\n\
      n,N:  file name(n), full file name(N)\n\
	    (name displayed using +N format)\n\
    d,f,F:  directory name(d), file name(f), full file name(F)\n\
	L:  Symbolic link target prefaced by '->'\n\
	C:  16 bit checksum (performed on regular files only and\n\
	    access times are preserved)\n\
\n\
    +T: TIME format, e.g.:  els -la +T^rD (give age of file in days)\n\
       ^a:  Modifier for absolute time since the epoch\n\
       ^d:  Modifier for delta time from now (i.e. difference)\n\
       ^r:  Modifier for relative time from now (i.e. age)\n\
       ^A:  Modifier for alpha dates instead of numeric\n\
       ^N:  Modifier for numeric dates instead of alpha\n\
       ^G:  Modifier for GMT dates instead of local\n\
       ^L:  Modifier for local dates instead of GMT\n\
       ^M:  Modifier for meridian instead of military time\n\
	I:  Iso8601 style (same as +T^N~YMD.hms~)\n\
	e:  els style (default, same as +TM%%_DYt)\n\
	l:  ls style (same as +TM%%_DQ)\n\
	d:  dos style (same as +T^N%%_M-D-y^M%%_h:~mp~)\n\
	w:  windows style (same as +T\"^N%%_M/D/y^M%%_tP'M'\")\n\
	v:  verbose style (same as +TWM%%_DTZY)\n\
  Y,M,D,W:  year(Y), month(M), day(D), weekday(W)\n\
    h,m,s:  hour(h), minutes(m), seconds(s)\n\
      t,T:  time as h:m(t), h:m:s(T)\n\
	Q:  time or year depending on age\n\
	p:  'a' or 'p' depending on meridian\n\
	    (meaningful only with ^M modifier)\n\
	P:  'A' or 'P' depending on meridian\n\
	    (meaningful only with ^M modifier)\n\
	y:  year modulo 100\n\
	Z:  zone name\n\
\n\
    +N: NAME format, e.g.:  els -laR +NF\n\
       ^q:  Modifier for quoting unusual file names (same as +q)\n\
	F:  Full file name (same as +Nd/f)\n\
	d:  Directory name\n\
	f:  File name\n\
	l:  Symbolic link target\n\
	L:  Symbolic link target prefaced by '->'\n\
\n\
    +G, +T, +N format controls\n\
	\\:  Output following character verbatim\n\
	~:  Toggle spacing off/on\n\
       %%%%:  Output a single %% character\n\
       %%D:  Output directive 'D' using default width and default padding\n\
      %%_D:  Pad left side of output with blanks\n\
      %%0D:  Pad left side of output with zeros\n\
      %%-D:  Suppress all padding and use minimum width\n\
     %%0nD:  Output a zero padded field 'n' characters wide\n\
     %%_nD:  Output a blank padded field 'n' characters wide\n\
      %%nD:  Output a right justified field 'n' characters wide\n\
     %%-nD:  Output a left justified field 'n' characters wide\n\
     %%+nD:  Output a field 'n' characters wide regardless of ~ spacing\n\
\n\
	If a string occurs within an inner set of quotes then the string\n\
	is output verbatim (except for any directives within the inner\n\
	quotes prefaced by a %%).  Thus, the following are equivalent:\n\
\n\
	+T\"'DATE: '^N%%M/D/Y\"    # M/D/Y in outer quotes (%% optional)\n\
	+T'\"DATE: \"^NM/%%D/Y'    # M/D/Y in outer quotes (%% optional)\n\
	+T'\"DATE: ^N%%M/%%D/%%Y\"'  # M/D/Y in INNER quotes (%% REQUIRED)\n\
\n\
");

    Void fprintf(out, "\
\n\
Enhanced LS -- FILTERING:\n\
\n\
    +I: Include specified files, e.g.:  els -laR +I\"*.{c,h,cc,hh}\"\n\
    +E: Exclude specified files, e.g.:  els -laR +E\"*.o\"\n\
    +i: Include specified directories, e.g.:  els -laR +i\"[a-m]*\"\n\
    +e: Exclude specified directories, e.g.:  els -laR +e\"SCCS\"\n\
\n\
    +F: FILTER EXPRESSION\n\
\n\
	A filter expression (fexpr) consists of one or more of the\n\
	following filter types separated by Boolean operators:\n\
		A{...} -- Access Filter\n\
		T{...} -- Type Filter\n\
		P{...} -- Permission Filter\n\
		Q{...} -- Quantity Filter\n\
		U{...} -- Unusual Filter\n\
\n\
	Each filter type consists of one or more terms separated by\n\
	Boolean operators.  The following lists each term appropriate\n\
	for the associated filter type:\n\
	    Access Filter Terms:\n\
		rwx: read, write, execute access\n\
	    Type Filter Terms:\n\
		rdcb: regular file, directory, char device, block device\n\
		plsD: pipe/fifo, symbolic link, socket, Door\n\
		S: Special/device file (same as 'b|c')\n\
	    Permission Filter Terms:\n\
		value: octal value <= 07777\n\
		ugo: user field, group field, other field\n\
		a: all fields (same as 'ugo')\n\
		+: test if indicated field(s) contain any attribute(s)\n\
		-: test if indicated field(s) missing any attribute(s)\n\
		=: test if indicated field(s) match attribute(s) exactly\n\
		rwx: read attribute, write attribute, execute attribute\n\
		s: setuid/setgid attribute (applies to u and/or g fields)\n\
		t: sticky attribute (applies to u field)\n\
		l: mandatory lock attribute (applies to g field)\n\
	    Quantity Filter Terms:\n\
		iAlugs: inode, ACL count, link count, uid, gid, size\n\
		mac: time modified, accessed, status changed\n\
		== != ~= : equals, not equals, not equals\n\
		> >= : greater than, greater than or equals\n\
		< <= : less than, less than or equals\n\
		value: positive integer to be compared against\n\
		YMWDhms: Years, Months, Weeks, Days, hours, minutes, seconds\n\
	    Unusual Filter Terms:\n\
		t: unusual type (i.e. !regular & !directory & !symlink)\n\
		p: unusual permissions (i.e. o+w & !+t & !symlink |\n\
		   setuid | setgid | mandatory_locking)\n\
		P: unusual permissions (i.e. access(o) > access(g) |\n\
		   access(o) > access(u) | access(g) > access(u))\n\
		A: ACL_count > 1\n\
		l: link_count > 1\n\
		u: nobody/noaccess UID (i.e. UID = 60001, 60002, or 65534)\n\
		   and/or unassigned UID (i.e. not listed in /etc/passwd)\n\
		g: nobody/noaccess GID (i.e. GID = 60001, 60002, or 65534)\n\
		   and/or unassigned GID (i.e. not listed in /etc/group)\n\
		mac: modification, access, change time is in the future\n\
		n: name containing unusual or troublesome characters\n\
		N: full pathname containing unusual or troublesome characters\n\
		L: symbolic link pointing to non-existent file\n\
		s: sparse file (i.e. partially filled w/data)\n\
		G: General tests (i.e. perform all the above unusual tests)\n\
		S: Security related tests (i.e. !directory & o+w |\n\
		   directory & o+w & !+t | setuid | setgid | device_file)\n\
\n\
	Boolean operators are as follows:\n\
		! or ~ -- Boolean NOT (evaluated first)\n\
		& or , -- Boolean AND (evaluated second)\n\
		| or : -- Boolean OR (evaluated last)\n\
	(The reason for allowing multiple symbols for Boolean operations is\n\
	that characters such as '!' have special meaning within the shell\n\
	and are awkward to use.  Also, characters such as ',' can improve\n\
	readability when grouping lists of terms, e.g.: 'u+rw,g-rw,o-rw'\n\
	is more readable than 'u+rw&g-rw&o-rw'.)\n\
\n\
	The syntax for each term is based upon its associated filter type.\n\
	In all cases, any filter having multiple terms must use Boolean\n\
	operators to separate each of its terms.  The 'Access' and 'Type'\n\
	Filters are the simplest cases when it comes to syntax, as each of\n\
	their terms consists of a single letter.\n\
\n\
	The syntax for each term of the permission filter is very similar to\n\
	Unix's chmod command with the following exceptions: if the left-hand\n\
	side of a term is blank then it is taken to mean ANY instead of ALL\n\
	fields; the symbol '+' is taken to mean HAS instead of GIVE; the\n\
	symbol '-' is taken to mean MISSING instead of TAKE; and the symbol\n\
	'=' is taken to mean MATCH EXACTLY instead of SET EXACTLY.  Thus,\n\
	the filter '+FP{+rw}' means ANY FIELD HAVING READ-WRITE PERMISSION,\n\
	while the filter '+FP{a=}' means ALL FIELDS HAVING EXACTLY NO\n\
	PERMISSIONS.\n\
\n\
	The syntax for each term of the quantity filter consists of a\n\
	single letter and an integer quantity separated by a comparison\n\
	operator.  Moreover, if the quantity being compared represents the\n\
	file modification, access, or status change time (i.e. one of\n\
	'mac'), then the integer quantity must be followed by a qualifier\n\
	designating Years, Months, Weeks, Days, hours, minutes, or seconds\n\
	(i.e. one of 'YMWDhms').\n\
\n\
	Examples:\n\
	    Recursively list regular files in /usr/bin that are\n\
	    setuid/setgid having a UID/GID of less than 10:\n\
		els -laR +F'T{r}&P{u+s|g+s}&A{x}&Q{u<10|g<10}' /usr/bin\n\
\n\
	    Recursively list all non-directory files that share a common\n\
	    inode (i.e. hard-link) in /etc and /dev:\n\
		els +GilN +F'Q{l>1}&T{~d}' -R /etc /dev | sort\n\
\n\
	    Recursively list all files modified on or after April 1, 1999\n\
	    belonging to either the user 'markb' or the group 'projectx':\n\
		els -laR +F'Q{m>=19990401}&Q{u=markb|g=projectx}'\n\
\n\
	    Display all files where g or o have more permission than u:\n\
		els -laR +F'P{u-w,+w}|P{u-x,+x}|P{u-r,+r}' +NF\n\
\n\
	    Display all files having 'unusual' types or permissions\n\
	    (e.g., world-writable or sticky files excluding symlinks,\n\
	    setuid/setgid/locking files excluding symlinks and dirs,\n\
	    device/pipe/socket/door files):\n\
		els -laR +NF +F'P{o+w:+t},T{~l}|P{+s,+x:+l},~T{d:l}|~T{r:d:l}'\n\
\n\
	    List all files greater than 100K bytes that have not been\n\
	    accessed within the last 180 days and were modified over\n\
	    365 days ago:\n\
		els +GsamN +T^rD +d -AR +F'Q{s>=100000 & a>180D & m>365D}'\n\
\n\
	    List all files having troublesome characters in their name:\n\
		els -laR -Q +NF +F'U{N}'\n\
\n\
");

    Void fprintf(out, "\
\n\
Enhanced LS -- UTILITIES:\n\
\n\
    +S: STAMP: The output from this option is used for recording the state\n\
	of a hierarchy for future comparison to determine any changes.\n\
\n\
	C:  Stamp a source code hierarchy excluding *.[ao], *.ln, *~,\n\
	    and SCCS/RCS/CVS files, e.g.:  els +SC -R /home/myfiles/src\n\
	c:  Same as +SC, except nothing is excluded\n\
\n\
	O:  Stamp an operating system hierarchy (no files are excluded and\n\
	    more detail given), e.g.:  els +SO +z -R  /  /usr /var /opt\n\
	    (Note that +z is recommended to zero volatile data)\n\
	o:  Same as +SO, except checksumming is also performed\n\
\n\
	F:  Stamp a file hierarchy (same as +SO except that UIDs and GIDs\n\
	    are excluded), e.g.:  els +SF +z -R /etc /dev /usr /var\n\
	    (note that +z is recommended to zero volatile data)\n\
	f:  Same as +SF, except checksumming is also performed\n\
\n\
    +U: UNTOUCH: Create a script which can be saved into a file for later\n\
	recovery of modification dates following a /bin/touch, e.g.:\n\
	    els +U -R * > untouch  # Create untouch script file\n\
	    /bin/touch *           # Clobber modification dates\n\
	    sh untouch             # Recover modification dates\n\
\n\
    +W: WATCH: Watch the progress of how many files have been examined.\n\
	This option is best used in conjunction with +F filtering when\n\
	output is sparse and there are numerous files being examined or\n\
	when stdout is being redirected to a file, e.g.:\n\
	    els +W +F'Q{s=0}' +NF -laR\n\
	    els +W -laR > /tmp/output\n\
	Additionally, +W will intercept SIGQUIT (i.e. ^\\) while listing\n\
	and display the name of the directory currently being processed.\n\
\n\
    +X: EXECUTE: The output of the +G format will be executed instead of\n\
	listed for each file.  If +X is followed by a number then els will\n\
	terminate whenever the command returns a status of some other value,\n\
	otherwise the return status will be ignored.  Additionally, if +V is\n\
	specified then each command will be echoed before being executed.\n\
\n\
	E.g., fix the mode of any .c, .h, .cc, .hh file having 'x' mode set\n\
	in any of its fields:\n\
	    els +F'P{+x}' +I'*.{c,h,cc,hh}' +G'\"chmod a-x %%N\"' +XV -RA\n\
\n\
	E.g., fix the mode of any file having 'x' mode set in the 'u' field\n\
	but is missing from the 'g' or 'o' fields:\n\
	    els +F'P{u+x,g-x|u+x,o-x}' +G'\"chmod go+x %%N\"' +X0 +V -RA\n\
\n\
");
  }
  
  if (more)
    pclose(out);
  
  return;
}


void analyze_options()
{
  char *cp;

  if (stamping) stamp_setup(2);

  if (untouch > 0) untouch_setup(2);

  if (execute_mode)
  {
    if (!G_option) error_msg("+X option requires +G option");
    /* Dissallow +X if running as "ls": */
    if (strcmp(Progname, "ls") == 0) error_msg("+X option disallowed");
  }


  /* When filtering without the +NF option being specified, dot-dirs can
     make for very sloppy output! */
  if (filtering) list_dotdir = FALSE;


  /* root gets special treatment under BSD: */
  if (ANYBIT(Sem,SEM_BSD) && Whoami == 0) list_hidden = TRUE;


  /* The -g flag implies long listing for SYS5: */
  if (ANYBIT(Sem,SEM_SYS5) && g_flag) list_long = TRUE;


  /* Define Name and Time formats for a long listing, unless either +N
     or +T were explicitly defined on the command-line: */
  if (list_long)
  {
    /* Defer to any previous N option from the command-line: */
    if (!N_option)
    {
      do_options("+Nf");
      N_option = FALSE;	/* reset deferrability */
    }

    /* Defer to any previous T option from the command-line: */
    if (!T_option)
    {
      if (ANYBIT(Sem,SEM_ELS))
	do_options("+TM%_DYt");
      else
	do_options("+TM%_DQ");
      T_option = FALSE;	/* reset deferrability */
    }
  }


  /* Generate G_format if no implicit G_format given on the command-line: */
  if (!G_option && munge == 0)
  {
    static char cmd[32];
    if (list_long)
    {
      char *gfmt;
      char *uid, *gid;

      /* NOTE: SYS5 interprets the -g flag to mean "exclude the UID *and*
	 perform a long listing", whereas BSD interprets the -g flag to mean
	 "include GID *if* performing a long listing".

	 It's also true that BSD has no -n or -o flag, however, just for fun
	 let's also support -n and -o for BSD with the understanding that
	 when combined -n and -o that the -g flag will always follow SYS5
	 semantics. */

      if (list_long_numeric && list_long_omit_gid)
      {
	uid = "%-8U";
	gid = "";
      }
      else if (list_long_numeric)
      {
	uid = "%-8U";
	gid = "%-8G";
      }
      else if (list_long_omit_gid)
      {
	uid = "u";
	gid = "";
      }
      else
      {
	uid = "u";
	gid = "g";
      }

      if (ANYBIT(Sem,SEM_SYS5) || list_long_numeric || list_long_omit_gid)
      {
	/* Take -g flag to mean omit UID [sic] in this context: */
	if (g_flag) uid = "";
      }
      else if (ANYBIT(Sem,SEM_BSD))
      {
	/* Handle -g flag a la BSD: */
	if (!g_flag) gid = "";
      }

      /* Handle -G flag a la GNU: */
      if (G_flag) gid = "";

      /* If ACLs are supported on this platform, then scrunch the link-count
	 so as to maintain columns following ACL '+'.  This can violate els'
	 philosophy of having every item separated by a space by causing the
	 link-count to butt-up against the '+' if an ACL is present and the
	 link-count is greater than 99: */
      {
	Boole gfmt_with_acl;
# if defined(HAVE_ACL)
	/* ACLs are supported on this platform: */
	gfmt_with_acl = TRUE;
#   if defined(SUNOS)
	/* SunOS decides at run-time whether or not ACLs are supported: */
	if (osVersion < 50500) gfmt_with_acl = FALSE;
#   endif
# else
	/* ACLs are NOT supported on this platform: */
	gfmt_with_acl = FALSE;
# endif
	if (gfmt_with_acl)
	{
	  if (fielding)
	    gfmt = "+G%s%s~tpA~l%s%ss%cnL";
	  else
	    gfmt = "+G%s%s~tp%%+A%%+l~%s%ss%cnL";
	}
	else
	  gfmt = "+G%s%s~tp~l%s%ss%cnL";
      }
      
      sprintf(cmd, gfmt,
	      list_inode ? "i" : "",
	      list_size_in_blocks ? "S" : "",
	      uid,
	      gid,

	      list_atime ? TIME_ACCESSED :
	      list_ctime ? TIME_MODE_CHANGED :
	      /*list_mtime*/ TIME_MODIFIED);
    }
    else
      sprintf(cmd, "+G%s%sn",
	      list_inode ? "i" : "",
	      list_size_in_blocks ? "S" : "");
    
    do_options(cmd);
    G_option = FALSE; /* G_format was not specified on the command-line */
  }


  /* Munge-levels 3 and 4 deal with access times.  It is much more
     interesting to see the access time of the file being pointed at rather
     than the access time of symbolic link itself.  Munge level 4 changes
     the modification dates, thus if the file is a symbolic link we want to
     use the modification time of the file being pointed at. */
  if (munge >= 3 && munge <= 4) do_options("-L");

  /* Munge level 2 displays access times of the files that are within
     a hierarchy--avoid symbolic links that point elsewhere: */
  if (munge == 2) do_options("+FT{~l}");


  /* Unless +N was explicitly defined on the command line, examine +G format
     for the "N" directive which if found then the desired +N format is "F": */
  if (!N_option)
  {
    char list[2];
    list[0] = FULL_NAME_FORMAT;
    list[1] = CNULL;
    if (find_first_directive(list, G_format) != NULL)
      N_format = "F";
  }
    
  /* Determine whether we will be using full file names by looking for
     'F' or 'd' directives in either the +N or +G formats: */
  {
    char list[3];
    list[0] = FULL_NAME;
    list[1] = DIR_NAME;
    list[2] = CNULL;
    using_full_names = ((find_first_directive(list, N_format) != NULL) ||
			(find_first_directive(list, G_format) != NULL));
  }

  /* Determine the default sort-by-time field by finding the first
     occurence of 'm', 'a', or 'c' in the +G format: */
  {
    cp = find_first_directive("mac", G_format);
    if (cp == NULL)
      first_mac = CNULL;
    else
      first_mac = *cp;
  }

    
  /* Avoid "ls" trimmings if using full names or executing output: */
  avoid_trimmings = using_full_names || execute_mode;

  /* dotdirs are very annoying and very redundant when avoiding trimmings
     as the parent directory has already been listed: */
  if (list_dotdir && avoid_trimmings) list_dotdir = FALSE;

  /* If following /bin/ls semantics and if traverse option NOT explicitly
     specified, then enable all traversing: */
  /* Should this be followed for BSD, SYS5, GNU, semantics ??? */
  if (!traverse_specified && ANYBIT(Sem,SEM_BSD|SEM_SYS5|SEM_GNU|SEM_LS))
  {
    traverse_mp = TRUE;
    traverse_expanded_symlink = TRUE;
  }

  /* BSD's (i.e. SunOS4's) and GNU's /bin/ls truncates long USER/GROUP IDs: */
  if (ANYBIT(Sem,SEM_BSD|SEM_GNU))
  {
    TruncateId = TRUE;
  }

  return;
}


void append_dir(Dir_List *dlist,
		char *fname,
		int fname_size)
{
  register Dir_Item *file;
  
  /* NB: Rather than have to calculate strlen(fname), I use d_reclen
     as a rough guess.  It is assumed that d_reclen >= strlen(fname). */

  /* Append file name to directory list: */
  file = dir_item_alloc();
  if (dlist->head == NULL)
    dlist->head = file;
  else
    dlist->tail->next = file;
  dlist->tail = file;
  file->next = NULL;
  
  if (fname_size > 0)
  {
    file->mem = mem_alloc(fname_size+1); /* +1 required by Linux */
    file->fname = file->mem;
    strcpy(file->fname, fname);
  }
  else
  {
    /* Command line arguments do not need to have mem alloc'ed: */
    file->mem = NULL;
    file->fname = fname;
  }
  
  file->readdir_errno = 0;
  file->stat_errno = 0;
  file->sort = file;
  
  return;
}


void copy_info(Dir_Item *dest, Dir_Item *src)
{
  /* Copy just the control values and info portion of a Dir_Item:
     (This wouldn't be necessary if the same Dir_Item could exist on
     more than one Dir_List or if we didn't need to know the stat info
     of the Dir_Item in the secondary list.) */
  dest->islink = src->islink;
  dest->isdir = src->isdir;
  dest->isdev = src->isdev;

  dest->dotdir = src->dotdir;
  dest->hidden = src->hidden;
  dest->listable = src->listable;
  dest->searchable = src->searchable;

  dest->readdir_errno = src->readdir_errno;
  dest->stat_errno = src->stat_errno;
  dest->info = src->info;

  return;
}


void read_dir(Dir_List *dlist, char *dname, DIR *dir)
{  
  register struct dirent *dp;
  int save_errno = errno;
  int readdir_errno;

  do {
    errno = 0;
    dlist->head = NULL;
    sigEvent = FALSE;
    while (!sigEvent && (dp = readdir(dir)) != NULL)
      append_dir(dlist, dp->d_name, dp->d_reclen);
    
    /* Note: Some OSes set the final errno from readdir to indicate no more
       entries, in which case it shouldn't be considered an error.  Only in
       cases where no reads occurred should errno be considered valid.  */
    readdir_errno = errno;
    
    /* Re-read directory if signal took place during read: */
    if (sigEvent)
    {
      /*fprintf(stderr, "readdir errno == %d, dlist == %p\n",
	errno, dlist->head);*/
      if (dlist->head != NULL) free_dir(dlist);
      rewinddir(dir);
    }
  } while (sigEvent);

  /* If readdir() fails on the first access then dlist->head will be NULL
     in which case append a dummy directory head of "".  There are two
     known cases during which this can occur: 1) listing a corrupt
     directory on a damaged local disk, and 2) while logged in as "root"
     listing a directory that is NFS mounted from a host that doesn't
     export with "root=myhost" and the directory has settings "chmod 700;
     chown root:root" (stat() detects errors only for unprivileged users
     in this situation). */
  if (dlist->head == NULL)
  {
    /* Append dummy directory head and record errno: */
    append_dir(dlist, "", 0);
    dlist->head->readdir_errno = readdir_errno;

    /* Something's very off if errno NOT set and directory is empty: */
    if (readdir_errno == 0)
    {
      fprintf(stderr, "%s: %s: Directory has no \".\" or \"..\"\n",
	      Progname, dname);
      listingError = TRUE;
    }
  }
  else
  {
    /* The "." dir is always the very first item in the list, but search
       for it anyway just in case: */
    register Dir_Item *file = dlist->head;
    while (file != NULL && strcmp(file->fname, ".") != 0)
      file = file->next;
    CwdItem = file;
  }

  errno = save_errno;
  return;
}


void stat_dir(Dir_List *dlist, char *dname)
{
  register Dir_Item *file;
  char *path;
  int sts;
  int save_errno = errno;
  
  dir_block_total = 0;
  dir_file_count = 0;

  for (file = dlist->head; file != NULL; file = file->next)
  {
    char *slash;

    /* Start by setting as many attributes as possible.  These attributes
       must not depend on any information from stat(): */

    path = full_name(dname, file->fname);

    /* Deal with cases such as: "els ../dir ../.. ../../.xyz" */
    slash = strrchr(file->fname, '/');
    if (slash)
      slash++;
    else
      slash = file->fname;
    file->dotdir = strcmp(slash,".") == 0 || strcmp(slash,"..") == 0;
    file->hidden = !file->dotdir && slash[0] == '.';
    
    /* Set the remaining attributes that are dependent on stat(): */

    errno = 0;
# if defined(S_IFLNK)
    /* System has symbolic links: */
    sts = sigSafe_lstat(path, &file->info);
    file->islink = (sts == 0 && S_ISLNK(file->info.st_mode));
    if (file->islink)
    {
      if (expand_symlink)
      {
	/* Return info regarding the actual file rather than the link: */
	sts = sigSafe_stat(path, &file->info);
      }
      else if (zero_st_info && VersionLevel >= 143)
      {
	/* If not expanding symlinks then modification and access times can
	   be annoying as st_mtime gets changed whenever a symlink gets
	   copied via dump/restore, tar, etc.; also, st_atime gets updated
	   whenever listing a symlink. */
	if ((zero_st_mask & ZERO_LINK_MTIME) != 0) file->info.st_mtime = 0;
	if ((zero_st_mask & ZERO_LINK_ATIME) != 0) file->info.st_atime = 0;
	/* When first created, the uid/gid of a symbolic link is that
	   of the user, but after being copied via dump/restore, tar,
	   etc. the symbolic link ends up owned by root: */
	if ((zero_st_mask & ZERO_LINK_OWNER) != 0)
	{
	  file->info.st_uid = 0;
	  file->info.st_gid = 0;
	}
      }
    }
# else
    /* System does not have symbolic links: */
    sts = sigSafe_stat(path, &file->info);
    file->islink = FALSE;
# endif
    file->stat_errno = errno;
    if (Debug&4)
      printf("st_blksize = %lu (%s)\n", (Ulong)file->info.st_blksize, path);

    if (!file->stat_errno)
    {
      file->isdir = S_ISDIR(file->info.st_mode);
      if (file->isdir && zero_st_info && VersionLevel >= 143)
      {
	/* While directory modification times are usually preserved after
	   being copied via dump/restore, tar, etc., they can still be a
	   distraction when changes occur: */
	if ((zero_st_mask & ZERO_DIR_MTIME) != 0) file->info.st_mtime = 0;
	/* Access time can be annoying when listing directories as st_atime
	   gets changed whenever listing the contents of a directory: */
	if ((zero_st_mask & ZERO_DIR_ATIME) != 0) file->info.st_atime = 0;
	/* Directories can change size over time or after being copied
	   via dump/restore, tar, etc.: */
	if ((zero_st_mask & ZERO_DIR_SIZE) != 0) file->info.st_size = 0;
      }

      file->isdev = (S_ISBLK(file->info.st_mode) ||
		     S_ISCHR(file->info.st_mode));
      if (file->isdev && zero_st_info && VersionLevel >= 144)
      {
	if ((zero_st_mask & ZERO_DEV_MTIME) != 0) file->info.st_mtime = 0;
	if ((zero_st_mask & ZERO_DEV_ATIME) != 0) file->info.st_atime = 0;
	/* Note: Sometime the owner of a device file matters, sometimes
	   it doesn't; thus, "owner" should never be zeroed. */
      }

      filter_file(file, path);
      if (file->listable) dir_block_total += get_stat_blocks(&file->info);
    }
    else
    {
      file->isdir = FALSE;
      file->isdev = FALSE;
      /* Even though an error was detected during stat(), we have enough
	 information to see if it is to be filtered from our display list.
	 If the file qualifies for filtering, then no error needs be reported
	 during the output stage! */
      filter_file(file, path);
    }

    if (file->listable) dir_file_count++;

    if (watch_progress)
    {
      static int nfiles = 0;
      static int nreport = 0;
      if (nfiles == nreport)
      {
	sigAskAbort = FALSE; /* Ask to abort via ^C only once an interval */
	fprintf(ttyout, "%d\r", nfiles);
	fflush(ttyout);
	nreport += 1000;
      }
      nfiles++;
    }
  }

  errno = save_errno;
  return;
}


void name_dir(Dir_List *dlist, char *dname)
{
  Boole skip_total = filtering && (dir_file_count == 0);

  if (!avoid_trimmings && !skip_total)
  {
    Boole list_total = list_long;
    Boole unreadable = (dlist == NULL || dlist->head->fname[0] == CNULL);
    if (output_nlines > 0) putchar('\n');

    if (ALLBITS(Sem,SEM_SYS5|SEM_LS))
    {
      /* Trimmings a la SYS5/LS: */
      /* SunOS5,HPUX10 list both dname and total, even if unreadable(!),
	 except for OSF1 which behaves more like BSD if unreadable. */
      if (dname[0] != CNULL && (recursive || multiple_file_args))
      {
	if (zero_file_args && recursion_level > 1 &&
	    strncmp(dname, "./", 2) != 0)
	  printf("./%s:\n", dname);
	else
	  printf("%s:\n", dname);
      }
    }
    else if (ALLBITS(Sem,SEM_GNU|SEM_LS))
    {
      /* Trimmings a la GNU/LS: */
      if (dname[0] != CNULL && recursive)
      {
	/* GNU doesn't list dname or total if unreadable: */
	if (unreadable)
	  list_total = FALSE;
	else
	  printf("%s:\n", dname);
      }
    }
    else
    {
      /* Trimmings a la BSD/LS and ALL/ELS: */
      if (multiple_file_args || recursion_level > 1)
      {
	/* BSD and ELS don't list dname or total if unreadable: */
	if (unreadable)
	{
	  /* Unreadable directory a la BSD and ELS writes an informational
	     message to stdout: */
	  printf("%s unreadable\n", dname);
	  list_total = FALSE;
	}
	else
	  printf("%s:\n", dname);
      }
    }

    if (list_total)
    {
      char nblocks[64];
      sprintf(nblocks, F_st_blocks(FALSE,0),
	      ALLBITS(Sem,SEM_SYS5|SEM_LS) ?
	      dir_block_total : dir_block_total/2);
      printf("total %s\n", nblocks);
    }
    output_nlines++;
  }

  if (dlist == NULL)
  {
    if (!warning_suppress) fprintf(stderr, "%s: %s: Permission denied\n",
				   Progname, dname);
    listingError = TRUE;
  }
  else if (dlist->head->readdir_errno != 0)
  {
    /* SunOS's /bin/ls lets readdir() errors slide-by undectected (only
       stat() errors get reported): */
    if (ANYBIT(Sem,SEM_ELS))  /* What does SEM_GNU do??? */
    {
      char msg[MAX_MSG];
      int save_errno = errno;
      sprintf(msg, "%s: %s", Progname, dname);
      errno = dlist->head->readdir_errno;
      if (!warning_suppress) perror(msg);
      errno = save_errno;
      listingError = TRUE;
    }
  }

  return;
}	


void sort_dir(Dir_List *dlist, char *dname)
{
  if (!unsort)
  {
    if (time_sort)
      sort_dir_time(dlist, dname);
    
    else if (ANYBIT(Sem,SEM_BSD|SEM_GNU) && list_ctime)
      /* BSD's and GNU's ls -c implies ls -ct, but SYS5's does NOT! */
      sort_dir_time(dlist, dname);
    
    else
      sort_dir_name(dlist, dname);
  }
    
  return;
}


void sort_dir_time(Dir_List *dlist, char *dname)
{
  register Dir_Item *ptr1, *ptr2;
  register Dir_Item *dp;
  char sort_by;
  
  if (list_atime)
    /* Sort by time accessed regardless of what's in G_format: */
    sort_by = TIME_ACCESSED;
  else if (list_ctime)
    /* Sort by time mode changed regardless of what's in G_format: */
    sort_by = TIME_MODE_CHANGED;
  else
  {
    /* Use first occurrence of either m, a, or c as found in G_format: */
    if (first_mac == CNULL)
      sort_by = TIME_MODIFIED;	/* default */
    else
      sort_by = first_mac;
  }
  
  for (ptr1 = dlist->head; ptr1 != NULL; ptr1 = ptr1->next)
  {
    for (ptr2 = ptr1->next; ptr2 != NULL; ptr2 = ptr2->next)
    {
      if (sort_by == TIME_MODIFIED)
      {
	if (reverse_sort ?
	    (Ulong)ptr2->sort->info.st_mtime < (Ulong)ptr1->sort->info.st_mtime :
	    (Ulong)ptr2->sort->info.st_mtime > (Ulong)ptr1->sort->info.st_mtime)
	{
	  /* Switch the entry pointers */
	  dp = ptr1->sort;
	  ptr1->sort = ptr2->sort;
	  ptr2->sort = dp;
	}
      }

      else if (sort_by == TIME_ACCESSED)
      {
	if (reverse_sort ?
	    (Ulong)ptr2->sort->info.st_atime < (Ulong)ptr1->sort->info.st_atime :
	    (Ulong)ptr2->sort->info.st_atime > (Ulong)ptr1->sort->info.st_atime)
	{
	  /* Switch the entry pointers */
	  dp = ptr1->sort;
	  ptr1->sort = ptr2->sort;
	  ptr2->sort = dp;
	}
      }

      else if (sort_by == TIME_MODE_CHANGED)
      {
	if (reverse_sort ?
	    (Ulong)ptr2->sort->info.st_ctime < (Ulong)ptr1->sort->info.st_ctime :
	    (Ulong)ptr2->sort->info.st_ctime > (Ulong)ptr1->sort->info.st_ctime)
	{
	  /* Switch the entry pointers */
	  dp = ptr1->sort;
	  ptr1->sort = ptr2->sort;
	  ptr2->sort = dp;
	}
      }
    }
  }
  
  return;
}


void sort_dir_name(Dir_List *dlist, char *dname)
{
  register Dir_Item *ptr1, *ptr2;
  register Dir_Item *dp;
  
  for (ptr1 = dlist->head; ptr1 != NULL; ptr1 = ptr1->next)
  {
    for (ptr2 = ptr1->next; ptr2 != NULL; ptr2 = ptr2->next)
    {
      if (reverse_sort ?
	  strcmp(ptr2->sort->fname, ptr1->sort->fname) > 0 :
	  strcmp(ptr2->sort->fname, ptr1->sort->fname) < 0)
      {
	/* Switch the entries(by pointer only) */
	dp = ptr1->sort;
	ptr1->sort = ptr2->sort;
	ptr2->sort = dp;
      }
    }
  }
  
  return;
}


void list_dir(Dir_List *dlist,
	      char *dname)
{
  register Dir_Item *ptr;
  Dir_List sub_dlist;
  DIR *dir;
  
  recursion_level++;
  sub_dlist.head = NULL;
  
  for (ptr = dlist->head; ptr != NULL; ptr = ptr->next)
  {
    list_file(ptr->sort, dname);

    /* Recursively list directory (if appropriate): */
    if (ptr->sort->searchable ||
	(recursion_level == 1 && ptr->sort->dotdir))
    {
      if (expand_directories && (recursion_level == 1 || recursive))
      {
	if (recursion_level == 1 || !ptr->sort->dotdir)
	{
	  append_dir(&sub_dlist, ptr->sort->fname, 0);
	  copy_info(sub_dlist.tail, ptr->sort);
	}
      }
    }
  }
  
  if (sub_dlist.head != NULL)
  {
    for (ptr = sub_dlist.head; ptr != NULL; ptr = ptr->next)
    {
      char cur_dname[MAX_FULL_NAME];
      if (dname[0] == CNULL || strcmp(dname, ".") == 0)
	Void strcpy(cur_dname, ptr->sort->fname);
      else
      {
	char *fmt;
	if (dname[strlen(dname)-1] == '/')
	  fmt = "%s%s";
	else
	  fmt = "%s/%s";
	Void sprintf(cur_dname, fmt, dname, ptr->sort->fname);
      }
      
      if ((dir = opendir(cur_dname)) == NULL)
      {
	name_dir(NULL, cur_dname);
      }
      else
      {
	Boole ok_to_cross = TRUE;	/* allow crossing by default */
	struct stat *old_mp = NULL;

	/* Save top-most mount-point: */
	if (recursion_level == 1) set_mp(&ptr->sort->info);

	if (ptr->sort->islink)
	{
	  /* Designate a temporary mount-point for expanded symlinked
	     directory and then cross if option enabled: */
	  old_mp = set_mp(&ptr->sort->info);
	  ok_to_cross = traverse_expanded_symlink;
	}
	else if (!traverse_mp)
	  /* If not traversing mount-points then allow crossing only if
	     the mount-point hasn't changed since last set: */
	  ok_to_cross = diff_mp(&ptr->sort->info) == 0;
	
	if (ok_to_cross)
	{
	  Dir_List cur_dlist;
	  char *saveCwdPath = CwdPath;
	  CwdPath = cur_dname;
	  read_dir(&cur_dlist, cur_dname, dir);
	  stat_dir(&cur_dlist, cur_dname);
	  name_dir(&cur_dlist, cur_dname);
	  sort_dir(&cur_dlist, cur_dname);
	  list_dir(&cur_dlist, cur_dname);
	  free_dir(&cur_dlist);
	  closedir(dir);
	  CwdPath = saveCwdPath;
	}
	else if (verbose_mode)
	{
	  char *path = full_name(dname, ptr->sort->fname);
	  fprintf(stderr, "%s: %s: %s not traversed\n",
		  Progname, path,
		  ptr->sort->islink ? "symlink" : "mount-point");
	}

	/* Restore old mount-point prior to symlink traversal (if any): */
	if (old_mp != NULL) set_mp(old_mp);
      }
    }
    free_dir(&sub_dlist);
  }
  
  recursion_level--;

  return;
}


void free_dir(Dir_List *dlist)
{
  register Dir_Item *ptr1, *ptr2;

  /* free_dir() invalidates the Cwd* global variables: */
  CwdItem = NULL;
  CwdBlocksize = 0;

  ptr1 = dlist->head;
  while (ptr1 != NULL)
  {
    ptr2 = ptr1->next;
    if (ptr1->mem != NULL) mem_free(ptr1->mem);
    dir_item_free(ptr1);
    ptr1 = ptr2;
  }
  dlist->head = NULL;

  return;
}


struct stat *getCwdInfo(char *path)
{
  static struct stat info;
  static Boole lastCwdWasDot = FALSE;
  if (CwdItem != NULL)
    return(&CwdItem->info);
  else
  {
    char *cwd;
    char *slash = strrchr(path, '/');
    if (slash != NULL)
    {
      *slash = CNULL;
      cwd = path;
    }
    else
    {
      if (lastCwdWasDot) return(&info);
      cwd = ".";
    }
      
    /* Ignore return status, as stat_dir() will eventually detect any errno: */
# if defined(S_IFLNK)
    /* System has symbolic links: */
    sigSafe_lstat(cwd, &info);
# else
    /* System does not have symbolic links: */
    sigSafe_stat(cwd, &info);
# endif

    if (slash != NULL)
    {
      *slash = '/';
      lastCwdWasDot = FALSE;
    }
    else
    {
      lastCwdWasDot = TRUE;
    }
    return(&info);
  }
}


#if defined(HPUX)
# define LOCAL_BLOCKSIZE 1024
#else
# define LOCAL_BLOCKSIZE  512
#endif

/* NB: The blocksize calculated by this routine has *nothing* to do with
   the st_blksize element returned by stat()!  Regreatably st_blksize
   refers to the preferred blocksize for file system I/O and almost never
   equals the size of each st_blocks unit, whereas getCwdBlocksize()
   attempts to heuristically calculate the size of each st_blocks unit
   using the characteristics of the current working directory (cwd). */

int getCwdBlocksize(char *path)
{
  static int lastCwdBlocksize = LOCAL_BLOCKSIZE;
# ifdef NO_STAT_BLOCKS
  CwdBlocksize = LOCAL_BLOCKSIZE;
# else
  /* FIXME???: If +z is specified this routine returns lastCwdBlocksize */
  if (CwdBlocksize == 0)
  {
    struct stat *cwd_info = getCwdInfo(path);
    int bs;

    /* ANOMALOUS CASES:
       1) Most HPUX directories are a multiple of 1024, but it is possible
       for directories under HPUX's vxfs filesystem to be as small as 26 bytes.
       2) Directory sizes of ClearCase's mvfs filesystem under SunOS5 can be
       *any* non-negative integer.  Moreover, a ClearCase mvfs directory
       containing 0 files will have 0 blocks and 0 bytes.
       3) SunOS5.6+'s directory sizes of an indirect autofs map are 1 byte if
       the option -nobrowse is not specified and the filesystem is not mounted.
       */

    if ((Debug&2) && (cwd_info->st_size <= 0 || cwd_info->st_blocks == 0 ||
		      (cwd_info->st_size & 511) != 0))
      printf("STRANGE DIRECTORY: blocks=%lu, size=%lu (%s)\n",
	     (Ulong)cwd_info->st_blocks, (Ulong)cwd_info->st_size, path);

    /* Round up to the next power of 2 >= 512: */
    if (cwd_info->st_blocks == 0)
    {
      /* Files with errno set will probably end up here (this makes 
	 the file seem sparse which is OK so that perror() gets called): */
      bs = lastCwdBlocksize;  /* Best available guess??? */
    }
    else
    {
      bs = cwd_info->st_size / cwd_info->st_blocks;
      if (bs < 256)
	/* Attempt to handle HPUX vxfs, ClearCase mvfs, and SunOS5.6+
	   browsable autofs filesystems: */
	bs = lastCwdBlocksize;  /* Best available guess??? */
      else if (bs <= 512)
	bs = 512;
      else if (bs <= 1024)
	bs = 1024;
      else if (bs <= 2048)
	bs = 2048;
      else if (bs <= 4096)
	bs = 4096;
      else if (bs <= 8192)
	bs = 8192;
      else
      {
	/* A blocksize should never bigger than this!?: */
	fprintf(stderr, "Block size error\n");
	exit(LISTING_ERROR);
      }
    }

    if (Debug&1)
      printf("Blocksize = %d, %lu = %lu/%lu (%s)\n", bs,
	     (Ulong)cwd_info->st_size / (Ulong)cwd_info->st_blocks,
	     (Ulong)cwd_info->st_size, (Ulong)cwd_info->st_blocks,
	     path);
    if ((Debug&2) && lastCwdBlocksize != bs)
      printf("BLOCKSIZE CHANGING FROM %d TO %d (%s)\n",
	     lastCwdBlocksize, bs, path);
    
    CwdBlocksize = bs;
    lastCwdBlocksize = bs;
  }
# endif
  return(CwdBlocksize);
}


void filter_file(Dir_Item *file, char *path)
{
  int i;
  Boole file_is_listable = TRUE;
  Boole file_is_searchable = FALSE;

  if (file->fname[0] == CNULL)
  {
    /* This situation is unlikely!?: */
    file_is_listable = FALSE;
  }
  else
  {
    /*****************************************/
    /* Check for ".", "..", and ".xxx" files */
    /*****************************************/
    
    /* The user asked for the file explicitly on the command-line
       whenever recursion_level == 0, thus, always grant such a
       request; otherwise, hidden files must meet the list_* criteria.
       */

    if (file->dotdir)
      file_is_listable = list_dotdir || recursion_level == 0;
    else if (file->hidden)
      file_is_listable = list_hidden || recursion_level == 0;
    
    /******************************************/
    /* Check file against "F"ilter expression */
    /******************************************/

    if (file_is_listable && fexpr_count > 0)
    {
      file_is_listable = FALSE;
      for (i = 0; i < fexpr_count && !file_is_listable; i++)
	file_is_listable = filter_fexpr(file, path,
					fexpr_filter[i], NULL);
    }

    /********************************************/
    /* Check file name against "E"xclude filter */
    /********************************************/

    if (file_is_listable && exc_file_count > 0 && !file->isdir)
    {
      Boole found = FALSE;
      for (i = 0; i < exc_file_count && !found; i++)
	found = wildcard_match(file->fname, exc_file_filter[i]);
      file_is_listable = !found;
    }
    
    /*******************************************/
    /* Check dir name against "e"xclude filter */
    /*******************************************/

    if (file_is_listable && exc_dir_count > 0 && file->isdir)
    {
      Boole found = FALSE;
      for (i = 0; i < exc_dir_count && !found; i++)
	if (strchr(exc_dir_filter[i], '/'))
	  /* Exclude a directory having a path component (this ad hoc
	     code facilitates cases such as "els +e/usr/local /usr"): */
	  found = wildcard_match(path, exc_dir_filter[i]);
	else
	  /* Exclude a directory having *no* path component: */
	  found = wildcard_match(file->fname, exc_dir_filter[i]);
      file_is_listable = !found;
    }
    
    /********************************************/
    /* Check file name against "I"nclude filter */
    /********************************************/

    if (file_is_listable && inc_file_count > 0 && !file->isdir)
    {
      Boole found = FALSE;
      for (i = 0; i < inc_file_count && !found; i++)
	found = wildcard_match(file->fname, inc_file_filter[i]);
      file_is_listable = found;
    }

    /* Special circumstances warrant not listing directories when explicitly
       "I"ncluding files (although directories will still be considered
       searchable; moreover, the search filters "e" and "i" will still play
       a role): */
    if (file_is_listable && inc_file_count > 0 && file->isdir)
      file_is_listable = FALSE;

    /*******************************************/
    /* Check dir name against "i"nclude filter */
    /*******************************************/

    if (file_is_listable && inc_dir_count > 0 && file->isdir)
    {
      Boole found = FALSE;
      for (i = 0; i < inc_dir_count && !found; i++)
	found = wildcard_match(file->fname, inc_dir_filter[i]);
      file_is_listable = found;
    }

    /************************************/
    /* Decide whether dir is searchable */
    /************************************/

    /* When filtering a subdirectory, we want to search the subdirectory's
       contents, even if it doesn't meet our listing criterion, otherwise
       we will end up with hardly any files or subdirectories to search!
       The exceptions to this rule are: 1) when the user explicitly
       specifies directories to include using the "+i" option; 2) when
       listing normally or via exclusion, we want to search subdirectories
       only if they didn't get scratched from our search list; 3) the
       directory is hidden and list_hidden has not been enabled.  Even with
       these exceptions we still end up searching quite a few files. */

    if (file->isdir)
    {
      if (file_is_listable)
	file_is_searchable = TRUE;
      else if (file->hidden && !list_hidden && VersionLevel >= 144)
	file_is_searchable = FALSE;
      else
      {
	file_is_searchable = TRUE;
	if (file_is_searchable && exc_dir_count > 0)
	{
	  Boole found = FALSE;
	  for (i = 0; i < exc_dir_count && !found; i++)
	    if (strchr(exc_dir_filter[i], '/'))
	      /* Exclude a directory having a path component (this ad hoc
		 code facilitates cases such as "els +e/usr/local /usr"): */
	      found = wildcard_match(path, exc_dir_filter[i]);
	    else
	      /* Exclude a directory having *no* path component: */
	      found = wildcard_match(file->fname, exc_dir_filter[i]);
	  file_is_searchable = !found;
	}
	if (file_is_searchable && inc_dir_count > 0)
	{
	  Boole found = FALSE;
	  for (i = 0; i < inc_dir_count && !found; i++)
	    found = wildcard_match(file->fname, inc_dir_filter[i]);
	  file_is_searchable = found;
	}
      }
    }
  }

  /* FINALLY, file is listable if it met the filtering criteria and
     1) it isn't a dir, 2) it's a non-expandable dir, or 3) the list_*
     criteria specify directories (unless this is the very first recursion
     level and trimmings aren't being avoided).  All this is needed to
     mimic "ls" behavior: */
  
  file->listable =
    file_is_listable && (!file->isdir ||
			 !expand_directories ||
			 (list_directories && (recursion_level > 0 ||
					       avoid_trimmings)));

  file->searchable = file_is_searchable;

  return;
}


Boole filter_fexpr(Dir_Item *file, char *path, char *ptr, char **end)
{
  /******************************************/
  /* Check file against "F"ilter expression */
  /******************************************/
  
  Boole file_is_listable = FALSE;
  Boole looking_expr;
  static int level = 0;
  if (level++ == 0)
  {
    Current_Opt = "+F";
    Current_Arg = ptr;
  }

  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole_Function filter = NULL;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }
      
      switch (*ptr)
      {
      case 'A':
	filter = filter_access;
	break;
      case 'T':
	filter = filter_type;
	break;
      case 'P':
	filter = filter_perm;
	break;
      case 'Q':
	filter = filter_quantity;
	break;
      case 'U':
	filter = filter_unusual;
	break;
      default:
	expecting_error(!negate, "ATPQU", FALSE, FALSE, ptr);
	break;
      }

      ptr++;
      if (*ptr != fexpr_BEGIN) expecting_error(FALSE, "{", FALSE, FALSE, ptr);
      ptr++;
      
      term_good &=
	negate ? !filter(file, path, ptr, &ptr)
	  :       filter(file, path, ptr, &ptr);
      
      ptr--;
      if (*ptr != fexpr_END) expecting_error(FALSE, "}", FALSE, FALSE, ptr+1);
      ptr++;
      
      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case CNULL:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	expecting_error(FALSE, NULL, TRUE, FALSE, ptr);
	break;
      }

      if (*ptr) ptr++;
    }

    file_is_listable |= term_good;
  }
  if (--level == 0)
  {
    Current_Opt = NULL;
    Current_Arg = NULL;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


Boole filter_access(Dir_Item *file, char *path, char *ptr, char **end)
{
  /**************************************/
  /* Check file against "A"ccess filter */
  /**************************************/
  
  Boole file_is_listable = FALSE;
  Boole looking_expr;
  
  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole test_result = FALSE;
      Uint have_access = 0;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }

      while (IS_MEMBER(*ptr, "rwx"))
      {
	switch (*ptr)
	{
	case 'r':
	  have_access |= R_OK;
	  break;
	case 'w':
	  have_access |= W_OK;
	  break;
	case 'x':
	  have_access |= X_OK;
	  break;
	}
	ptr++;
      }

      if (have_access == 0)
	expecting_error(!negate, "rwx", FALSE, FALSE, ptr);

      /* ??? may need to use getaccess() for ACLs under HPUX9 ??? */
      test_result = access(path, have_access) == 0;
      term_good &= negate ? !test_result : test_result;

      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case fexpr_END:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	expecting_error(FALSE, NULL, TRUE, TRUE, ptr);
	break;
      }
      ptr++;
    }

    file_is_listable |= term_good;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


Boole filter_type(Dir_Item *file, char *path, char *ptr, char **end)
{
  /************************************/
  /* Check file against "T"ype filter */
  /************************************/
  
  Boole file_is_listable = FALSE;
  Boole looking_expr;
  int fmode = file->info.st_mode;
  int ftype = (fmode & S_IFMT);
  
  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;
    Boole redundancy_check = FALSE;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole test_result = FALSE;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }

      switch (*ptr)
      {
      case symtype_REG:
      case type_REG:	/* Regular */
	test_result = ftype == S_IFREG;
	break;
	
      case symtype_DIR:
      case type_DIR:	/* Directory */
	test_result = ftype == S_IFDIR;
	break;
	
      case type_CHR:	/* Char Special */
	test_result = ftype == S_IFCHR;
	break;
	
      case type_BLK:	/* Block Special */
	test_result = ftype == S_IFBLK;
	break;

      case type_FIFO:	/* Fifo */
#ifdef S_IFIFO
	test_result = ftype == S_IFIFO;
#endif
	break;
      case symtype_LNK:
      case type_LNK:	/* Symbolic Link */
#ifdef S_IFLNK
	if (expand_symlink)
	  /* If expanding symlinks and the user wants symlinks filtered(!)
	     then do what is asked (even though it's a contradiction): */
	  test_result = file->islink;
	else
	  test_result = ftype == S_IFLNK;
#endif
	break;
      case symtype_SOCK:
      case type_SOCK:	 /* Socket */
#ifdef S_IFSOCK
	test_result = ftype == S_IFSOCK;
#endif
	break;
      case symtype_DOOR:
      case type_DOOR:	 /* Door */
#ifdef S_IFDOOR
	test_result = ftype == S_IFDOOR;
#endif
	break;
      case type_SPECIAL:	/* Any Device (c or b) */
	test_result = filter_type(file, path, "c:b", NULL);
	break;

      default:
	expecting_error(!negate, "rdcbplsDS", FALSE, FALSE, ptr);
	break;
      }

      if (!negate)
      {
	if (redundancy_check)
	  redundancy_warning("types", ptr);
	redundancy_check = TRUE;
      }
      ptr++;
      
      term_good &= negate ? !test_result : test_result;
      
      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case fexpr_END:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	expecting_error(FALSE, NULL, TRUE, TRUE, ptr);
	break;
      }
      ptr++;
    }

    file_is_listable |= term_good;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


Boole filter_perm(Dir_Item *file, char *path, char *ptr, char **end)
{
  /******************************************/
  /* Check file against "P"ermission filter */
  /******************************************/

  Boole file_is_listable = FALSE;
  Boole looking_expr;
  register int fmode = file->info.st_mode;
  
  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;
    Boole redundancy_check = FALSE;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole test_result = FALSE;
#define ALL_USER  (S_IWUSR|S_IRUSR|S_IXUSR|S_ISUID|S_ISVTX)
#define ALL_GROUP (S_IWGRP|S_IRGRP|S_IXGRP|S_ISGID)
#define ALL_OTHER (S_IWOTH|S_IROTH|S_IXOTH)
      Boole perm_any = FALSE;
      Boole perm_have = FALSE;
      Boole perm_missing = FALSE;
      Boole perm_exact = FALSE;
      Uint perm_field = 0;
      Uint perm_allow = 0;
      Uint perm_deny = 0;
      Boole octal_value = FALSE;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }
      
      /* Attempt to read an absolute mode (base 8): */
      if (isdigit(*ptr) && *ptr < '8')
      {
	perm_allow = strtoul(ptr, &ptr, 8);
	perm_field = (ALL_USER|ALL_GROUP|ALL_OTHER);
	perm_exact = TRUE;
	octal_value = TRUE;

	if (perm_allow > 07777)
	  filter_error_msg("Out of bounds [MAX == 07777]", ptr-1);
	if (!negate)
	{
	  if (redundancy_check)
	    redundancy_warning("permission fields", ptr-1);
	}
      }
      else
      {
	while (IS_MEMBER(*ptr, "ugoa"))
	{
	  switch (*ptr)
	  {
	  case 'u':
	    perm_field |= ALL_USER;
	    break;
	  case 'g':
	    perm_field |= ALL_GROUP;
	    break;
	  case 'o':
	    perm_field |= ALL_OTHER;
	    break;
	  case 'a':
	    perm_field |= (ALL_USER|ALL_GROUP|ALL_OTHER);
	    break;
	  }
	  ptr++;
	}

	switch (*ptr)
	{
	  case '+':
	    perm_have = TRUE;
	    break;
	  case '-':
	    perm_missing = TRUE;
	    break;
	  case '=':
	    perm_exact = TRUE;
	    break;
	  default:
	    if (perm_field == 0)
	      expecting_error(!negate, "ugoa+-=' or 'octal-value",
			      FALSE, FALSE, ptr);
	    else
	      expecting_error(FALSE, "ugoa+-=",
			      FALSE, FALSE, ptr);
	    break;
	}
	ptr++;
	
	/* If no permission field has been explicitly defined by now,
	   then "any" permission field is assumed: */
	if (perm_field == 0) perm_any = TRUE;
	
	while (IS_MEMBER(*ptr, "rwxstl"))
	{
	  switch (*ptr)
	  {
	  case 'r':
	    perm_allow |= (S_IRUSR|S_IRGRP|S_IROTH);
	    break;
	    
	  case 'w':
	    perm_allow |= (S_IWUSR|S_IWGRP|S_IWOTH);
	    break;
	    
	  case 'x':
	    perm_allow |= (S_IXUSR|S_IXGRP|S_IXOTH);
	    break;
	    
	  case 's':
	    perm_allow |= (S_ISUID|S_ISGID);
	    if (perm_any) perm_field |= (ALL_USER|ALL_GROUP);
	    if ((perm_field & ALL_OTHER) != 0)
	      filter_error_msg("Invalid context", ptr);
	    break;
	    
	  case 't':
	    perm_allow |= (S_ISVTX);
	    if (perm_any) perm_field |= ALL_USER;
	    if ((perm_field & (ALL_GROUP|ALL_OTHER)) != 0)
	      filter_error_msg("Invalid context", ptr);
	    break;
	    
	  case 'l':
	    perm_allow |= (S_ISGID);
	    perm_deny |= (S_IXGRP);
	    if (perm_any) perm_field |= ALL_GROUP;
	    if ((perm_field & (ALL_USER|ALL_OTHER)) != 0)
	      filter_error_msg("Invalid context", ptr);
	    break;
	  }
	  ptr++;
	}
	    
        if (perm_field == 0)
	  perm_field = (ALL_USER|ALL_GROUP|ALL_OTHER);
      }
      
      if (perm_any)
      {
	Uint fmode_user = fmode & ALL_USER;
	Uint fmode_group = fmode & ALL_GROUP;
	Uint fmode_other = fmode & ALL_OTHER;
	
	/* Test for having all specified permissions in any field: */
	if (perm_have)
	  test_result =
	    (((perm_field & ALL_USER) != 0 &&
	      (fmode_user & perm_deny) == 0 &&
	      (fmode_user & perm_allow) == (perm_allow & ALL_USER)) ||
	     ((perm_field & ALL_GROUP) != 0 &&
	      (fmode_group & perm_deny) == 0 &&
	      (fmode_group & perm_allow) == (perm_allow & ALL_GROUP)) ||
	     ((perm_field & ALL_OTHER) != 0 &&
	      (fmode_other & perm_deny) == 0 &&
	      (fmode_other & perm_allow) == (perm_allow & ALL_OTHER)));
	
	/* Test for missing all specified permissions from any field: */
	else if (perm_missing)
	  test_result = 
	    (((perm_field & ALL_USER) != 0 &&
	      ((fmode_user & perm_deny) == (perm_deny & ALL_USER) &&
	       (fmode_user & perm_allow) == 0)) ||
	     ((perm_field & ALL_GROUP) != 0 &&
	      ((fmode_group & perm_deny) == (perm_deny & ALL_GROUP) &&
	       (fmode_group & perm_allow) == 0)) ||
	     ((perm_field & ALL_OTHER) != 0 &&
	      ((fmode_other & perm_deny) == (perm_deny & ALL_OTHER) &&
	       (fmode_other & perm_allow) == 0)));
	
	/* Test for having nothing other than specified permissions
	   in any field: */
	else if (perm_exact)
	  test_result =
	    (((perm_field & ALL_USER) != 0 &&
	      (fmode_user & perm_deny) == 0 &&
	      (fmode_user) == (perm_allow & ALL_USER)) ||
	     ((perm_field & ALL_GROUP) != 0 &&
	      (fmode_group & perm_deny) == 0 &&
	      (fmode_group) == (perm_allow & ALL_GROUP)) ||
	     ((perm_field & ALL_OTHER) != 0 &&
	      (fmode_other & perm_deny) == 0 &&
	      (fmode_other) == (perm_allow & ALL_OTHER)));
      }
      else
      {
	/* Strip masks of undesired fields (IMPORTANT STEP!): */
	perm_allow &= perm_field;
	perm_deny &= perm_field;
	
	/* Test for having all specified permissions
	   in all specified fields: */
	if (perm_have)
	  test_result = ((fmode & perm_deny) == 0 &&
			 (fmode & perm_allow) == perm_allow);

	/* Test for missing all specified permissions
	   in all specified fields: */
	else if (perm_missing)
	  test_result = ((fmode & perm_deny) == perm_deny &&
			 (fmode & perm_allow) == 0);

	/* Test for having nothing other than specified permissions
	   in all specified fields: */
	else if (perm_exact)
	  test_result = ((fmode & perm_deny) == 0 &&
			 (fmode & perm_field) == perm_allow);
      }

      if (!negate)
	redundancy_check = TRUE;

      term_good &= negate ? !test_result : test_result;

      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case fexpr_END:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	if (octal_value)
	  expecting_error(FALSE, NULL, TRUE, TRUE, ptr);
	else
	  expecting_error(FALSE, "rwxstl", TRUE, TRUE, ptr);
	break;
      }
      ptr++;
    }

    file_is_listable |= term_good;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


Boole filter_quantity(Dir_Item *file, char *path, char *ptr, char **end)
{
  /****************************************/
  /* Check file against "Q"uantity filter */
  /****************************************/

  Boole file_is_listable = FALSE;
  Boole looking_expr;
  int fmode = file->info.st_mode;
  int ftype = (fmode & S_IFMT);
  char q_type = CNULL;
  
  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole test_result = FALSE;
      PhVal q_info = 0;
      PhVal q_val = 0;
      PhVal q_val2 = 0;
      int q_op = 0;
      PhVal q_relative = 0;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }
      
      q_type = *ptr;
      switch (*ptr)
      {
      case 'i':
	q_info = file->info.st_ino;
	break;
      case 'A':
#     if !defined(HAVE_ACL)
	filter_warning_msg("Program compiled without ACL support", ptr);
#     endif
	q_info = get_acl_count(file, path);
	break;
      case 'l':
	q_info = file->info.st_nlink;
	break;
      case 'u':
	q_info = file->info.st_uid;
	break;
      case 'g':
	q_info = file->info.st_gid;
	break;
      case 's':
	if (ftype == S_IFBLK || ftype == S_IFCHR)
	  q_info = 0;
	else
	  q_info = file->info.st_size;
	break;
      case 'm':
	q_info = file->info.st_mtime;
	break;
      case 'a':
	q_info = file->info.st_atime;
	break;
      case 'c':
	q_info = file->info.st_ctime;
	break;
      default:
	expecting_error(!negate, "iAlugsmac", FALSE, FALSE, ptr);
	break;
      }
      ptr++;
      
#define Q_EQ 1
#define Q_NE 2
#define Q_GT 3
#define Q_GE 4
#define Q_LT 5
#define Q_LE 6
      switch (*ptr)
      {
      case '=':
	q_op = Q_EQ;
	if (ptr[1] == '=')
	{
	  q_op = Q_EQ; ptr++;
	}
	break;
      case '!':
      case '~':
	q_op = Q_NE;
	ptr++;
	if (*ptr != '=')
	  expecting_error(FALSE, "=", FALSE, FALSE, ptr);
	break;
      case '<':
	q_op = Q_LT;
	if (ptr[1] == '=')
	{
	  q_op = Q_LE; ptr++;
	}
	else if (ptr[1] == '>')
	{
	  q_op = Q_NE; ptr++;
	}
	break;
      case '>':
	q_op = Q_GT;
	if (ptr[1] == '=')
	{
	  q_op = Q_GE; ptr++;
	}
	break;
      default:
	expecting_error(FALSE, "!~<>=", FALSE, FALSE, ptr);
	break;
      }
      ptr++;
      
      /* Read a comparison value: */
      if (IS_MEMBER(q_type, "ug") && *ptr == '.')
      {
	struct stat *cwd_info = getCwdInfo(path);
	if (q_type == 'u')
	  q_val = cwd_info->st_uid;
	else
	  q_val = cwd_info->st_gid;
	ptr++;
      }
      else if (!phLookup(ptr, &ptr, &q_val, &q_relative, &q_val2))
      {
	char *start = ptr;
	/* ??? Allow negative values ??? */
	/* Note: base10 is enforced here because of ambiguity caused
	   by +F'Q{m<0x100D}' */
	if (isdigit(*ptr))
	  q_val = strtoPhVal(ptr, &ptr, 10);
	else if (IS_MEMBER(q_type, "ug"))
	{
	  struct passwd *pwd = NULL;
	  struct group *grp = NULL;
	  char name[MAX_USER_GROUP_NAME+2];  /* allow 1 char overrun + CNULL */
	  int i = 0;
	  while (i <= MAX_USER_GROUP_NAME && IS_NOT_MEMBER(*ptr, "&,|:}") &&
		 *ptr != CNULL)
	    name[i++] = *ptr++;
	  name[i] = CNULL;
	  if (i > 0 && i <= MAX_USER_GROUP_NAME)
	  {
	    if (q_type == 'u')
	    {
	      pwd = getpwnam(name);
	      if (pwd != NULL)
		q_val = pwd->pw_uid;

	    }
	    else if (q_type == 'g')
	    {
	      grp = getgrnam(name);
	      if (grp != NULL)
		q_val = grp->gr_gid;
	    }
	  }
	  if (pwd == NULL && grp == NULL)
	  {
	    char msg[128];
	    sprintf(msg, "%s ID '%s' %s",
		    q_type == 'u' ? "User" : "Group",
		    i == 0 ? "?" : name,
		    i == 0 || i > MAX_USER_GROUP_NAME ? "is illegal" : "fails lookup");
	    filter_error_msg(msg, start);
	  }
	}
	else
	  filter_error_msg("Expected positive integer", ptr);
	
	if (IS_MEMBER(q_type, "mac"))
	{
	  if (isalpha(*ptr))
	  {
	    switch (*ptr)
	    {
	    case 'Y':
	      q_val *= SECS_PER_YEAR;
	      break;
	    case 'M':
	      q_val *= SECS_PER_MONTH;
	      break;
	    case 'W':
	      q_val *= SECS_PER_WEEK;
	      break;
	    case 'D':
	      q_val *= SECS_PER_DAY;
	      break;
	    case 'h':
	      q_val *= SECS_PER_HOUR;
	      break;
	    case 'm':
	      q_val *= SECS_PER_MIN;
	      break;
	    case 's':
	      /* Already in seconds */
	      break;
	    default:
	      expecting_error(FALSE, "YMWDhms", FALSE, FALSE, ptr);
	      break;
	    }
	    q_relative = The_Time; /* Compare using relative time */
	    ptr++;
	  }
	  else if (ptr-start < 6)
	    expecting_error(FALSE, "YMWDhms' or 'ISO8601-date",
			    FALSE, FALSE, ptr-1);
	  else if (ptr-start != 8)
	    /* For purposes of meaningful error messages, assume that 6, 7,
	       or 9+ digits was intended to be an ISO8601-date: */
	    filter_error_msg("Date not in YYYYMMDD[.hhmmss] form", ptr-1);
	  else
	  {
	    /* Re-parse ISO8601-style dates: */
	    ptr = start;
	    if (q_op == Q_EQ || q_op == Q_NE)
	    {
	      q_val = convert_iso8601(Current_Arg, &ptr, FALSE);
	      ptr = start;
	      q_val2 = convert_iso8601(Current_Arg, &ptr, TRUE);
	    }
	    else if (q_op == Q_GT || q_op == Q_LE)
	      q_val = convert_iso8601(Current_Arg, &ptr, TRUE);
	    else
	      q_val = convert_iso8601(Current_Arg, &ptr, FALSE);
	    q_relative = 0; /* Compare using absolute time */
	  }
	}

	phStore(start, ptr, q_val, q_relative, q_val2);
      }

      if (q_relative != 0)
	q_info = q_relative - q_info;

      switch (q_op)
      {
      case Q_EQ:
	if (q_val2 == 0)
	  test_result = q_info == q_val;
	else
	  test_result = q_info >= q_val && q_info <= q_val2;
	break;
      case Q_NE:
	if (q_val2 == 0)
	  test_result = q_info != q_val;
	else
	  test_result = q_info < q_val || q_info > q_val2;
	break;
      case Q_GT:
	test_result = q_info > q_val;
	break;
      case Q_GE:
	test_result = q_info >= q_val;
	break;
      case Q_LT:
	test_result = q_info < q_val;
	break;
      case Q_LE:
	test_result = q_info <= q_val;
	break;
      default:
	/* Can't happen!? */
	break;
      }

      term_good &= negate ? !test_result : test_result;

      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case fexpr_END:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	expecting_error(FALSE, NULL, TRUE, TRUE, ptr);
	break;
      }
      ptr++;
    }

    file_is_listable |= term_good;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


Boole filter_unusual(Dir_Item *file, char *path, char *ptr, char **end)
{
  /***************************************/
  /* Check file against "U"nusual filter */
  /***************************************/
  
  Boole file_is_listable = FALSE;
  Boole looking_expr;
  
  looking_expr = TRUE;
  while (looking_expr && (!file_is_listable || end))
  {
    Boole term_good = TRUE;
    Boole looking_term;

    looking_term = TRUE;
    while (looking_term)
    {
      Boole negate = FALSE;
      Boole test_result = FALSE;
      
      if (*ptr == fexpr_FNOT || *ptr == fexpr_NOT)
      {
	negate = TRUE;
	ptr++;
      }

      /* Maybe list GID != parent GID,
	 symbolic links to files that don't exist */

      switch (*ptr)
      {
      case 't':
	test_result = filter_fexpr(file, path,
				   "~T{r:d:l}", NULL);
	break;

      case 'p':
	test_result = filter_fexpr(file, path, "\
P{o+w:+t},T{~l}|P{+s,+x:+l},~T{d:l}", NULL);
	break;

      case 'P':
	{
	  Uint u = (file->info.st_mode >> 6) & 007;
	  Uint g = (file->info.st_mode >> 3) & 007;
	  Uint o = (file->info.st_mode     ) & 007;
	  test_result = o > g || o > u || g > u;
	}
	break;

      case 'A':
	test_result = get_acl_count(file, path) > 0;
	break;

      case 'l':
	test_result = filter_fexpr(file, path,
				   "Q{l>1}&T{~d}", NULL);
	break;

      case 'u':
	/* Look for NFS nobody/noaccess UIDs and/or
	   UIDs not listed in /etc/passwd: */
	test_result =
	  file->info.st_uid == 60001 ||
	    file->info.st_uid == 60002 ||
	      file->info.st_uid == 65534 ||
		uid2name(file->info.st_uid) == NULL;
	break;

      case 'g':
	/* Look for NFS nobody/noaccess GIDs and/or
	   GIDs not listed in /etc/group: */
	test_result =
	  file->info.st_gid == 60001 ||
	    file->info.st_gid == 60002 ||
	      file->info.st_gid == 65534 ||
		gid2name(file->info.st_gid) == NULL;
	break;

      case 'm':
      case 'a':
      case 'c':
	/* Look for future dates: */
	{
	  Ulong info = (Ulong)(*ptr == 'm' ? file->info.st_mtime :
			       *ptr == 'a' ? file->info.st_atime :
			       file->info.st_ctime);
	  test_result = info > (Ulong)The_Time_in_an_hour || info == 0;
	}
	break;

      case 'n':
	test_result = quote_str(file->fname, NULL, QS_FILE) != NULL;
	break;

      case 'N':
	if (using_full_names)
	  test_result = quote_str(path, NULL, QS_FILE) != NULL ||
	    wildcard_match(path, "*/-*");
	else
	  test_result = quote_str(file->fname, NULL, QS_FILE) != NULL;
	break;

      case 'L':
	/* ??? Should the 'L' result get quote tested ??? */
	test_result = FALSE;
# if defined(S_IFLNK)
	/* System has symbolic links: */
	if (file->islink)
	{
	  if (expand_symlink)
	    test_result = file->stat_errno != 0;
	  else
	  {
	    Void sigSafe_stat(path, &file->info);
	    test_result = errno != 0;
	  }
	}
# endif
	break;

      case 's':
	/* Look for sparse files (i.e. partially filled w/data): */
# ifdef NO_STAT_BLOCKS
	filter_warning_msg("Filtering for sparse files unsupported", ptr);
# else
	/* One strange case to beware of is regular files located on a NetApp
	   file server whose size is <= 64 bytes (these peculiar files occupy
	   0 blocks!); thus, as a precaution avoid consideration of any file
	   smaller than 512 bytes. */
	if (file->stat_errno != 0 || /* FIXME???: too forgiving? */
	    (file->info.st_mode & S_IFMT) != S_IFREG ||
	    file->info.st_size <= 512)
	{
	  test_result = FALSE;
	}
	else
	{
	  int blksize = getCwdBlocksize(path);
	  BLKCNT_T blkbytes = file->info.st_blocks * blksize;

	  if (blksize == 1024)
	  {
	    /* Eliminate non-sparse files from possible HPUX file server: */
	    /* CONSIDER: What would happen if non-HPUX file server with file
	       size > 2^31 were encountered on a non-LARGEFILE system?.
	       blkbytes would overflow, but that wouldn't matter as EFBIG
	       error should occur first!?  Just to be safe make sure
	       LARGEFILES are enabled (i.e. sizeof(BLKCNT_T) > 4) or fewer
	       than 0x200000 blocks exist before testing for HPUX sparse
	       files. */
	    if (sizeof(BLKCNT_T) == 4 && file->info.st_blocks >= 0x200000)
	      test_result = FALSE;
	    else
	      test_result = file->info.st_size > blkbytes;
	  }
	  else
	  {
	    /* Detect sparse files from non-HPUX file servers: */
	    test_result = file->info.st_size > blkbytes;
	  }
	}
# endif
	break;

      case 'G':
	/* General tests: */
	test_result = filter_fexpr(file, path,
				   "U{t:p:P:A:l:u:g:m:a:c:N:L:s}", NULL);
	/* Tests 'n' and 'S' were deliberately excluded from 'G'
	   as these are subsets of other tests. */
	break;

      case 'S':
	/* Security tests: */
	test_result = filter_fexpr(file, path, "\
~T{d:l},P{o+w}|T{d},P{o+w,-t}|T{r},P{+s,+x}|T{b:c}", NULL);
	break;

      default:
	expecting_error(!negate, "tpPAlugmacnNGS", FALSE, FALSE, ptr);
	break;
      }
      ptr++;
      
      term_good &= negate ? !test_result : test_result;
      
      switch (*ptr)
      {
      case fexpr_FAND:
      case fexpr_AND:
	break;
      case fexpr_FOR:
      case fexpr_OR:
	looking_term = FALSE;
	break;
      case fexpr_END:
	looking_term = FALSE;
	looking_expr = FALSE;
	break;
      default:
	expecting_error(FALSE, NULL, TRUE, TRUE, ptr);
	break;
      }
      ptr++;
    }

    file_is_listable |= term_good;
  }

  if (end) *end = ptr;
  return(file_is_listable);
}


void list_file(Dir_Item *file,
	       char *dname)
{
  /* Perform G_print only if no errors were detected while getting the file
     status and if the file is of interest (i.e. it needs to satisfy our
     listing criteria): */
  if (!file->stat_errno && file->listable)
  {
    char *bp = G_print(output_buff, G_format, dname, file);
    strcat(bp, "\n");
  }

  if (file->stat_errno)
  {
    /* An error was detected while getting the file status or during G_print.
       Report the error only if the file is of interest (i.e. it needs to
       satisfy either our listing or search criteria): */
    if (file->listable || file->searchable)
    {
      char msg[MAX_MSG];
      int save_errno = errno;
      sprintf(msg, "%s: %s", Progname, full_name(dname, file->fname));
      errno = file->stat_errno;
      if (!warning_suppress) perror(msg);
      errno = save_errno;
      listingError = TRUE;
    }
  }
  else
  {
    /* No errors were detected while getting the file status or during
       G_print.  List the file only if it is of interest (i.e. it needs to
       satisfy our listing criteria): */
    if (file->listable)
    {
      if (execute_mode)
      {
	Uint sts;
	if (file->islink && !execute_symlinks)
	{
	  Void fprintf(stderr, "%s: %s: +X ignoring symbolic link\n",
		       Progname, full_name(dname, file->fname));
	  Void fprintf(stderr, "\
  (NOTE: +h flag must be specified for +X to act on symbolic links)\n");
/*(+X requires that +h flag be specified before acting on symbolic links)*/
	  sts = 1;
	}
	else
	{
	  if (verbose_mode) Void fputs(output_buff, stdout);
	  sts = system(output_buff);
	}
	if (execute_sts_check)
	{
	  if ((sts & 0xff) == 0) sts >>= 8;
	  if (sts != execute_sts_good) exit(sts);
	}
      }
      else
	Void fputs(output_buff, stdout);
      output_nlines++;
    }
  }
  
  return;
}


char *find_first_directive(char *list, char *fmt)
{
  if (fmt != NULL)
  {
    char *cp = fmt;
    while (*cp != CNULL)
    {
      /* Look for "'%...'" directive inside string: */
      if (*cp == '"' || *cp == '\'')
      {
	char delimiter = *cp++;
	while (*cp != CNULL && *cp != delimiter) 
	{
	  /* Skip a quoted character inside string: */
	  if (*cp == '\\')
	    cp++;
	  /* Look for "'%...'" directive inside string (ignore "%%"): */
	  else if (*cp == '%')
	  {
	    cp++;
	    if (*cp != '%')
	    {
	      while (IS_MEMBER(*cp, "0_+-")) cp++;
	      while (isdigit(*cp)) cp++;	/* Scan past width (if any) */
	      if (IS_MEMBER(*cp, list))
		return(cp);
	    }
	  }
	  if (*cp != CNULL) cp++;
	}
      }
      /* Skip a quoted character outside of string: */
      else if (*cp == '\\')
	cp++;
      /* Look for directive outside of string: */
      else if (IS_MEMBER(*cp, list))
	return(cp);
      if (*cp != CNULL) cp++;
    }
  }
  return(NULL);
}


#define ZERO_PAD_DEFAULT  FALSE
char *G_print(char *buff,
	      char *fmt,
	      char *dname,
	      Dir_Item *file)
{
  register char *bp;
  char icase;
  char delimiter = CNULL;
  Boole quote_mode = FALSE;
  Boole percent_specified = FALSE;
  Boole zero_pad;
  Boole hard_width = FALSE;
  Boole width_specified = FALSE;
  int width = 0;
  Boole directive_seen = FALSE;
  Boole modifier = FALSE;
  Boole ttoggle = FALSE;
  register struct stat *info = &file->info;
  int fmode = file->info.st_mode;
  int ftype = (fmode & S_IFMT);

  bp = buff;
  *bp = CNULL;

  zero_pad = ZERO_PAD_DEFAULT;
  squeeze_cnt = 0;
  squeeze = FALSE;
  as_is = FALSE;
  separated = FALSE;

  while ((icase = *fmt++) != CNULL)
  {
    if (modifier)
    {
      /* Modifiers generate no output: */
      switch (icase)
      {
      default:
	/* G_format currently has no modifiers(!): */
	carrot_msg(NULL, "+G", G_format, "Unrecognized modifier", fmt-1);
	exit(USAGE_ERROR);
	break;
      }
      zero_pad = ZERO_PAD_DEFAULT;
      modifier = FALSE; /* reset for next time */
    }
    else
    {
      switch (icase)
      {
      case '"':
      case '\'':
	if (!quote_mode)
	{
	  quote_mode = TRUE;
	  delimiter = icase;
	}
	else if (delimiter == icase)
	{
	  quote_mode = FALSE;
	}
	break;

      case '%':
	if (*fmt == '%')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  Boole minus = FALSE;
	  percent_specified = TRUE;
	  if (*fmt == '0') {zero_pad = TRUE; fmt++;}
	  if (*fmt == '_') {zero_pad = FALSE; fmt++;}
	  if (*fmt == '+') {hard_width = TRUE; fmt++;}
	  if (*fmt == '-') {minus = TRUE; fmt++;}
	  width_specified = (isdigit(*fmt) != 0);
	  if (width_specified)
	  {
	    width = (*fmt++ - '0');
	    while (isdigit(*fmt))
	      width = width * 10 + (*fmt++ - '0');
	    if (width < 0 || width > F_MAX_WIDTH)
	    {
	      /* NB: width < 0 implies overflow! */
	      carrot_msg(NULL, "+G", G_format, "Too big", fmt-1);
	      exit(USAGE_ERROR);
	    }
	    if (minus)
	    {
	      width = -width;
	      /* Make "%0*d",-9 behave like "%0-9d" for negative widths: */
	      zero_pad = FALSE;
	    }
	  }
	  else if (minus)
	  {
	    /* Make "%-d" mimic Linux's date +format: */
	    width = 0;
	    width_specified = TRUE;
	  }
	}
	break;

      case '^':
	if (*fmt == '^')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  modifier = TRUE;
	}
	break;

      case '~':
	ttoggle = !ttoggle;
	if (ttoggle)
	  squeeze_cnt++;
	else
	  squeeze_cnt--;
	squeeze = (squeeze_cnt != 0);
	if (!squeeze) bp = separate(bp, icase);
	break;

      case '\\':
	/* Print any character following the \quote as-is: */
	as_is = TRUE;
	if (*fmt != CNULL)
	{
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	break;

      default:
	/* A character within quotes is a directive only when preceded by %: */
	if (quote_mode && !percent_specified)
	{
	  as_is = TRUE;
	  bp = separate(bp, icase);
	  break;
	}

	/* Remove extraneous spacing if appropriate: */
	if ((fielding || (squeeze && !hard_width)) &&
	    (!zero_pad || width != 0))
	{
	  width = 0;
	  width_specified = TRUE;
	}

	/* Process a directive: */
	switch (icase)
	{
	case TYPE_IN_ALPHA:
	  {
	    char type;
	    if      (ftype == S_IFREG)  /* Regular */
	      type = symtype_REG; /* sic! */
	    else if (ftype == S_IFDIR)  /* Directory */
	      type = type_DIR;
	    else if (ftype == S_IFCHR)  /* Char Special */
	      type = type_CHR;
	    else if (ftype == S_IFBLK)  /* Block Special */
	      type = type_BLK;
#ifdef S_IFIFO
	    else if (ftype == S_IFIFO)  /* Fifo */
	      type = type_FIFO;
#endif
#ifdef S_IFLNK
	    else if (ftype == S_IFLNK)  /* Symbolic Link */
	      type = type_LNK;
#endif
#ifdef S_IFSOCK
	    else if (ftype == S_IFSOCK) /* Socket */
	      type = type_SOCK;
#endif
#ifdef S_IFDOOR
	    else if (ftype == S_IFDOOR) /* Door */
	      type = type_DOOR;
#endif
	    else
	      type = '?';

	    if (!width_specified) width = 1;
	    sprintf(bp, "%*c",width, type);
	  }
	  break;

	  /* Add various flavorings: */
#ifdef SUNOS
#  if SUNOS >= 5
#    define  HAVE_MANDATORY_LOCKING
	  /*   Actually, only SunOS5.5 and above uses a fifo mark but it shouldn't
	       hurt for all versions of SunOS5 to use it: */
#    define  FIFO_MARK  symtype_FIFO
#  endif
#endif
#ifdef HPUX
#  define  FIFO_MARK  symtype_FIFO
#endif

#ifdef HAVE_MANDATORY_LOCKING
#  define  S_OR_l     'l'
#else
#  define  S_OR_l     'S'
#endif

#ifndef FIFO_MARK
#  define  FIFO_MARK  ' '
#endif

	case TYPE_IN_SYMBOLIC:
	  {
	    char type;
	    if      (ftype == S_IFREG)  /* Regular */
	      type = (fmode & (S_IXUSR|S_IXGRP|S_IXOTH) ? '*' : ' ');
	    else if (ftype == S_IFDIR)  /* Directory */
	      type = symtype_DIR;
	    else if (ftype == S_IFCHR)  /* Char Special */
	      type = ' ';
	    else if (ftype == S_IFBLK)  /* Block Special */
	      type = ' ';
#ifdef S_IFIFO
	    else if (ftype == S_IFIFO)  /* Fifo */
	      type = FIFO_MARK;
#endif
#ifdef S_IFLNK
	    else if (ftype == S_IFLNK)  /* Symbolic Link */
	      type = symtype_LNK;
#endif
#ifdef S_IFSOCK
	    else if (ftype == S_IFSOCK) /* Socket */
	      type = symtype_SOCK;
#endif
#ifdef S_IFDOOR
	    else if (ftype == S_IFDOOR) /* Door */
	      type = symtype_DOOR;
#endif
	    else
	      type = '?';

	    if (!width_specified) width = 1;
	    sprintf(bp, "%*c",width, type);
	  }
	  break;

#ifdef TYPE_IN_NUMERIC
	case TYPE_IN_NUMERIC:
	  /* This is probably non-portable to some of the more obscure Unixes: */
	  if (!width_specified) width = 2;
	  sprintf(bp, "%*d",width, ftype >> 12);
	  break;
#endif /*TYPE_IN_NUMERIC*/

	case PERMISSION_IN_ALPHA:
	  {
	    char perm[16], *pp;
	    pp = perm;
	    if (fmode & S_IRUSR)
	      *pp++ = 'r';
	    else
	      *pp++ = '-';

	    if (fmode & S_IWUSR)
	      *pp++ = 'w';
	    else
	      *pp++ = '-';

	    if ((fmode & S_IXUSR) && (fmode & S_ISUID))
	      *pp++ = 's';
	    else if (fmode & S_ISUID)
	      *pp++ = 'S';
	    else if (fmode & S_IXUSR)
	      *pp++ = 'x';
	    else
	      *pp++ = '-';

	    if (fmode & S_IRGRP)
	      *pp++ = 'r';
	    else
	      *pp++ = '-';

	    if (fmode & S_IWGRP)
	      *pp++ = 'w';
	    else
	      *pp++ = '-';

	    if ((fmode & S_IXGRP) && (fmode & S_ISGID))
	      *pp++ = 's';
	    else if (fmode & S_ISGID)
	      *pp++ = S_OR_l;
	    else if (fmode & S_IXGRP)
	      *pp++ = 'x';
	    else
	      *pp++ = '-';

	    if (fmode & S_IROTH)
	      *pp++ = 'r';
	    else
	      *pp++ = '-';

	    if (fmode & S_IWOTH)
	      *pp++ = 'w';
	    else
	      *pp++ = '-';

	    if ((fmode & S_IXOTH) && (fmode & S_ISVTX))
	      *pp++ = 't';
	    else if (fmode & S_ISVTX)
	      *pp++ = 'T';
	    else if (fmode & S_IXOTH)
	      *pp++ = 'x';
	    else
	      *pp++ = '-';

	    *pp = CNULL;

	    if (!width_specified) width = 9;
	    sprintf(bp, "%*s",width, perm);
	  }
	  break;

	case PERMISSION_IN_NUMERIC:
	  if (!width_specified || width < 4) width = 4;
	  sprintf(bp, "%0*o",width, (fmode & 07777));
	  break;

	case PERMISSION_IN_SYMBOLIC:
	  {
	    char perm[32], *pp;
	    pp = perm;
	    *pp++ = 'u'; *pp++ = '=';
	    if (fmode & S_IRUSR) *pp++ = 'r';
	    if (fmode & S_IWUSR) *pp++ = 'w';
	    if (fmode & S_IXUSR) *pp++ = 'x';
	    *pp++ = ','; *pp++ = 'g'; *pp++ = '=';
	    if (fmode & S_IRGRP) *pp++ = 'r';
	    if (fmode & S_IWGRP) *pp++ = 'w';
	    if (fmode & S_IXGRP) *pp++ = 'x';
	    *pp++ = ','; *pp++ = 'o'; *pp++ = '=';
	    if (fmode & S_IROTH) *pp++ = 'r';
	    if (fmode & S_IWOTH) *pp++ = 'w';
	    if (fmode & S_IXOTH) *pp++ = 'x';

	    {
#ifdef HAVE_MANDATORY_LOCKING
	      Boole lock = (fmode & S_ISGID) &&
		!(fmode & S_IXGRP);
	      Boole setgid = (fmode & S_ISGID) &&
		(fmode & S_IXGRP);
#else
	      Boole lock = FALSE;
	      Boole setgid = (fmode & S_ISGID) != 0;
#endif
	      Boole setuid = (fmode & S_ISUID) != 0;
	      Boole sticky = (fmode & S_ISVTX) != 0;

	      if (setuid || setgid)
	      {
		*pp++ = ',';
		if (setuid)  *pp++ = 'u';
		if (setgid)  *pp++ = 'g';
		*pp++ = '+'; *pp++ = 's';
	      }

	      if (file->isdir && !setgid)
	      {
		*pp++ = ','; *pp++ = 'g'; *pp++ = '-'; *pp++ = 's';
	      }

	      if (lock || sticky)
	      {
		*pp++ = ','; *pp++ = '+';
		if (lock)    *pp++ = 'l';
		if (sticky)  *pp++ = 't';
	      }
	    }

	    *pp = CNULL;

	    /* Worst cases: u=rwxs,g=rwxs,o=rwx,+t    (files) */
	    /*              u=rwx,g=rwx,o=rwx,ug+s,+t (all) */
	    /*              u=rwx,g=rw,o=rwx,u+s,+lt  (sunos5 files) */
	    /*              u=rwx,g=rwx,o=rwx,u+s,g-s,+t (dir without g+s) */
	    /*              1234567890123456789012345 */
	    if (!width_specified) width = -25;
	    sprintf(bp, "%*s",width, perm);
	  }
	  break;

	case ACL_INDICATOR:
	  if (!width_specified) width = 1;
	  sprintf(bp, "%*s",width,
		  get_acl_count(file, full_name(dname, file->fname)) > 0 ?
		  "+" : "");
	  break;

	case INODE:
	  /* SYS5-R2 == 5
	     SunOS4.1.3 == 6, SunOS5.0-5.5.1 == 5(!), SunOS5.6-5.8 == 10
	     HPUX10.20 == 6
	     Linux2.2 == 7 */
	  /* Most modern UNIXes assume inodes take at least 6 places: */
	  if (!width_specified) width = 6;
	  sprintf(bp, F_st_ino(zero_pad,width), info->st_ino);
	  break;

	case SIZE_IN_BYTES:
	  if (!width_specified) width = (ANYBIT(Sem,SEM_BSD|SEM_GNU|SEM_ELS) ? 8 : 7);
	  if (file->isdev)
	  {
	    char major_dev[16], minor_dev[16], str[32];
	    sprintf(major_dev, "%lu", (Ulong)major(info->st_rdev));
	    sprintf(minor_dev, "%lu", (Ulong)minor(info->st_rdev));
	    /* Leading spaces are added to the major field for device numbers
	       less than 100 (unless squeeze or width specified) so as to mimic
	       /bin/ls when the minor field exceeds 999 (or 9999 for BSD) thus
	       keeping it from pushing its way into the major field.  If true
	       mimicing of BSD's or SYS5's /bin/ls is desired when listing
	       devices, then specify either +4l or +5l flags respectively. */
	    sprintf(str, "%*s,%*s",
		    (squeeze || width_specified ? 0 : 3),
		    major_dev,
		    (squeeze ? 0 : (ANYBIT(Sem,SEM_BSD|SEM_GNU|SEM_ELS) ? 4 : 3)),
		    minor_dev);
	    sprintf(bp, "%*s",width, str);
	  }
	  else
	    sprintf(bp, F_st_size(zero_pad,width), info->st_size);
	  break;

	case SIZE_IN_BLOCKS:
	  if (!width_specified) width = 4;
	  sprintf(bp, F_st_blocks(zero_pad,width),
		  ANYBIT(Sem,SEM_BSD|SEM_GNU|SEM_ELS) ?
		  get_stat_blocks(&file->info)/2 :
		  get_stat_blocks(&file->info));
	  break;

	case LINK_COUNT:
	  /* ELS version > 1.45 always uses 3 for link count and reserves
	     1 for possible ACL unless LS and BSD (i.e. +l4) specified.
	     ELS version <= 1.45 uses 2 for link count except where support
	     for ACLs exists or LS and not BSD (i.e. +l) specified. */
	  if (!width_specified) width = (ALLBITS(Sem,SEM_BSD|SEM_LS) ? 2 : 3);
# if !defined(HAVE_ACL)
	  if (VersionLevel <= 145 && !width_specified)
	    width = (ANYBIT(Sem,SEM_BSD|SEM_ELS) ? 2 : 3);
# endif
	  sprintf(bp, F_st_nlink(zero_pad,width), info->st_nlink);
	  break;

	case TIME_MODIFIED:
	case TIME_ACCESSED:
	case TIME_MODE_CHANGED:
	  bp = T_print_width(bp, T_format,
			     icase == TIME_MODIFIED ? info->st_mtime :
			     icase == TIME_ACCESSED ? info->st_atime :
			     /*icase == TIME_MODE_CHANGED ?*/ info->st_ctime,
			     NULL, FALSE, FALSE, width);
	  break;

	case UID_IN_ALPHA:
	  {
	    char *pw_name = uid2name(info->st_uid);
	    int trunc = MAX_USER_GROUP_NAME;
	    if (!width_specified) width = -8;
	    if (width_specified || TruncateId)
	      trunc = (width > 0 ? width : -width);
	    if (pw_name != NULL)
	      sprintf(bp, "%*.*s",width,trunc, pw_name);
	    else
	      /* No name, print the number instead: */
	      sprintf(bp, "%*lu",width, (Ulong)info->st_uid);
	  }
	  break;

	case GID_IN_ALPHA:
	  {
	    char *gr_name = gid2name(info->st_gid);
	    int trunc = MAX_USER_GROUP_NAME;
	    if (!width_specified) width = -8;
	    if (width_specified || TruncateId)
	      trunc = (width > 0 ? width : -width);
	    if (gr_name != NULL)
	      sprintf(bp, "%*.*s",width,trunc, gr_name);
	    else
	      /* No name, print the number instead: */
	      sprintf(bp, "%*lu",width, (Ulong)info->st_gid);
	  }
	  break;

	case UID_IN_NUMERIC:
	  if (!width_specified) width = 5;
	  sprintf(bp, "%*lu",width, (Ulong)info->st_uid);
	  break;

	case GID_IN_NUMERIC:
	  if (!width_specified) width = 5;
	  sprintf(bp, "%*lu",width, (Ulong)info->st_gid);
	  break;

	case _16_BIT_CHECKSUM:
	  /* Perform checksum on regular files and preserve the access times: */
	  {
	    Ulong cksum = 0;
	    if (ftype == S_IFREG)
	    {
	      int save_errno = errno;
	      char *path = full_name(dname, file->fname);
	      struct utimbuf tb;
	      /* Perform checksum and record any errno generated by checksum: */
	      errno = 0;
	      cksum = checksum(path);
	      if (errno != 0)
		file->stat_errno = errno;
	      else
	      {
		/* Attempt to restore access times: */
		tb.actime = info->st_atime;
		tb.modtime = info->st_mtime;
		Void utime(path, &tb);
		/* ??? if (verbose_mode && errno != 0) then warn about access
		   times being clobbered ??? */
	      }
	      errno = save_errno;
	    }
	    if (!width_specified || width < 5) width = 5;
	    sprintf(bp, "%0*lu",width, cksum);
	  }
	  break;

	case NAME_FORMAT:
	case FULL_NAME_FORMAT:
	case FULL_NAME:
	case DIR_NAME:
	case FILE_NAME:
	  {
	    char *N_format_to_use;
	    char N_format_case[2];
	    if (icase == NAME_FORMAT || icase == FULL_NAME_FORMAT)
	    {
	      N_format_to_use = N_format;
	    }
	    else
	    {
	      N_format_case[0] = icase;
	      N_format_case[1] = CNULL;
	      N_format_to_use = N_format_case;
	    }

	    if (!width_specified)
	      bp = N_print(bp, N_format_to_use, dname, file);
	    else
	    {
	      char tmp[MAX_FULL_NAME];
	      Void N_print(tmp, N_format_to_use, dname, file);
	      sprintf(bp, "%*s",width, tmp);
	    }
	  }
	  break;

	case LINK_NAME:
#     if defined(S_IFLNK)
	  if (file->islink && !expand_symlink)
	  {
	    char *N_format_to_use;
	    char N_format_case[2];

	    N_format_case[0] = icase;
	    N_format_case[1] = CNULL;
	    N_format_to_use = N_format_case;

	    if (!width_specified)
	      bp = N_print(bp, N_format_to_use, dname, file);
	    else
	    {
	      char tmp[MAX_FULL_NAME];
	      Void N_print(tmp, N_format_to_use, dname, file);
	      sprintf(bp, "%*s",width, tmp);
	    }
	  }
#     endif
	  break;

	default:
	  if (isalnum(icase))
	  {
	    carrot_msg(NULL, "+G", G_format, "Unrecognized directive", fmt-1);
	    exit(USAGE_ERROR);
	  }
	  else
	    /* Output non-directive/non-alphanumeric as-is: */
	    as_is = TRUE;
	  break;
	}

	percent_specified = FALSE;
	zero_pad = ZERO_PAD_DEFAULT;
	hard_width = FALSE;
	width_specified = FALSE;
	width = 0;
	if (!as_is) directive_seen = TRUE;
	bp = separate(bp, icase);
	break;
      }
    }
  }
  bp = finish(bp);

  /* Remove trailing spaces caused by marking files and/or symbolic
     links.  This is caused by: 1) the -F option when marking a regular
     file 2) the +GL option when listing a non-symbolically linked
     file.  Regrettably, this removes legitimate trailing spaces from
     the end of file names, etc.: */
  while (bp != buff && isspace(bp[-1])) bp--;
  *bp = CNULL;

  if (quote_mode) error_msg("Missing delimiter in +G format");
  if (!directive_seen) error_msg("Incomplete +G format");

  return(bp);
}
#undef  ZERO_PAD_DEFAULT


#define ZERO_PAD_DEFAULT  FALSE
char *N_print(char *buff, char *fmt,
	      char *dname, Dir_Item *file)
{
  register char *bp;
  char icase;
  char delimiter = CNULL;
  Boole quote_mode = FALSE;
  Boole percent_specified = FALSE;
  Boole zero_pad;
  Boole hard_width = FALSE;
  Boole width_specified = FALSE;
  int width = 0;
  Boole directive_seen = FALSE;
  Boole modifier = FALSE;
  Boole ttoggle = FALSE;
  Boole use_quotes = plus_q;
  char *fname = file->fname;	/* Use "fname" to avoid side-effects! */
  char *slash = NULL;
  Boole markable = FALSE;

  /* NB: You must reference "fname" instead of "file->fname" for the
     duration of this procedure to avoid probable side-effects! */

  /*****************************************************/
  /* Put dir name and/or file name into canonical form */
  /*****************************************************/

  if (using_full_names)
  {
    if (strcmp(dname, "/") == 0)
    {
      /* Do nothing */;
    }
    else if ((slash = strrchr(fname, '/')) != NULL)
    {
      if (slash == fname)
	dname = "/";
      else
	dname = fname;
      *slash = CNULL;
      fname = slash+1;
    }
    else
    {
      int last_dir_char = strlen(dname) - 1;
      if (last_dir_char > 0 && dname[last_dir_char] == '/')
      {
	slash = &dname[last_dir_char];
	*slash = CNULL;
      }
    }
  }

  bp = buff;
  *bp = CNULL;

  zero_pad = ZERO_PAD_DEFAULT;
  separated = FALSE;

  while ((icase = *fmt++) != CNULL)
  {
    if (modifier)
    {
      /* Modifiers generate no output: */
      switch (icase)
      {
      case 'Q':		/* '+N^Q' is DEPRECATED, use '+N^q' instead */
      case QUOTE_NAME:
	use_quotes = TRUE;
	break;

      default:
	carrot_msg(NULL, "+N", N_format, "Unrecognized modifier", fmt-1);
	exit(USAGE_ERROR);
	break;
      }
      zero_pad = ZERO_PAD_DEFAULT;
      modifier = FALSE; /* reset for next time */
    }
    else
    {
      switch (icase)
      {
      case '"':
      case '\'':
	if (!quote_mode)
	{
	  quote_mode = TRUE;
	  delimiter = icase;
	}
	else if (delimiter == icase)
	{
	  quote_mode = FALSE;
	}
	break;

      case '%':
	if (*fmt == '%')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  Boole minus = FALSE;
	  percent_specified = TRUE;
	  if (*fmt == '0') {zero_pad = TRUE; fmt++;}
	  if (*fmt == '_') {zero_pad = FALSE; fmt++;}
	  if (*fmt == '+') {hard_width = TRUE; fmt++;}
	  if (*fmt == '-') {minus = TRUE; fmt++;}
	  width_specified = (isdigit(*fmt) != 0);
	  if (width_specified)
	  {
	    width = (*fmt++ - '0');
	    while (isdigit(*fmt))
	      width = width * 10 + (*fmt++ - '0');
	    if (width < 0 || width > F_MAX_WIDTH)
	    {
	      /* NB: width < 0 implies overflow! */
	      carrot_msg(NULL, "+N", N_format, "Too big", fmt-1);
	      exit(USAGE_ERROR);
	    }
	    if (minus)
	    {
	      width = -width;
	      /* Make "%0*d",-9 behave like "%0-9d" for negative widths: */
	      zero_pad = FALSE;
	    }
	  }
	  else if (minus)
	  {
	    /* Make "%-d" mimic Linux's date +format: */
	    width = 0;
	    width_specified = TRUE;
	  }
	}
	break;

      case '^':
	if (*fmt == '^')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  modifier = TRUE;
	}
	break;

      case '~':
	ttoggle = !ttoggle;
	if (ttoggle)
	  squeeze_cnt++;
	else
	  squeeze_cnt--;
	squeeze = (squeeze_cnt != 0);
	if (!squeeze) bp = separate(bp, icase);
	break;

      case '\\':
	/* Print any character following the \quote as-is: */
	as_is = TRUE;
	if (*fmt != CNULL)
	{
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	break;

      case 'Q':		/* '+NQ' is DEPRECATED, use '+N^q' instead */
      case QUOTE_NAME:	/* '+Nq' is DEPRECATED, use '+N^q' instead */
	if (quote_mode)
	{
	  /* Don't confuse a quoted alpha character for a legacy modifier
	     (this bug exists in version 145 and earlier): */
	  as_is = TRUE;
	  bp = separate(bp, icase);
	  break;
	}
	else
	{
	  /* Handle legacy modifiers before "^" was introduced by treating
	     these as if they were prefaced by "^" by re-entering while-loop
	     with "modifier" flag set TRUE and with fmt reset to original
	     character: */
	  modifier = TRUE;
	  fmt--;
	}
	break;	/* modifiers generate no output */

      default:
	/* A character within quotes is a directive only when preceded by %: */
	if (quote_mode && !percent_specified)
	{
	  as_is = TRUE;
	  bp = separate(bp, icase);
	  break;
	}

	/* Remove extraneous spacing if appropriate: */
	if ((fielding || (squeeze && !hard_width)) &&
	    (!zero_pad || width != 0))
	{
	  width = 0;
	  width_specified = TRUE;
	}

	/* Process a directive: */
	switch (icase)
	{
	case FULL_NAME:
	  {
	    char *d, *s;
	    if (strcmp(dname, "/") == 0)
	    {
	      d = ""; s = "/";
	    }
	    else if (dname[0] != CNULL && strcmp(dname, ".") != 0)
	    {
	      d = dname; s = "/";
	    }
	    else
	    {
	      d = ""; s = "";
	    }

	    if (!width_specified)
	    {
	      sprintf(bp, "%s%s%s", d, s, fname);
	      if ((d = quote_fname(bp, dash_b, dash_Q, use_quotes)) != NULL)
		sprintf(bp, "%s", d);
	    }
	    else
	    {
	      char tmp[MAX_FULL_NAME];
	      sprintf(tmp, "%s%s%s", d, s, fname);
	      if ((d = quote_fname(tmp, dash_b, dash_Q, use_quotes)) != NULL)
		sprintf(bp, "%*s",width, d);
	      else
		sprintf(bp, "%*s",width, tmp);
	    }
	    markable = TRUE;
	  }
	  break;

	case DIR_NAME:
	  {
	    char *d;
	    if (dname[0] == CNULL) dname = ".";
	    if ((d = quote_fname(dname, dash_b, dash_Q, use_quotes)) != NULL)
	      sprintf(bp, "%*s",width, d);
	    else
	      sprintf(bp, "%*s",width, dname);
	    markable = TRUE;
	  }
	  break;

	case FILE_NAME:
	  {
	    char *f;
	    if ((f = quote_fname(fname, dash_b, dash_Q, use_quotes)) != NULL)
	      sprintf(bp, "%*s",width, f);
	    else
	      sprintf(bp, "%*s",width, fname);
	    markable = TRUE;
	  }
	  break;

	case LINK_NAME:
	case LINK_NAME2:
#     if defined(S_IFLNK)
	  if (file->islink && !expand_symlink)
	  {
	    int i;
	    char lname[MAX_FULL_NAME];
	    char *fmt;
	    i = readlink(full_name(dname, fname),
			 lname, MAX_FULL_NAME);
	    if (i < 0 || i >= MAX_FULL_NAME) i = 0;
	    lname[i] = CNULL;
	    if (fielding || icase == LINK_NAME2)
	      fmt = "%*s";
	    else
	      fmt = (squeeze ? "->%*s" : "-> %*s");
	    {
	      char *l;
	      if ((l = quote_fname(lname, dash_b, dash_Q, plus_q)) != NULL)
		sprintf(bp, fmt,width, l);
	      else
		sprintf(bp, fmt,width, lname);
	    }

	    if (mark_files || mark_dirs)
	    {
	      /* Mark file using info from actual file rather than link: */
	      Dir_Item tmp;
	      tmp = *file;
	      if (sigSafe_stat(full_name(dname, tmp.fname), &tmp.info) == 0)
	      {
		bp += strlen(bp); /* Find where we are */
		if (mark_files)
		  bp = G_print(bp, "T", dname, &tmp);
		else if (mark_dirs && S_ISDIR(tmp.info.st_mode))
		{
		  *bp++ = '/';
		  *bp = CNULL;
		}
	      }
	      markable = FALSE;
	    }
	  }
#     endif
	  break;

	default:
	  if (isalnum(icase))
	  {
	    carrot_msg(NULL, "+N", N_format, "Unrecognized directive", fmt-1);
	    exit(USAGE_ERROR);
	  }
	  else
	    /* Output non-directive/non-alphanumeric as-is: */
	    as_is = TRUE;
	  break;
	}
	
	if (markable)
	{
	  bp += strlen(bp); /* Find where we are */
	  if (mark_files)
	  {
	    /* Short listings (i.e. not long) always mark files,
	       long listings mark files unless a non-expanded symbolic link: */
	    if (!list_long || !(file->islink && !expand_symlink))
	      bp = G_print(bp, "T", dname, file);
	  }
	  else if (mark_dirs && file->isdir)
	  {
	    *bp++ = '/';
	    *bp = CNULL;
	  }
	  markable = FALSE;
	}
	
	percent_specified = FALSE;
	zero_pad = ZERO_PAD_DEFAULT;
	hard_width = FALSE;
	width_specified = FALSE;
	width = 0;
	if (!as_is) directive_seen = TRUE;
	bp = separate(bp, icase);
	break;
      }
    }
  }
  bp = finish(bp);

  if (slash != NULL)
    *slash = '/';

  if (quote_mode) error_msg("Missing delimiter in +N format");
  if (!directive_seen) error_msg("Incomplete +N format");

  return(bp);
}
#undef  ZERO_PAD_DEFAULT


#define ZERO_PAD_DEFAULT  (!tics)
char *T_print(char *buff, char *fmt, time_t ftime,
	      struct tm *fdate, Boole gmt, Boole meridian)
{
  register char *bp;
  char icase;
  char delimiter = CNULL;
  Boole quote_mode = FALSE;
  Boole percent_specified = FALSE;
  Boole zero_pad;
  Boole hard_width = FALSE;
  Boole width_specified = FALSE;
  int width = 0;
  Boole directive_seen = FALSE;
  Boole modifier = FALSE;
  Boole delta_time = FALSE;
  Boole future = FALSE;
  Boole tics = FALSE;
  Boole numeric = FALSE;
  Boole ttoggle = FALSE;
  time_t abs_ftime = ftime;	/* Avoid possible side-effects */
  static char *dow[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
  static char *moy[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
  char legacy_modifier = CNULL;
# define BOOLE_XOR(a,b) ((a) != (b))

  bp = buff;
  *bp = CNULL;

  zero_pad = ZERO_PAD_DEFAULT;
  separated = FALSE;

  while ((icase = *fmt++) != CNULL)
  {
    if (modifier)
    {
      /* Make legacy_modifier substitution if needed: */
      if (legacy_modifier != CNULL)
      {
	icase = legacy_modifier;
	legacy_modifier = CNULL;
      }
      /* Modifiers generate no output: */
      switch (icase)
      {
      case ABS_TIME:
	tics = TRUE;
	numeric = FALSE;
	ftime = abs_ftime;
	fdate = NULL; /* Recalculate fdate */
	break;

      case DELTA_TIME:
      case REL_TIME:
	tics = TRUE;
	numeric = FALSE;
	delta_time = (icase == DELTA_TIME);
	if ((Ulong)The_Time < (Ulong)abs_ftime)
	{
	  future = TRUE;
	  ftime = abs_ftime - The_Time;
	}
	else
	{
	  future = FALSE;
	  ftime = The_Time - abs_ftime;
	}
	fdate = NULL; /* Recalculate fdate */
	break;

      case ALPHA_DATE:
	tics = FALSE;
	numeric = FALSE;
	break;

      case NUM_DATE:
	tics = FALSE;
	numeric = TRUE;
	break;

      case GMT_DATE:
	gmt = TRUE;
	fdate = NULL; /* Recalculate fdate */
	break;

      case LOCAL_DATE:
	gmt = FALSE;
	fdate = NULL; /* Recalculate fdate */
	break;

      case MERIDIAN_TIME:
	meridian = TRUE;
	break;

      default:
	carrot_msg(NULL, "+T", T_format, "Unrecognized modifier", fmt-1);
	exit(USAGE_ERROR);
	break;
      }
      zero_pad = ZERO_PAD_DEFAULT;
      modifier = FALSE; /* reset for next time */
    }
    else
    {
      switch (icase)
      {
      case '"':
      case '\'':
	if (!quote_mode)
	{
	  quote_mode = TRUE;
	  delimiter = icase;
	}
	else if (delimiter == icase)
	{
	  quote_mode = FALSE;
	}
	break;

      case '%':
	if (*fmt == '%')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  Boole minus = FALSE;
	  percent_specified = TRUE;
	  if (*fmt == '0') {zero_pad = TRUE; fmt++;}
	  if (*fmt == '_') {zero_pad = FALSE; fmt++;}
	  if (*fmt == '+') {hard_width = TRUE; fmt++;}
	  if (*fmt == '-') {minus = TRUE; fmt++;}
	  width_specified = (isdigit(*fmt) != 0);
	  if (width_specified)
	  {
	    width = (*fmt++ - '0');
	    while (isdigit(*fmt))
	      width = width * 10 + (*fmt++ - '0');
	    if (width < 0 || width > F_MAX_WIDTH)
	    {
	      /* NB: width < 0 implies overflow! */
	      carrot_msg(NULL, "+T", T_format, "Too big", fmt-1);
	      exit(USAGE_ERROR);
	    }
	    if (minus)
	    {
	      width = -width;
	      /* Make "%0*d",-9 behave like "%0-9d" for negative widths: */
	      zero_pad = FALSE;
	    }
	  }
	  else if (minus)
	  {
	    /* Make "%-d" mimic Linux's date +format: */
	    width = 0;
	    width_specified = TRUE;
	  }
	}
	break;

      case '^':
	if (*fmt == '^')
	{
	  as_is = TRUE;
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	else
	{
	  modifier = TRUE;
	}
	break;

      case '~':
	ttoggle = !ttoggle;
	if (ttoggle)
	  squeeze_cnt++;
	else
	  squeeze_cnt--;
	squeeze = (squeeze_cnt != 0);
	if (!squeeze) bp = separate(bp, icase);
	break;

      case '\\':
	/* Print any character following the \quote as-is: */
	as_is = TRUE;
	if (*fmt != CNULL)
	{
	  icase = *fmt++;
	  bp = separate(bp, icase);
	}
	break;

      case 'A':		/* '+TA' is DEPRECATED, use '+T^a' instead */
      case 'R':		/* '+TR' is DEPRECATED, use '+T^r' instead */
      case NUM_DATE:	/* '+TN' is DEPRECATED, use '+T^N' instead */
      case GMT_DATE:	/* '+TG' is DEPRECATED, use '+T^G' instead */
      case LOCAL_DATE:	/* '+TL' is DEPRECATED, use '+T^L' instead */
	if (quote_mode)
	{
	  /* Don't confuse a quoted alpha character for a legacy modifier
	     (this bug exists in version 145 and earlier): */
	  as_is = TRUE;
	  bp = separate(bp, icase);
	  break;
	}
	else
	{
	  /* Handle legacy modifiers before "^" was introduced by treating
	     these as if they were prefaced by "^" by re-entering while-loop
	     with "modifier" flag set TRUE and with fmt reset to original
	     character: */
	  modifier = TRUE;
	  fmt--;
	  /* Special case: substitute +TA, +TR  with  +T^a, +T^r */
	  if (icase == 'A') legacy_modifier = 'a';
	  if (icase == 'R') legacy_modifier = 'r';
	  /* No other legacy modifiers need substitution. */
	}
	break;	/* modifiers generate no output */

      default:
	/* A character within quotes is a directive only when preceded by %: */
	if (quote_mode && !percent_specified)
	{
	  as_is = TRUE;
	  bp = separate(bp, icase);
	  break;
	}

	/* Remove extraneous spacing if appropriate: */
	if ((fielding || (squeeze && !hard_width)) &&
	    (!zero_pad || width != 0))
	{
	  if (icase == MINS || icase == SECS || icase == YEARS_MOD_100)
	    /* MINS, SECS, YEARS_MOD_100 should never have a width less
	       than 2 so as to allow zero_pad: */
	    width = 2;
	  else
	    width = 0;
	  width_specified = TRUE;
	}

	/* Determine file's date using local time (if needed): */
	if (fdate == NULL)
	{
	  if (gmt)
	    fdate = gmtime32(&ftime);
	  else
	    fdate = localtime32(&ftime);
	}

	/* Process a directive: */
	switch (icase)
	{
	case ISO8601_DATE:
	  if (zero_st_info && ftime == 0)
	  {
	    /* If zero_st_info option specified, use all zeroes instead of
	       whatever the zero date translates to in the local time zone: */
	    if (!width_specified) width = 0;
	    sprintf(bp, "%*s",width, "00000000.000000");
	  }
	  else
	    bp = T_print_width(bp, "^N~YMD.hms~", ftime, fdate,
			       gmt, meridian, width);
	  break;

	case ELS_DATE:
	  bp = T_print_width(bp, "M%_DYt", ftime, fdate, gmt, meridian, width);
	  break;

	case LS_DATE:
	  bp = T_print_width(bp, "M%_DQ", ftime, fdate, gmt, meridian, width);
	  break;

	case DOS_DATE:
	  bp = T_print_width(bp, "^N%_M-D-y^M%_h:~mp~",
			     ftime, fdate, gmt, meridian, width);
	  break;

	case WINDOWS_DATE:
	  bp = T_print_width(bp, "^N%_M/D/y^M%_tP'M'",
			     ftime, fdate, gmt, meridian, width);
	  break;

	case VERBOSE_DATE:
	  bp = T_print_width(bp, "WM%_DTZY",
			     ftime, fdate, gmt, meridian, width);
	  break;

	case TIME_OR_YEAR:
#       define HALF_A_YEAR  (SECS_PER_YEAR/2)
	  if ((Ulong)The_Time - (Ulong)ftime < (Ulong)HALF_A_YEAR)
	    bp = T_print_width(bp, zero_pad ? "%0t" : "%_t",
			       ftime, fdate, gmt, meridian, width);
	  else
	  {
	    if (!width_specified && !(fielding || squeeze)) width = 5;
	    bp = T_print_width(bp, "Y", ftime, fdate, gmt, meridian, width);
	  }
	  break;

	case YEARS:
	case YEARS_MOD_100:
	  if (tics)
	  {
	    long r_years = ftime / SECS_PER_YEAR;
	    if (BOOLE_XOR(future, delta_time)) r_years = -r_years;
	    if (!width_specified) width = 2;
	    sprintf(bp, F_LD(zero_pad,width), r_years);
	  }
	  else
	  {
	    int year = fdate->tm_year;
	    if (icase == YEARS)
	    {
	      year += 1900;
	      if (!width_specified) width = 4;
	    }
	    else
	    {
	      year %= 100;
	      if (!width_specified) width = 2;
	    }
	    sprintf(bp, F_D(zero_pad,width), year);
	  }
	  break;

	case MONTHS:
	  if (tics)
	  {
	    long r_months = ftime / SECS_PER_MONTH;
	    if (BOOLE_XOR(future, delta_time)) r_months = -r_months;
	    if (!width_specified) width = 3;
	    sprintf(bp, F_LD(zero_pad,width), r_months);
	  }
	  else if (numeric)
	  {
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), fdate->tm_mon + 1);
	  }
	  else
	  {
	    if (!width_specified) width = 3;
	    sprintf(bp, "%*s",width, moy[fdate->tm_mon]);
	  }
	  break;

	case WEEKS:
	  if (tics)
	  {
	    long r_weeks = ftime / SECS_PER_WEEK;
	    if (BOOLE_XOR(future, delta_time)) r_weeks = -r_weeks;
	    if (!width_specified) width = 4;
	    sprintf(bp, F_LD(zero_pad,width), r_weeks);
	  }
	  else if (numeric)
	  {
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), fdate->tm_wday + 1);
	  }
	  else
	  {
	    if (!width_specified) width = 3;
	    sprintf(bp, "%*s",width, dow[fdate->tm_wday]);
	  }
	  break;

	case DAYS:
	  if (tics)
	  {
	    long r_days = ftime / SECS_PER_DAY;
	    if (BOOLE_XOR(future, delta_time)) r_days = -r_days;
	    if (!width_specified) width = 5;
	    sprintf(bp, F_LD(zero_pad,width), r_days);
	  }
	  else
	  {
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), fdate->tm_mday);
	  }
	  break;

	case HOURS:
	  if (tics)
	  {
	    long r_hours = ftime / SECS_PER_HOUR;
	    if (BOOLE_XOR(future, delta_time)) r_hours = -r_hours;
	    if (!width_specified) width = 6;
	    sprintf(bp, F_LD(zero_pad,width), r_hours);
	  }
	  else
	  {
	    int hour = fdate->tm_hour;
	    if (meridian)
	    {
	      if (hour > 12)
		hour -= 12;
	      else if (hour == 0)
		hour  = 12;
	    }
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), hour);
	  }
	  break;

	case 'p':
	case 'P':
	  if (! meridian)
	  {
	    carrot_msg(NULL, "+T", T_format,
		       "Must first specify meridian modifier", fmt-1);
	    exit(USAGE_ERROR);
	  }
	  else
	  {
	    char ap;
	    if (icase == 'p')
	      ap = (fdate->tm_hour < 12 ? 'a' : 'p');
	    else
	      ap = (fdate->tm_hour < 12 ? 'A' : 'P');
	    if (!width_specified) width = 1;
	    sprintf(bp, "%*c",width, ap);
	  }
	  break;

	case MINS:
	  if (tics)
	  {
	    long r_mins = ftime / SECS_PER_MIN;
	    if (BOOLE_XOR(future, delta_time)) r_mins = -r_mins;
	    if (!width_specified) width = 8;
	    sprintf(bp, F_LD(zero_pad,width), r_mins);
	  }
	  else
	  {
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), fdate->tm_min);
	  }
	  break;

	case SECS:
	  if (tics)
	  {
	    char *rsp, r_secs[32];
	    rsp = r_secs;
	    if (BOOLE_XOR(future, delta_time)) *rsp++ = '-';
	    if (!width_specified) width = 10;
	    /* Use "unsigned" format as this value may use all 32-bits: */
	    sprintf(rsp, F_time_t(FALSE,0), ftime);
	    sprintf(bp, "%*s",width, r_secs);
	  }
	  else
	  {
	    if (!width_specified) width = 2;
	    sprintf(bp, F_D(zero_pad,width), fdate->tm_sec);
	  }
	  break;

	case '2': /* '+T2' is DEPRECATED, use '+Tt' instead */
	case TIME_2:
	  bp = T_print_width(bp, zero_pad ? "%0h:m" : "%_h:m",
			     ftime, fdate, gmt, meridian, width);
	  break;

	case '3': /* '+T3' is DEPRECATED, use '+TT' instead */
	case TIME_3:
	  bp = T_print_width(bp, zero_pad ? "%0h:m:s" : "%_h:m:s",
			     ftime, fdate, gmt, meridian, width);
	  break;

	case ZONE_NAME:
	  if (!width_specified) width = 3;
#ifdef HAVE_TM_ZONE
	  /* tm_zone is right no matter what: */
	  sprintf(bp, "%*s",width, fdate->tm_zone);
#else
	  /* tzname is not always what we want: */
	  if (gmt)
	    sprintf(bp, "%*s",width, "GMT");
	  else
	  {
	    extern char *tzname[2];
	    sprintf(bp, "%*s",width, tzname[fdate->tm_isdst]);
	  }
#endif
	  break;

	default:
	  if (isalnum(icase))
	  {
	    carrot_msg(NULL, "+T", T_format, "Unrecognized directive", fmt-1);
	    exit(USAGE_ERROR);
	  }
	  else
	    /* Output non-directive/non-alphanumeric as-is: */
	    as_is = TRUE;
	  break;
	}

	percent_specified = FALSE;
	zero_pad = ZERO_PAD_DEFAULT;
	hard_width = FALSE;
	width_specified = FALSE;
	width = 0;
	if (!as_is) directive_seen = TRUE;
	bp = separate(bp, icase);
	break;
      }
    }
  }
  bp = finish(bp);

  if (quote_mode) error_msg("Missing delimiter in +T format");
  if (!directive_seen) error_msg("Incomplete +T format");

  return(bp);
}
#undef  ZERO_PAD_DEFAULT


char *T_print_width(char *buff, char *fmt, time_t ftime,
		    struct tm *fdate, Boole gmt, Boole meridian, int width)
{
  if (width == 0)
    buff = T_print(buff, fmt, ftime, fdate, gmt, meridian);
  else
  {
    char tmp[MAX_INFO];
    Void T_print(tmp, fmt, ftime, fdate, gmt, meridian);
    sprintf(buff, "%*s",width, tmp);
    buff += strlen(buff);
  }
  return(buff);
}


char *separate(char *bp, char icase)
{
  bp += strlen(bp); /* Find where we are */
  if (as_is)
  {      
    /* Remove separation when it precedes an "as-is" character: */
    if (separated)
    {
      bp--;
      separated = FALSE;
    }
    
    /* Add the character as-is: */
    *bp++ = icase;  *bp = CNULL;
    as_is = FALSE;
  }
  else if (squeeze)
  {
    /* If squeezing then don't inject a separator: */
    separated = FALSE;
  }
  else
  {
    *bp++ = separator; *bp = CNULL;
    separated = TRUE;
  }

  return(bp);
}


char *finish(char *bp)
{
  bp += strlen(bp); /* Find where we are */
  /* Remove final separation: */
  if (separated)
  {
    bp--; *bp = CNULL;
    separated = FALSE;
  }

  return(bp);
}


char *full_name(char *dname,
		char *fname)
{
  /* Be careful that the return value doesn't get overworked! */
  static char full_name[MAX_FULL_NAME];
  
  /* Gather any information on this parameter */
  if (dname[0] == CNULL || strcmp(dname, ".") == 0)
  {
    return(fname);
  }
  else
  {
    if (strcmp(dname, "/") == 0) dname = "";
    Void sprintf(full_name, "%s/%s", dname, fname);
    return(full_name);
  }
}


#define DIR_LIST_CHUNK  (2048/sizeof(Dir_Item))
Dir_List dir_item_list = {NULL, NULL};
Dir_Item *dir_item_alloc(void)
{
  register Dir_Item *file;
  if (dir_item_list.head == NULL)
  {
    dir_item_list.head = mem_alloc(DIR_LIST_CHUNK*sizeof(Dir_Item));
    dir_item_list.tail = dir_item_list.head+DIR_LIST_CHUNK-1;
    for (file = dir_item_list.head; file != dir_item_list.tail; file++)
      file->next = file+1;
    file->next = NULL;
  }

  file = dir_item_list.head;
  dir_item_list.head = dir_item_list.head->next;
  return(file);
}


void dir_item_free(Dir_Item *file)
{
  if (dir_item_list.head == NULL)
    dir_item_list.head = file;
  else
    dir_item_list.tail->next = file;
    
  dir_item_list.tail = file;
  file->next = NULL;
  return;
}


void filter_error_msg(char *msg, char *ptr)
{
  carrot_msg(NULL, Current_Opt, Current_Arg, msg, ptr);
  exit(USAGE_ERROR);
}


Boole filter_warning_msg(char *msg, char *ptr)
{
# define MAX_WARN 16
  static char *warned[MAX_WARN];
  static int nwarned = 0;
  int i;
  if (nwarned < MAX_WARN)
  {
    Boole found = FALSE;
    for (i = 0; i < nwarned && !found; i++)
      found = warned[i] == ptr;
    if (!found)
    {
      carrot_msg("Warning", Current_Opt, Current_Arg, msg, ptr);
      warned[nwarned++] = ptr;
      return(TRUE);
    }
  }
  else
    filter_error_msg(msg, ptr);

  return(FALSE);
}


void redundancy_warning(char *types, char *ptr)
{
  if (filter_warning_msg("Redundant term", ptr))
    fprintf(stderr, "\
Perhaps you meant to OR the two preceding %s instead of ANDing?\n\
\n", types);
  return;
}


void expecting_error(Boole negate, char *items, Boole logic, Boole end,
		     char *ptr)
{
  char msg[128];
  char choices[128];
  choices[0] = CNULL;
  if (items)
  {
    /*if (negate) strcat(choices, "[!~]");*/
    if (negate) strcat(choices, "!~");
    strcat(choices, items);
  }
  if (logic)
  {
    if (items)
      strcat(choices, "' or '");
    strcat(choices, "&,|:");
  }
  if (end)
  {
    if (items || logic)
      strcat(choices, "' or '");
    strcat(choices, "}");
  }

  if (logic || (items && strlen(items) > 1))
    sprintf(msg, "Expected one of '%s'", choices);
  else
    sprintf(msg, "Expected '%s'", choices);
  filter_error_msg(msg, ptr);
  return;
}


void watchSigHandler(int sig)
{
#ifdef SIG_IGN
  signal(WATCH_SIGNAL, SIG_IGN);
#endif
  sigEvent = TRUE;
  fprintf(ttyout, "\rDIR == %s\n", CwdPath);

#if WATCH_SIGNAL == SIGINT
  if (sigAskAbort && !sigAskNoMore)
  {
    char resp[10];
    fprintf(ttyout, "Abort? [ynN]: ");
    fgets(resp, 9, ttyin);
    if (resp[0] == 'y')
    {
      if (defaultSigHandler != NULL) defaultSigHandler();
      exit(0);
    }
    else if (resp[0] == 'N')
      sigAskNoMore = TRUE;
  }
  /* Next time ask to abort if ^C occurs more than once during an interval: */
  sigAskAbort = TRUE;
#endif

  signal(WATCH_SIGNAL,  watchSigHandler);
  return;
}


int sigSafe_stat(char *path, struct stat *buf)
{
  int sts;
  do {
    int save_errno = errno;
    sigEvent = FALSE;
    sts = stat(path, buf);
    /*if (sigEvent) fprintf(ttyout, "SS_stat errno == %d\n", errno);*/
    if (sigEvent) errno = save_errno;
  } while (sigEvent);
  return(sts);
}


int sigSafe_lstat(char *path, struct stat *buf)
{
  int sts;
  do {
    int save_errno = errno;
    sigEvent = FALSE;
    sts = lstat(path, buf);
    /*if (sigEvent) fprintf(ttyout, "SS_lstat errno == %d\n", errno);*/
    if (sigEvent) errno = save_errno;
  } while (sigEvent);
  return(sts);
}


void spin_wheel(void)
{
  static int i = 0;
  char *wheel = "/-\\|";
  fprintf(ttyout, "%c\b", wheel[i]);
  fflush(ttyout);
  i = (i+1) & 3;
  return;
}
