/******************************************************************************
  elsMisc.c -- els miscellaneous functions
  
  Author: Mark Baranowski (markb@elen.utah.edu)

  Last significant change: August 9, 2000
  
  These functions are provided "as is" in the hopes that they might serve some
  higher purpose.  If you want these functions to serve some lower purpose,
  then that too is all right.  So as to keep these hopes alive you may
  freely distribute these functions so long as this header remains intact.
  You may also freely distribute modified versions of these functions so long
  as you indicate that such versions are modified and so long as you
  provide access to the unmodified original copy.
  
  Note: The most recent version of these functions may be obtained from
  ftp://auriga.elen.utah.edu/pub/markb/els.tar.Z

  ****************************************************************************/


#include "sysdefs.h"

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>	/* sys/stat.h needed by elsMisc.h */
#include <pwd.h>
#include <grp.h>
#include <string.h>

#include "defs.h"
#include "aux.h"
#include "els.h"
#include "elsMisc.h"


/******************************************************************************
 * wildcard_match: Test whether text matches a wildcard pattern using the
 * same syntax and semantics as is used by /bin/csh.
 *
 *  ?	Scan a single text character
 *  *	Scan zero or more text characters
 *  []	Test whether a text character matches a set or a range of chars.  For
 *	example, match A through M, all digits, or the set "!@#": [A-M0-9!@#]
 *  [^]	Same as above but negate the result.  For example, match all
 *	characters EXCEPT A through M, digits, or the set "!@#": [^A-M0-9!@#]
 *  {}	Match any of the enclosed substrings separated by commas
 *  \	Match following character verbatim
 *  All other characters not listed above are matched verbatim.
 *
 *  NB: The meta characters ?*[]{,}\ might need to be prefaced by \ in certain
 *	contexts if a LITERAL match is desired.  For example, if attempting
 *	to LITERALLY match any of ?*[{} then these must be prefaced by \ in
 *	all contexts except when inside [...].  If attempting to LITERALLY
 *	match "," then it needs to be prefaced by \ only when inside {...}.
 *	Finally, if attempting to LITERALLY match ] or \ then these should
 *	always be prefaced by \ regardless of the context.
 *
 * EXAMPLES:
 *  Test whether text starts with a letter or digit:
 *	wildcard_match(text, "[a-zA-Z0-9]*");
 *  Test whether text contains at least three or more characters:
 *	wildcard_match(text, "???*");
 *  Test whether text is the name of a *.c or *.h file:
 *	wildcard_match(text, "*.[ch]");
 *  Test whether text is the name of a *.c, *.h, *.cc, or *.hh file:
 *	wildcard_match(text, "*.{[ch],cc,hh}");
 *  Test whether text contains the substring "Markb" or "markb":
 *	wildcard_match(text, "*[Mm]arkb*");
 *	wildcard_match(text, "*{Markb,markb}*");
 *  Test whether text matches any of ab*ch, ab*dh, aeeffh, aeegg*h, or aeeh:
 *	wildcard_match(text, "a{b*[cd],ee{ff,gg*,}}h");
 */

Boole wildcard_match(char *text, char *pattern)
{
  for (;;)
  {
    switch (*pattern)
    {
    case NULL:
      /* END OF PATTERN: */
      return(*text == CNULL);
      break;

    case '?':
      /* SCAN ONE CHAR OF ANY KIND: */
      if (*text == CNULL) return(FALSE);
      text++;
      pattern++;
      break;

    case '[':
      /* MATCH ONE CHAR FROM A SET OR A RANGE: */
      {
	Boole match = FALSE;
	Boole first = TRUE;
	Boole negate;
	char lastpat = CNULL;
	pattern++;
	negate = (*pattern == '^');
	if (negate) pattern++;
	while (*pattern != CNULL && *pattern != ']')
	{
	  if (*pattern == '\\')
	  {
	    pattern++;
	    if (*pattern == CNULL) return(FALSE); /* Missing ] */
	    match |= (*text == *pattern);
	  }
	  else if (*pattern == '-' && !first)
	  {
	    pattern++;
	    if (*pattern == CNULL) return(FALSE); /* Missing ] */
	    match |= (lastpat <= *text && *text <= *pattern);
	  }
	  else
	  {
	    match |= (*text == *pattern);
	  }
	  lastpat = *pattern;
	  first = FALSE;
	  pattern++;
	}
	if (*pattern == CNULL) return(FALSE); /* Missing ] */
	match ^= negate;
	if (!match) return(FALSE);
      }
      text++;
      pattern++;
      break;

    case '*':
      /* SCAN ZERO OR MORE CHARS: */
      while (*pattern == '*') pattern++;
      if (*pattern == CNULL) return(TRUE);

      /* Speed things up by scanning ahead for the first match of a REGULAR
	 CHAR.  This step is optional and could be skipped without affecting
	 the final results of this algorithm: */
      if (IS_NOT_MEMBER(*pattern, "?[\\{"))
	while (*text != CNULL && *text != *pattern) text++;

      /* Recursively look for consecutive groups of matches.  If we return
	 to the top level without exhausting the text, we succeeded: */
      while (*text != CNULL && !wildcard_match(text, pattern)) text++;

      return(*text != CNULL);
      break;

    case '{':
      /* RECURSIVELY SCAN SUBSTRINGS: */
      {
	char *subptr;
	char *substart = NULL;
	int depth;
	char *poststart;
	int postlen;

	subptr = pattern;
	depth = 0;
	do
	{
	  if (*subptr == '\\')
	    subptr++;
	  else if (*subptr == '{')
	    depth++;
	  else if (*subptr == '}')
	    depth--;
	  if (*subptr == CNULL) return(FALSE); /* Missing } */
	  subptr++;
	} while (depth > 0);
	poststart = subptr;
	postlen = strlen(poststart);

	subptr = pattern;
	depth = 0;
	do
	{
	  Boole doit = FALSE;
	  if (*subptr == '\\')
	    subptr++;
	  else if (*subptr == '{')
	  {
	    depth++;
	    if (depth == 1) substart = subptr + 1;
	  }
	  else if (*subptr == ',')
	  {
	    if (depth == 1) doit = TRUE;
	  }
	  else if (*subptr == '}')
	  {
	    if (depth == 1) doit = TRUE;
	    depth--;
	  }
	  if (doit)
	  {
#	    define NEWBUF_SIZE  16
	    Boole match;
	    char *newpat, newbuf[NEWBUF_SIZE];
	    int sublen, newlen;

	    sublen = subptr - substart;
	    newlen = sublen + postlen + 1;
	    newpat = newbuf;
	    if (newlen > NEWBUF_SIZE) newpat = mem_alloc(newlen);

	    strncpy(newpat, substart, sublen);
	    strcpy(&newpat[sublen], poststart);
	    match = wildcard_match(text, newpat);

	    if (newlen > NEWBUF_SIZE) mem_free(newpat);
	    if (match) return(TRUE);
	    substart = subptr + 1;
	  }
	  subptr++;
	} while (depth > 0);
      }
      return(FALSE);
      break;

    case '\\':
      /* MATCH ONE SPECIAL CHAR: */
      pattern++;
      if (*text == CNULL) return(FALSE);
      if (*text != *pattern) return(FALSE);
      text++;
      if (*pattern != CNULL) pattern++; /* consider pattern of: "xyz\^@abc" */
      break;

    default:
      /* MATCH ONE REGULAR CHAR: */
#ifdef REQUIRE_RH_QUOTES
      /* Define REQUIRE_RH_QUOTES to  require \] and \} quoting
	 (wildcard matching under /bin/csh works unquoted): */
      if (*text == ']') return(FALSE); /* Missing [ */
      if (*text == '}') return(FALSE); /* Missing { */
#endif
      if (*text == CNULL) return(FALSE);
      if (*text != *pattern) return(FALSE);
      text++;
      pattern++;
      break;
    }
  }
  /*NOTREACHED*/
}


char *uid2name(uid_t uid)
{
/* UID_CACHE_SIZE must be a power of two: */
# define UID_CACHE_SIZE 64
  static uid_t uids[UID_CACHE_SIZE];
  static char  name[UID_CACHE_SIZE][MAX_USER_GROUP_NAME+1];
  static Boole first_time = TRUE;
  struct passwd *passwd;
  int i;

  if (first_time)
  {
    /* The zero-th entry must initially be non-zero and all other entries
       must initially be zero in order for the cache to function: */
    uids[0] = 1;
    first_time = FALSE;
  }

  i = uid & (UID_CACHE_SIZE-1);
  if (uids[i] != uid)
  {
    if ((passwd = getpwuid(uid)) != NULL)
      strncpy(name[i], passwd->pw_name, MAX_USER_GROUP_NAME);
    else
      name[i][0] = CNULL;
    uids[i] = uid;
  }

  if (name[i][0] == CNULL)
    return(NULL);
  else
    return(name[i]);
}


char *gid2name(gid_t gid)
{
/* GID_CACHE_SIZE must be a power of two: */
# define GID_CACHE_SIZE 32
  static gid_t gids[GID_CACHE_SIZE];
  static char  name[GID_CACHE_SIZE][MAX_USER_GROUP_NAME+1];
  static Boole first_time = TRUE;
  struct group *group;
  int i;

  if (first_time)
  {
    /* The zero-th entry must initially be non-zero and all other entries
       must initially be zero in order for the cache to function: */
    gids[0] = 1;
    first_time = FALSE;
  }

  i = gid & (GID_CACHE_SIZE-1);
  if (gids[i] != gid)
  {
    if ((group = getgrgid(gid)) != NULL)
      strncpy(name[i], group->gr_name, MAX_USER_GROUP_NAME);
    else
      name[i][0] = CNULL;
    gids[i] = gid;
  }

  if (name[i][0] == CNULL)
    return(NULL);
  else
    return(name[i]);
}


Ulong checksum(char *filename)
{
  FILE *fp;
  Ulong checksum;
  int c;

  checksum = 0;
  
  /* PERFORM 16-BIT CHECKSUM (same algorithm as /bin/sum): */
  fp = fopen(filename, "r");
  if (fp != NULL)
  {
    Boole rotate;
    while ((c = getc(fp)) != EOF)
    {
      rotate = (checksum & 1) != 0;
      checksum >>= 1;
      if (rotate) checksum |= 0x8000;
      checksum += c;
      checksum &= 0xffff;
    }
    fclose(fp);
  }
  
  return(checksum);
}


int strToVer(char *str, char **ptr)
{
  int major = 0;
  int minor = 0;
  major = strtol(str, &str, 10);
  if (major < 100) major *= 100;
  if (*str == '.')
  {
    str++;
    if (isdigit(*str)) minor += (*str++ - '0') * 10;
    if (isdigit(*str)) minor += (*str++ - '0');
  }
  if (ptr) *ptr = str;
  return(major + minor);
}


char *verToStr(int ver)
{
  char *ver_str = mem_alloc(16);
  int major = ver / 100;
  int minor = ver - major * 100;
  if (major > 99) major = 99;
  if (minor > 99) minor = 99;
  sprintf(ver_str, "%d.%02d", major, minor);
  return(ver_str);
}


void print_args(char **argv, int argc, int version)
{
  int i, nout;
  for (i = 0, nout = 0; i < argc && nout < 128; i++)
  {
    char *arg = quote_str(argv[i], NULL, QS_ARG);
    if (arg == NULL) arg = argv[i];
    printf(" %s", arg);
    nout += strlen(arg);
    if (i == 0 && version != 0) printf(" +v%s", verToStr(version));
  }
  if (i < argc) printf(" ...");
}
