/******************************************************************************
  sysdep.c -- System dependencies
  
  Author: Mark Baranowski (markb@elen.utah.edu)

  Last significant change: March 14, 2001
  
  These functions are provided "as is" in the hopes that they might serve some
  higher purpose.  If you want these functions to serve some lower purpose,
  then that too is all right.  So as to keep these hopes alive you may
  freely distribute these functions so long as this header remains intact.
  You may also freely distribute modified versions of these functions so long
  as you indicate that such versions are modified and so long as you
  provide access to the unmodified original copy.
  
  Note: The most recent version of these functions may be obtained from
  ftp://auriga.elen.utah.edu/pub/markb/els.tar.Z

  ****************************************************************************/

#include "sysdefs.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
#include <utime.h>
#include <dirent.h>
#include <sys/param.h>
#include <unistd.h>
#include <sys/utsname.h>

#if defined(HAVE_ACL)
#  if defined(SUNOS)
#    include <sys/acl.h>
#  endif
#  if defined(HPUX)
#    include <sys/acl.h>
#  endif
#  if defined(OSF1)
#    include <sys/acl.h>
#  endif
#  if defined(AIX)
#    include <acl.h>
#    include <sys/acl.h>
#  endif
#  if defined(IRIX)
#    include <sys/acl.h>
#  endif
#endif

#include "defs.h"
#include "aux.h"
#include "els.h"
#include "elsMisc.h"
#include "sysdep.h"

Ulong osVersion;


char *get_userAtHost(void)
{
  static char *userAtHost = NULL;
  if (userAtHost == NULL)
  {
    char *user = uid2name(Whoami);
    struct utsname host;
    uname(&host);
    userAtHost = mem_alloc(strlen(user) + strlen(host.nodename) + 4);
    sprintf(userAtHost, "%s@%s", user, host.nodename);
  }
  return(userAtHost);
}


Ulong get_osVersion(void)
{
  Ulong release, revision, minor_rev;
  char *cp;
  struct utsname name;
  uname(&name);
  cp = name.release;
  /* Some OSes (e.g. HPUX) preface the release with letters: */
  while (*cp && !isdigit(*cp)) cp++;
  release = strtoul(cp, &cp, 10);
  if (*cp == '.') cp++;
  revision = strtoul(cp, &cp, 10);
  if (*cp == '.') cp++;
  minor_rev = strtoul(cp, &cp, 10);
  osVersion = release * 10000 + revision * 100 + minor_rev;
  return(osVersion);
}


void show_options(void)
{
  if (get_osVersion() > 0)
    printf("	Currently running os version: %lu\n", osVersion);
#ifdef SUNOS
  printf("	Compiled with SUNOS=%d\n", SUNOS);
#endif
#ifdef NO_ST_BLOCKS
  printf("	Compiled with NO_ST_BLOCKS\n");
#endif
#ifdef HAVE_TM_ZONE
  printf("	Compiled with HAVE_TM_ZONE\n");
#endif
#ifdef HAVE_ACL
  printf("	Compiled with HAVE_ACL\n");
#else
  printf("	Compiled *without* HAVE_ACL\n");
#endif
#ifdef HAVE_STAT64
  printf("	Compiled with HAVE_STAT64\n");
#else
  printf("	Compiled *without* HAVE_STAT64\n");
#endif
  return;
}


int get_acl_count(Dir_Item *file, char *path)
{
  int aclcnt = 0;

  /* Special consideration for symbolic links:
     For now treat all symbolic links as having a 0 ACL count based upon
     the following information:
     1) SunOS5.7's /bin/ls treats the ACL count of all symbolic links
     as 0 unless -L specified.
     2) OSF1's acl_get_file() generates a Segmentation fault if acl_get_file()
     called on a symbolic link that points to a non-existent file. */
  if (file->islink && !expand_symlink) return(0);

#ifdef HAVE_ACL

# if defined(SUNOS)
  /* Determine at runtime whether the version of SunOS being run supports
     ACLs.  This is necessary so that if the executable is compiled under
     SunOS5.5 using shared objects, then the *same* executable can be used
     on other machines running SunOS5.4.  This feat is accomplished by
     avoiding calls to getacl whenever necessary: */
  if (osVersion >= 50500)
  {
    /* NB: acl returns -1 if error */
    aclcnt = acl(path, GETACLCNT, 0, NULL);
    if (aclcnt > MIN_ACL_ENTRIES)
      aclcnt -= MIN_ACL_ENTRIES;
    else
      aclcnt = 0;
  }
# endif

# if defined(HPUX)
  aclcnt = getacl(path, 0, NULL);
  if (aclcnt > NBASEENTRIES)
    aclcnt -= NBASEENTRIES;
  else
    aclcnt = 0;
# endif

# if defined(OSF1)
  {
    acl_t acl = acl_get_file(path, ACL_TYPE_ACCESS);
    aclcnt = acl->acl_num;
    if (aclcnt > ACL_MIN_ENTRIES)
      aclcnt -= ACL_MIN_ENTRIES;
    else
      aclcnt = 0;
  }
# endif

# if defined(AIX)
  /* ??? this needs to be retested ??? */
  {
    struct acl pAcl;
    statacl(path, STX_NORMAL, &pAcl, sizeof(pAcl));
    if (pAcl.acl_mode & S_IXACL)
      aclcnt = 1;	/* There may be more than 1 ACL ??? */
  }
# endif

#endif

  return(aclcnt);
}


BLKCNT_T get_stat_blocks(struct stat *info)
{
#ifdef NO_ST_BLOCKS
  /* Crude approximation to what the real st_blocks should be: */
  /* ??? this needs to be retested ??? */
  return((info->st_size + 511)/512);
#else
  return(info->st_blocks);
#endif
}

/*===========================================================================*/

#ifdef NO_ST_DEV
struct stat *set_mp(struct stat *dir)
{
  return;
}

int diff_mp(struct stat *dir)
{
  return(0);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
#else

Local struct stat *mp;
struct stat *set_mp(struct stat *dir)
{
  struct stat *old_mp;
  /*printf("SET: %x\n", dir->st_dev);*/
  old_mp = mp;
  mp = dir;
  return(old_mp);
}

int diff_mp(struct stat *dir)
{
  int diff;
  /*printf("DIFF: %x %x\n", mp->st_dev, dir->st_dev);*/
  diff = memcmp(&mp->st_dev, &dir->st_dev, sizeof(dev_t));
  return(diff);
}

#endif
/*===========================================================================*/
